package teajay.bytecode;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.fields.FieldAccessFlag;
import teajay.bytecode.methods.MethodAccessFlag;

/**
 * Takes care of correct I/O of access flags.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class AccessFlagsIO {
    
    @SafeVarargs
    @SuppressWarnings({"unchecked", "varargs"})
    public static <T extends AccessFlag> void write(DataOutputStream out, T... flags) throws IOException {
        write(out, Arrays.asList(flags));
    }
    
    /**
     * Writes access flags as a 16-bit bitmask.
     * @param <T> type of access flags
     * @param out the target stream
     * @param flags list of flags to be written
     * @throws IOException if <tt>out</tt> throws a IOException
     */    
    public static <T extends AccessFlag> void write(DataOutputStream out, Collection<T> flags) throws IOException {
        int result = 0;
        for (AccessFlag flag : flags) {
            result |= flag.getIntValue();
        }
        out.writeShort(result);
    }

    private static <T extends AccessFlag> Collection<T> parse(int bits, Class<T> type) throws ByteCodeException {
        if ((bits >> 16) != 0) {
            throw new ByteCodeException("Invalid access_flags mask: " + Integer.toHexString(bits));
        }
        
        List<T> result = new LinkedList<>();
        T[] flags = type.getEnumConstants();
        for (T flag : flags) {
            if ((flag.getIntValue() & bits) != 0) {
                result.add(flag);
            }
        }
        //PACKAGE_PRIVATE handling
        if (VisibilityAccessFlags.checkPackagePrivate(bits)) {
            for (T flag : flags) {
                if (flag.getIntValue() == VisibilityAccessFlags.PACKAGE_PRIVATE.intValue) {
                    result.add(flag);
                }
            }
        }
        return result;
    }
    /**
     * Reads ClassAccessFlags from a 16-bit bitmask.
     * 
     * @param bits a 16-bit bitmask
     * @return a collection of access flags that this this mask represents.
     * @throws ByteCodeException  if any bits above the 16th are set in <tt>bits</tt>
     */
    public static Collection<ClassAccessFlag> parseClassAccessFlags(int bits) throws ByteCodeException {
        return parse(bits, ClassAccessFlag.class);
    }
    /**
     * @see #parseClassAccessFlags(int) 
     * @param bits
     * @return
     * @throws ByteCodeException 
     */
    public static Collection<FieldAccessFlag> parseFieldAccessFlags(int bits) throws ByteCodeException {
        return parse(bits, FieldAccessFlag.class);
    }
    /**
     * @see #parseClassAccessFlags(int) 
     * @param bits
     * @return
     * @throws ByteCodeException 
     */
    public static Collection<MethodAccessFlag> parseMethodAccessFlags(int bits) throws ByteCodeException {
        return parse(bits, MethodAccessFlag.class);
    }
    /**
     * @see #parseClassAccessFlags(int) 
     * @param bits
     * @return
     * @throws ByteCodeException 
     */
    public static Collection<InnerClassAccessFlag> parseInnerClassAccessFlags(int bits) throws ByteCodeException {
        return parse(bits, InnerClassAccessFlag.class);
    }
}
