package teajay.bytecode;

import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.*;
import teajay.bytecode.VersionConstants.JavaMajorVersion;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.attributes.AttributesReader;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.MethodDescriptor;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.fields.FieldInfo;
import teajay.bytecode.methods.MethodInfo;
import teajay.bytecode.util.Tools;
import static teajay.bytecode.util.Tools.ccpt;
import static teajay.bytecode.util.Tools.checkUnqualifiedClassName;

/**
 * Implements a whole class file. There is no plausability checking for class
 * file versions other than 51.0 (java 7.0). The default major and minor version
 * numbers are 51 resp. 0.
 * Support for everything above Java 7 is limited!
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class ClassFile {

    public enum ReadMode {

        /**
         * The this and super class are always saved. This flag only exists to
         * make it possible to solely read the this and super class.
         */
        ThisAndSuper,
        /**
         * Signifies that the class file's interfaces will be saved.
         */
        Interfaces,
        /**
         * Signifies that the class file's fields will be saved.
         */
        Fields,
        /**
         * Signifies that the class file's methods will be saved.
         */
        Methods,
        /**
         * Signifies that the class file's attributes will be saved.
         */
        Attributes,
        /**
         * The default mode, all informations will be saved.
         */
        All,
        /**
         * As this framework is not yet complete, purging the constant pool of a
         * class file that has been read from a file will prevent the class file
         * to be written properly by the <tt>write(DataOutputStream)</tt>
         * method.
         */
        ForgetConstantPool;
    }
    private final ConstantPool cpool;
    private final Collection<ClassAccessFlag> accessFlags;
    private final ClassInfo thisClass;
    private ClassInfo superClass;
    private final Map<String, FieldInfo> fields;
    private final Map<String, List<MethodInfo>> methods;
    private final Map<String, AttributeInfo> attributes;
    private final Collection<ClassInfo> interfaces;
    private final U2 minorVersion;
    private final U2 majorVersion;

    /**
     * Defaults superName to <tt>java/lang/Object</tt> .
     *
     * @see #ClassFile(teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.VersionConstants.JavaMajorVersion, teajay.bytecode.U2,
     * teajay.bytecode.ClassAccessFlag[])
     * @param className
     * @param accessFlags
     * @throws ByteCodeException
     */
    public ClassFile(String className, ClassAccessFlag... accessFlags) throws ByteCodeException {
        this(className, "java/lang/Object", accessFlags);
    }

    /**
     * Defaults superName to <tt>java/lang/Object</tt> .
     *
     * @see #ClassFile(teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.VersionConstants.JavaMajorVersion, teajay.bytecode.U2,
     * teajay.bytecode.ClassAccessFlag[])
     * @param className
     * @param major
     * @param minor
     * @param accessFlags
     * @throws ByteCodeException
     */
    public ClassFile(String className, JavaMajorVersion major, U2 minor, ClassAccessFlag... accessFlags) throws ByteCodeException {
        this(new ClassInfo(className), new ClassInfo("java/lang/Object"), major, minor, accessFlags);
    }

    /**
     * Defaults superName to <tt>java/lang/Object</tt> .
     *
     * @see #ClassFile(teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.VersionConstants.JavaMajorVersion, teajay.bytecode.U2,
     * teajay.bytecode.ClassAccessFlag[])
     * @param className
     * @param major
     * @param minor
     * @param accessFlags
     * @throws ByteCodeException
     */
    public ClassFile(ClassInfo className, JavaMajorVersion major, U2 minor, ClassAccessFlag... accessFlags) throws ByteCodeException {
        this(className, new ClassInfo("java/lang/Object"), major, minor, accessFlags);
    }

    /**
     * Defaults superName to <tt>java/lang/Object</tt> .
     *
     * @see #ClassFile(teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.VersionConstants.JavaMajorVersion, teajay.bytecode.U2,
     * teajay.bytecode.ClassAccessFlag[])
     * @param className
     * @param accessFlags
     * @throws ByteCodeException
     */
    public ClassFile(ClassInfo className, ClassAccessFlag... accessFlags) throws ByteCodeException {
        this(className, new ClassInfo("java/lang/Object"), accessFlags);
    }

    /**
     * @see #ClassFile(teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.VersionConstants.JavaMajorVersion, teajay.bytecode.U2,
     * teajay.bytecode.ClassAccessFlag[])
     * @param className
     * @param superName
     * @param accessFlags
     * @throws ByteCodeException
     */
    public ClassFile(String className, String superName, ClassAccessFlag... accessFlags) throws ByteCodeException {
        this(new ClassInfo(className), superName != null ? new ClassInfo(superName) : null, accessFlags);
    }

    /**
     * @see #ClassFile(teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.VersionConstants.JavaMajorVersion, teajay.bytecode.U2,
     * teajay.bytecode.ClassAccessFlag[])
     * @param className
     * @param superName
     * @param accessFlags
     * @throws ByteCodeException
     */
    public ClassFile(ClassInfo className, ClassInfo superName, ClassAccessFlag... accessFlags) throws ByteCodeException {
        this(className, superName, JavaMajorVersion.Java7_0, VersionConstants.J7_MINOR, accessFlags);
    }

    /**
     * Creates a new class file. By default all class files have the ACC_SUPER
     * flag set. You can remove it using {@link #removeSuperAccessFlag() }.
     *
     * @param className the new class's fully qualified name.
     * @param superName The class's super class. This may only be null for the
     * class name java/lang/Object.
     * @param major the major version of this class file (all except Java7 are
     * experimental)
     * @param minor the minor version (all except 0 are experimental)
     * @param accessFlags the class's access flags
     * @throws ByteCodeException if the class name is invalid
     */
    public ClassFile(ClassInfo className, ClassInfo superName, JavaMajorVersion major, U2 minor, ClassAccessFlag... accessFlags) throws ByteCodeException {
        this.cpool = new ConstantPool();
        checkName(className);
        this.thisClass = className;
        if (superName != null) {
            checkName(superName);
        }
        this.superClass = superName;
        this.fields = new HashMap<>();
        this.methods = new HashMap<>();
        this.attributes = new HashMap<>();
        this.interfaces = new HashSet<>();
        this.accessFlags = Collections.unmodifiableCollection(Arrays.asList(VisibilityAccessFlags.handlePackagePrivate(accessFlags, ClassAccessFlag.class)));
        this.minorVersion = minor;
        this.majorVersion = major.value;
    }

    private ClassFile(ClassInfo thisClass, ClassInfo superClass, Collection<ClassAccessFlag> accessFlags, ConstantPool cpool, int major, int minor) throws ByteCodeException {
        this.cpool = cpool;
        this.thisClass = thisClass;
        this.superClass = superClass;
        this.fields = new HashMap<>();
        this.methods = new HashMap<>();
        this.attributes = new HashMap<>();
        this.interfaces = new HashSet<>();
        this.accessFlags = accessFlags;
        this.minorVersion = new U2(minor);
        this.majorVersion = new U2(major);
    }

    private static void checkName(ClassInfo className) throws ByteCodeException {
        if (!checkUnqualifiedClassName(className.getValue())) {
            throw new ByteCodeException("invalid class name: " + className.getClassName());
        }
    }

    public U2 getMinorVersion() {
        return minorVersion;
    }

    public U2 getMajorVersion() {
        return majorVersion;
    }

    public ClassInfo getThis() {
        return thisClass;
    }

    public ClassInfo getSuper() {
        return superClass;
    }

    public Collection<ClassAccessFlag> getAccessFlags() {
        return Collections.unmodifiableCollection(accessFlags);
    }

    public Collection<FieldInfo> getFields() {
        return Collections.unmodifiableCollection(fields.values());
    }

    public Collection<List<MethodInfo>> getMethods() {
        return Collections.unmodifiableCollection(methods.values());
    }

    public Collection<ClassInfo> getInterfaces() {
        return Collections.unmodifiableCollection(interfaces);
    }

    public Collection<AttributeInfo> getAttributes() {
        return Collections.unmodifiableCollection(attributes.values());
    }

    public AttributeInfo getAttribute(String name) {
        return attributes.get(name);
    }

    public void removeAttribute(String name) {
        attributes.remove(name);
    }

    public FieldInfo getField(String name) {
        return fields.get(name);
    }

    public List<MethodInfo> getMethods(String name) {
        return methods.get(name);
    }

    public boolean implementsInterface(ClassInfo name) {
        return interfaces.contains(name);
    }

    public void addInterface(ClassInfo iface) {
        interfaces.add(iface);
    }

    public void addInterfaces(ClassInfo... ifaces) {
        addInterfaces(Arrays.asList(ifaces));
    }

    public void addInterfaces(Collection<ClassInfo> ifaces) {
        for (ClassInfo iface : ifaces) {
            addInterface(iface);
        }
    }

    public boolean removeInterface(ClassInfo iface) {
        return interfaces.remove(iface);
    }

    public ConstantPool getConstantPool() {
        return cpool;
    }

    public void setSuper(ClassInfo superClass) throws ByteCodeException {
        if (superClass != null) {
            checkName(superClass);
        }
        this.superClass = superClass;
    }

    public void setSuper(String superClassName) throws ByteCodeException {
        setSuper(new ClassInfo(superClassName));
    }

    public void addField(FieldInfo fieldInfo) throws ByteCodeException {
        String name = fieldInfo.getName().getValue();
        if (fields.containsKey(name)) {
            throw new ByteCodeException("class file " + thisClass.getClassName() + " already contains a field named " + fieldInfo.getName().getValue());
        } else {
            fields.put(name, fieldInfo);
        }
    }
    
    public boolean hasMethod(String name, MethodDescriptor descr) {
        List<MethodInfo> vals = methods.get(name);
        if (vals != null) {
            for (MethodInfo info : vals) {
                if (info.getDescriptor().equals(descr)) {
                    return true;
                }
            }
        }
        return false;
    }

    public void addMethod(MethodInfo method) throws ByteCodeException {
        if (!method.getDefiningClass().equals(thisClass)) {
            throw new ByteCodeException("defining class missmatch while adding method " + method.getName());
        }
        String name = method.getName().getValue();
        List<MethodInfo> vals = methods.get(name);
        if (vals == null) {
            vals = new ArrayList<>();
            methods.put(name, vals);
        }
        if (!vals.contains(method)) {
            vals.add(method);
        } else {
            throw new ByteCodeException("class file " + thisClass.getClassName() + " already contains a method " + method);
        }
    }

    public void addAttribute(AttributeInfo attribute) throws ByteCodeException {
        String name = attribute.getName().getValue();
        if (attributes.containsKey(name)) {
            throw new ByteCodeException("class file " + thisClass.getClassName() + " already contains a attribute named " + name);
        } else {
            attributes.put(attribute.getName().getValue(), attribute);
        }
    }

    public boolean isInterface() {
        return accessFlags.contains(ClassAccessFlag.INTERFACE);
    }

    public boolean isAbstract() {
        return accessFlags.contains(ClassAccessFlag.ABSTRACT);
    }

    public boolean isPublic() {
        return accessFlags.contains(ClassAccessFlag.PUBLIC);
    }

    public boolean isPackagePrivate() {
        return accessFlags.contains(ClassAccessFlag.PACKAGE_PRIVATE);
    }

    public boolean isFinal() {
        return accessFlags.contains(ClassAccessFlag.FINAL);
    }

    public boolean isSynthetic() {
        return accessFlags.contains(ClassAccessFlag.SYNTHETIC);
    }

    public boolean isEnum() {
        return accessFlags.contains(ClassAccessFlag.ENUM);
    }

    public boolean isAnnotation() {
        return accessFlags.contains(ClassAccessFlag.ANNOTATION);
    }

    public boolean isSuper() {
        return accessFlags.contains(ClassAccessFlag.SUPER);
    }

    private void fillCPool() throws ByteCodeException {
        thisClass.add(cpool);
        if (superClass != null) {
            superClass.add(cpool);
        }
        for (FieldInfo field : fields.values()) {
            field.add(cpool);
        }
        for (ClassInfo iface : interfaces) {
            iface.add(cpool);
        }
        for (List<MethodInfo> methodList : methods.values()) {
            for (MethodInfo method : methodList) {
                method.add(cpool);
            }
        }
        for (AttributeInfo attr : attributes.values()) {
            attr.add(cpool);
        }
    }

    public void write(File file) throws IOException, ByteCodeException {
        try (DataOutputStream out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(file)))) {
            write(out);
        }
    }

    public void write(DataOutputStream out) throws IOException, ByteCodeException {
        fillCPool();
        VersionConstants.JAVA_MAGIC.write(out, cpool);
        minorVersion.write(out, cpool);
        majorVersion.write(out, cpool);
        //write the constant pool
        cpool.write(out, cpool);
        //write the access flags
        AccessFlagsIO.write(out, accessFlags);
        //write this- and superclass
        thisClass.writeIndex(out, cpool);
        if (superClass != null) {
            superClass.writeIndex(out, cpool);
        } else {
            //An empty super class is only allowed for java.lang.Object
            U2.ZERO.write(out, cpool);
        }
        //write interfaces
        new U2(interfaces.size()).write(out, cpool);
        for (ClassInfo iface : interfaces) {
            iface.writeIndex(out, cpool);
        }
        //write fields
        new U2(fields.size()).write(out, cpool);
        for (FieldInfo field : fields.values()) {
            field.write(out, cpool);
        }
        //write methods
        new U2(countMethods()).write(out, cpool);

        for (List<MethodInfo> methodList : methods.values()) {
            for (MethodInfo method : methodList) {
                method.write(out, cpool);
            }
        }
        //write attributes
        new U2(attributes.size()).write(out, cpool);
        for (AttributeInfo attr : attributes.values()) {
            attr.write(out, cpool);
        }
    }

    private int countMethods() {
        int count = 0;
        for (List<MethodInfo> methodList : methods.values()) {
            count += methodList.size();
        }
        return count;
    }

    public static ClassFile read(String file) throws IOException, ByteCodeException {
        try (DataInputStream in = new DataInputStream(new FileInputStream(file))) {
            return read(in);
        }
    }

    public static ClassFile read(String file, ReadMode... modes) throws IOException, ByteCodeException {
        try (DataInputStream in = new DataInputStream(new FileInputStream(file))) {
            return read(in, modes);
        }
    }

    public static ClassFile read(File file) throws IOException, ByteCodeException {
        try (DataInputStream in = new DataInputStream(new FileInputStream(file))) {
            return read(in);
        }
    }

    public static ClassFile read(File file, ReadMode... modes) throws IOException, ByteCodeException {
        try (DataInputStream in = new DataInputStream(new FileInputStream(file))) {
            return read(in, modes);
        }
    }

    public static ClassFile read(byte[] b) throws IOException, ByteCodeException {
        try (DataInputStream in = new DataInputStream(new ByteArrayInputStream(b))) {
            return read(in);
        }
    }

    public static ClassFile read(byte[] b, ReadMode... modes) throws IOException, ByteCodeException {
        try (DataInputStream in = new DataInputStream(new ByteArrayInputStream(b))) {
            return read(in, modes);
        }
    }

    public static ClassFile read(DataInputStream in) throws IOException, ByteCodeException {
        return read(in, ReadMode.All);
    }

    public static ClassFile read(DataInputStream in, ReadMode... modes) throws IOException, ByteCodeException {
        boolean readAll = true;
        if (!Tools.contains(modes, ReadMode.All)) {
            readAll = false;
        }
        long magic = in.readInt() & 0xFFFFFFFFL;
        if (magic == VersionConstants.JAVA_MAGIC.getValue()) {
            int minor = in.readUnsignedShort();
            int major = in.readUnsignedShort();

            ConstantPool cpool = ConstantPool.read(in, major, minor);
            Collection<ClassAccessFlag> flags = AccessFlagsIO.parseClassAccessFlags(in.readUnsignedShort());
            ClassInfo thisClass = ccpt(cpool.get(in.readUnsignedShort()), ClassInfo.class, cpool);
            int nindex = in.readUnsignedShort();
            ClassInfo superClass = null;
            if (nindex != 0) {
                superClass = ccpt(cpool.get(nindex), ClassInfo.class, cpool);
            }
            ClassFile classFile = new ClassFile(thisClass, superClass, flags, cpool, major, minor);

            boolean readIfaces = Tools.contains(modes, ReadMode.Interfaces) | readAll;
            int ifaceCount = in.readUnsignedShort();

            for (int i = 0; i < ifaceCount; i++) {
                int index = in.readUnsignedShort();
                if (readIfaces) {
                    ClassInfo iface = ccpt(cpool.get(index), ClassInfo.class, cpool);
                    classFile.interfaces.add(iface);
                }
            }

            boolean readFields = Tools.contains(modes, ReadMode.Fields) | readAll;
            int fieldCount = in.readUnsignedShort();

            for (int i = 0; i < fieldCount; i++) {
                FieldInfo field = FieldInfo.read(in, cpool, major, minor);
                if (readFields) {
                    classFile.fields.put(field.getName().getValue(), field);
                }
            }

            boolean readMethods = Tools.contains(modes, ReadMode.Methods) | readAll;
            int methodCount = in.readUnsignedShort();

            for (int i = 0; i < methodCount; i++) {
                MethodInfo method = MethodInfo.read(in, cpool, major, minor, thisClass);
                if (readMethods) {
                    classFile.addMethod(method);
                }
            }

            boolean readAttributes = Tools.contains(modes, ReadMode.Attributes) | readAll;
            int attributeCount = in.readUnsignedShort();

            for (int i = 0; i < attributeCount; i++) {
                AttributeInfo attr = AttributesReader.read(in, cpool, major, minor);
                if (readAttributes) {
                    classFile.attributes.put(attr.getName().getValue(), attr);
                }
            }

            if (Tools.contains(modes, ReadMode.ForgetConstantPool)) {
                cpool.purge();
            }
            return classFile;
        } else {
            throw new ByteCodeException("Encountered an invalid magic number while reading the class file: " + Integer.toHexString((int) magic));
        }
    }
}
