package teajay.bytecode;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 * This enum represents all available opcodes for the java 7 VM.
 * It allows, with certain exceptions, to check a provided operand list for compliance with a certain instruction.
 * It does also record (at best effort) how many operands the instruction pops and pushes from the operand stack.
 * A pop or push count of -1 signifies that the effect of the instruction on the operand stack depends on the operands.
 * @author Michael Kußmann <michi@outx.de>
 * @author Michael Burzan
 */
public enum OpCode implements Operand {

    aaload(0x32,2,1),
    aastore(0x53,3,0),
    aconst_null(0x1,0,1),
    aload(0x19,0,1,OpSize.U1),
    aload_0(0x2a,0,1),
    aload_1(0x2b,0,1),
    aload_2(0x2c,0,1),
    aload_3(0x2d,0,1),
    anewarray(0xbd,1,1, OpSize.U2),
    areturn(0xb0,1,0),
    arraylength(0xbe,1,1),
    astore(0x3a,1,0,OpSize.U1),
    astore_0(0x4b,1,0),
    astore_1(0x4c,1,0),
    astore_2(0x4d,1,0),
    astore_3(0x4e,1,0),
    athrow(0xbf,1,1),
    baload(0x33,2,1),
    bastore(0x54,3,0),
    bipush(0x10,0,1, OpSize.U1),
    caload(0x34,2,1),
    castore(0x55,3,1),
    checkcast(0xc0,1,1, OpSize.U2),
    d2f(0x90,2,1),
    d2i(0x8e,2,1),
    d2l(0x8f,2,2),
    dadd(0x63,4,2),
    daload(0x31,2,2),
    dastore(0x52,4,0),
    dcmpg(0x98,4,1),
    dcmpl(0x97,4,1),
    dconst_0(0xe,0,2),
    dconst_1(0xf,0,2),
    ddiv(0x6f,4,2),
    dload(0x18,0,2, OpSize.U1),
    dload_0(0x26,0,2),
    dload_1(0x27,0,2),
    dload_2(0x28,0,2),
    dload_3(0x29,0,2),
    dmul(0x6b,4,2),
    dneg(0x77,2,2),
    drem(0x73,4,2),
    dreturn(0xaf,2,0),
    dstore(0x39,2,0, OpSize.U1),
    dstore_0(0x47,2,0),
    dstore_1(0x48,2,0),
    dstore_2(0x49,2,0),
    dstore_3(0x4a,2,0),
    dsub(0x67,4,2),
    dup(0x59,1,2),
    dup_x1(0x5a,2,3),
    dup_x2(0x5b,3,4),
    dup2(0x5c,2,4),
    dup2_x1(0x5d,3,5),
    dup2_x2(0x5e,4,6),
    f2d(0x8d,1,2),
    f2i(0x8b,1,1),
    f2l(0x8c,1,2),
    fadd(0x62,2,1),
    faload(0x30,2,1),
    fastore(0x51,3,0),
    fcmpg(0x96,2,1),
    fcmpl(0x95,2,1),
    fconst_0(0xb,0,1),
    fconst_1(0xc,0,1),
    fconst_2(0xd,0,1),
    fdiv(0x6e,2,1),
    fload(0x17,0,1, OpSize.U1),
    fload_0(0x22,0,1),
    fload_1(0x23,0,1),
    fload_2(0x24,0,1),
    fload_3(0x25,0,1),
    fmul(0x6a,2,1),
    fneg(0x76,1,1),
    frem(0x72,2,1),
    freturn(0xae,1,0),
    fstore(0x38,1,0, OpSize.U1),
    fstore_0(0x43,1,0),
    fstore_1(0x44,1,0),
    fstore_2(0x45,1,0),
    fstore_3(0x46,1,0),
    fsub(0x66,2,1),
    getfield(0xb4,1,-1, OpSize.U2),
    getstatic(0xb2,0,-1, OpSize.U2),
    vm_goto(0xa7,0,0, OpSize.U2),
    goto_w(0xc8,0,0, OpSize.U4),
    i2b(0x91,1,1),
    i2c(0x92,1,1),
    i2d(0x87,1,2),
    i2f(0x86,1,1),
    i2l(0x85,1,2),
    i2s(0x93,1,1),
    iadd(0x60,2,1),
    iaload(0x2e,2,1),
    iand(0x7e,2,1),
    iastore(0x4f,3,0),
    iconst_m1(0x2,0,1),
    iconst_0(0x3,0,1),
    iconst_1(0x4,0,1),
    iconst_2(0x5,0,1),
    iconst_3(0x6,0,1),
    iconst_4(0x7,0,1),
    iconst_5(0x8,0,1),
    idiv(0x6c,2,1),
    if_acmpeq(0xa5,2,0, OpSize.U2),
    if_acmpne(0xa6,2,0, OpSize.U2),
    if_icmpeq(0x9f,2,0, OpSize.U2),
    if_icmpne(0xa0,2,0, OpSize.U2),
    if_icmplt(0xa1,2,0, OpSize.U2),
    if_icmpge(0xa2,2,0, OpSize.U2),
    if_icmpgt(0xa3,2,0, OpSize.U2),
    if_icmple(0xa4,2,0, OpSize.U2),
    ifeq(0x99,1,0, OpSize.U2),
    ifne(0x9a,1,0, OpSize.U2),
    iflt(0x9b,1,0, OpSize.U2),
    ifge(0x9c,1,0, OpSize.U2),
    ifgt(0x9d,1,0, OpSize.U2),
    ifle(0x9e,1,0, OpSize.U2),
    ifnonnull(0xc7,1,0, OpSize.U2),
    ifnull(0xc6,1,0, OpSize.U2),
    iinc(0x84,0,0, OpSize.U1, OpSize.U1),
    iload(0x15,0,1, OpSize.U1),
    iload_0(0x1a,0,1),
    iload_1(0x1b,0,1),
    iload_2(0x1c,0,1),
    iload_3(0x1d,0,1),
    imul(0x68,2,1),
    ineg(0x74,1,1),
    vm_instanceof(0xc1,1,1, OpSize.U2),
    //We know that push is between 0 and 1
    invokedynamic(0xba,-1,-1, OpSize.U2, OpSize.U2),
    invokeinterface(0xb9,-1,-1, OpSize.U2, OpSize.U1, OpSize.U1),
    invokespecial(0xb7,-1,-1, OpSize.U2),
    invokestatic(0xb8,-1,-1, OpSize.U2),
    invokevirtual(0xb6,-1,-1, OpSize.U2),
    // END 
    ior(0x80,2,1),
    irem(0x70,2,1),
    ireturn(0xac,1,0),
    ishl(0x78,2,1),
    ishr(0x7a,2,1),
    istore(0x36,1,0, OpSize.U1),
    istore_0(0x3b,1,0),
    istore_1(0x3c,1,0),
    istore_2(0x3d,1,0),
    istore_3(0x3e,1,0),
    isub(0x64,2,1),
    iushr(0x7c,2,1),
    ixor(0x82,2,1),
    // Not supported vor java 7 and above
    jsr(0xa8,0,1, OpSize.U2),
    // END
    jsr_w(0xc9,0,1, OpSize.U4),
    l2d(0x8a,2,2),
    l2f(0x89,2,1),
    l2i(0x88,2,1),
    ladd(0x61,4,2),
    laload(0x2f,2,2),
    land(0x7f,4,2),
    lastore(0x50,4,0),
    lcmp(0x94,4,1),
    lconst_0(0x9,0,2),
    lconst_1(0xa,0,2),
    ldc(0x12,0,1, OpSize.U1),
    ldc_w(0x13,0,1, OpSize.U2),
    ldc2_w(0x14,0,2, OpSize.U2),
    ldiv(0x6d,4,2),
    lload(0x16,0,2, OpSize.U1),
    lload_0(0x1e,0,2),
    lload_1(0x1f,0,2),
    lload_2(0x20,0,2),
    lload_3(0x21,0,2),
    lmul(0x69,4,2),
    lneg(0x75,2,2),
    lookupswitch(0xab,1,0),
    lor(0x81,4,2),
    lrem(0x71,4,2),
    lreturn(0xad,2,0),
    lshl(0x79,3,2),
    lshr(0x7b,3,2),
    lstore(0x37,2,0, OpSize.U1),
    lstore_0(0x3f,2,0),
    lstore_1(0x40,2,0),
    lstore_2(0x41,2,0),
    lstore_3(0x42,2,0),
    lsub(0x65,4,2),
    lushr(0x7d,3,2),
    lxor(0x83,4,2),
    monitorenter(0xc2,1,0),
    monitorexit(0xc3,1,0),
    multianewarray(0xc5,-1,1, OpSize.U2, OpSize.U1),
    vm_new(0xbb,0,1, OpSize.U2),
    newarray(0xbc,1,1, OpSize.U1), // Parameter: Table 6.1
    nop(0x0,0,0),
    pop(0x57,1,0),
    pop2(0x58,2,0),
    putfield(0xb5,2,0, OpSize.U2),
    putstatic(0xb3,1,0, OpSize.U2),
    // Not supported vor java 7 and above
    ret(0xa9,0,0, OpSize.U1),
    //END
    vm_return(0xb1,0,0),
    saload(0x35,2,1),
    sastore(0x56,3,0),
    sipush(0x11,0,1, OpSize.U2),
    swap(0x5f,2,2),
    tableswitch(0xaa,1,0),
    wide1(0xc4,-1,-1, OpSize.U1, OpSize.U2),
    wide2(0xc4,-1,-1, OpSize.U1, OpSize.U2, OpSize.U2);

    public final int value;
    public final int num_pop;
    public final int num_push;
    private final OpSize[] valueOP;

    private OpCode(int n, int num_pop, int num_push, OpSize... ops) {
        this.value = n;
        this.valueOP = ops;
        this.num_push = num_push;
        this.num_pop = num_pop;
    }
    
    @Override
    public OpSize getDefaultLengthAsOperand() {
        return OpSize.U1;
    }

    @Override
    public long getAddress(ConstantPool cpool) {
        return value;
    }

    @Override
    public void writeIndex(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        this.write(out, cpool);
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        out.writeByte(value);
    }

    public enum OpSize {

        U1(1),
        U2(2),
        U4(4),
        /*S1(1),
        S2(2),
        S4(4)*/;
        
        public final int size;
        
        private OpSize(int n) {
            this.size = n;
        }
    }
    
    public int getValue() {
        return value;
    }
    
    public void validate(Operand... operands) throws RuntimeException {
        if (this == lookupswitch || this == tableswitch) {
            return; //cannot verify this instructions
        }
        if (valueOP.length != operands.length) {
            throw new RuntimeException("Parameter count mismatch for " + this);
        }
        for (int i = 0; i < operands.length; i++) {
            OpSize size = operands[i].getDefaultLengthAsOperand();
            if (size != valueOP[i]) {
                throw new RuntimeException("Invalid parameter value for " + this);
            }
        }
    }
    
    public static OpCode getFromCode(long value) {
        for (OpCode code : OpCode.values()) {
            if (code.getValue() == value) {
                return code;
            }
        }
        return null;
    }
    
    public static OpCode read(DataInputStream in) throws IOException, ByteCodeException {
        int c = in.readUnsignedByte();
        OpCode opcode = getFromCode(c);
        if (opcode == null) {
            throw new ByteCodeException("unknown opcode encountered: " + Integer.toHexString(c));
        }
        return opcode;
    }
}
