package teajay.bytecode.attributes;

import teajay.bytecode.attributes.signature.Signature;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.lang.reflect.Field;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import teajay.bytecode.attributes.annotations.AnnotationDefault;
import teajay.bytecode.attributes.annotations.RuntimeInvisibleAnnotations;
import teajay.bytecode.attributes.annotations.RuntimeInvisibleParameterAnnotations;
import teajay.bytecode.attributes.annotations.RuntimeVisibleAnnotations;
import teajay.bytecode.attributes.annotations.RuntimeVisibleParameterAnnotations;
import teajay.bytecode.attributes.bootstrap.BootstrapMethods;
import teajay.bytecode.attributes.innerclass.EnclosingMethod;
import teajay.bytecode.attributes.innerclass.InnerClasses;
import teajay.bytecode.attributes.nest.NestHost;
import teajay.bytecode.attributes.nest.NestMembers;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.MalformedAttributeException;
import static teajay.bytecode.util.Tools.ccpt;

/**
 * This is a tool for reading JVM attributes of any kind. This reader does not
 * yet have special handling for the Code-attribute. <tt>ClassFile</tt>s the
 * were read will lack the Code-attribute.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class AttributesReader {

    private static final Map<String, MethodHandle> ADDITIONAL_ATTRIBUTES = new HashMap<>();
    private static final Collection<String> DEFAULTS = new HashSet<>(Arrays.asList(
            ConstantValue.NAME, Synthetic.NAME, Deprecated.NAME, Signature.NAME, SourceFile.NAME,
            Exceptions.NAME, InnerClasses.NAME, EnclosingMethod.NAME, BootstrapMethods.NAME,
            SourceDebugExtension.NAME, AnnotationDefault.NAME, RuntimeInvisibleAnnotations.NAME,
            RuntimeVisibleAnnotations.NAME, RuntimeInvisibleParameterAnnotations.NAME,
            RuntimeVisibleParameterAnnotations.NAME, NestHost.NAME, NestMembers.NAME));

    public synchronized static void registerAttribute(Class<? extends AttributeInfo> cl) throws MalformedAttributeException {
        MethodType read1 = null;
        MethodType read2 = null;
        try {
            Field f = cl.getField("NAME");
            Object obj = f.get(null);
            if (obj == null) {
                throw new MalformedAttributeException("NAME attribute in " + cl.getName() + " may not be null.");
            }
            if (!(obj instanceof String)) {
                throw new MalformedAttributeException("NAME attribute in " + cl.getName() + " is not String.");
            }
            if (!ADDITIONAL_ATTRIBUTES.containsKey( (String) obj) && !DEFAULTS.contains( (String) obj)) {
                read1 = MethodType.methodType(cl, byte[].class, ConstantPool.class, int.class, int.class, Utf8Info.class);
                read2 = MethodType.methodType(cl, DataInputStream.class, ConstantPool.class, int.class, int.class, int.class, Utf8Info.class);

                MethodHandle handle;
                try {
                    handle = MethodHandles.lookup().findStatic(cl, "read", read1);
                } catch (NoSuchMethodException ex) {
                    handle = MethodHandles.lookup().findStatic(cl, "read", read2);
                }
                ADDITIONAL_ATTRIBUTES.put((String) obj, handle);
            } else {
                throw new MalformedAttributeException("An attribute with the name " + obj + " is already registered");
            }
        } catch (IllegalAccessException | SecurityException ex) {
            throw new RuntimeException("Cannot access fields or methods on the supplied attribute: " + ex.getMessage());
        } catch (NoSuchFieldException ex) {
            throw new MalformedAttributeException("Missing static String NAME attribute in " + cl.getName() + ".");
        } catch (NoSuchMethodException ex) {
            throw new MalformedAttributeException("No such methods read " + read1 + "\n\t" + " and read " + read2);
        }
    }

    public static AttributeInfo read(DataInputStream in, ConstantPool cpool, int major, int minor) throws ByteCodeException, IOException {
        Utf8Info name = ccpt(cpool.get(in.readUnsignedShort()), Utf8Info.class, cpool);
        long llen = in.readInt() & 0xFFFFFFFFL;
        if (llen > Integer.MAX_VALUE) {
            throw new ByteCodeException("Cannot handle attributes whose length exceeds signed 32-bit integer.");
        }
        /**
         * Signed integer range must be enough here, as otherwise the attribute
         * wont fit into a <tt>byte[]</tt> anymore.
         */
        int length = (int) llen;
        AttributeInfo info;
        MethodHandle read = ADDITIONAL_ATTRIBUTES.get(name.getValue());
        if (read != null) {
            try {
                if (read.type().parameterType(1).equals(byte[].class)) {
                    info = (AttributeInfo) read.invoke(read(in, length), cpool, major, minor, name);
                } else {
                    //Don't trust foreign attributes
                    DataInputStream stream = new DataInputStream(new ByteArrayInputStream(read(in, length)));
                    info = (AttributeInfo) read.invoke(stream, cpool, major, minor, length, name);
                }
            } catch (Throwable ex) {
                if (ex instanceof ByteCodeException) {
                    throw (ByteCodeException) ex;
                } else if (ex instanceof IOException) {
                    throw (IOException) ex;
                } else {
                    throw new ByteCodeException("custom attribute has thrown an exception", ex);
                }
            }
        } else {
            switch (name.getValue()) {
                case ConstantValue.NAME:
                    info = ConstantValue.read(in, cpool, major, minor, length, name);
                    break;
                case Synthetic.NAME:
                    info = Synthetic.read(in, cpool, major, minor, length, name);
                    break;
                case Deprecated.NAME:
                    info = Deprecated.read(in, cpool, major, minor, length, name);
                    break;
                case Signature.NAME:
                    info = Signature.read(in, cpool, major, minor, length, name);
                    break;
                case SourceFile.NAME:
                    info = SourceFile.read(in, cpool, major, minor, length, name);
                    break;
                case Exceptions.NAME:
                    info = Exceptions.read(in, cpool, major, minor, length, name);
                    break;
                case InnerClasses.NAME:
                    info = InnerClasses.read(in, cpool, major, minor, length, name);
                    break;
                case EnclosingMethod.NAME:
                    info = EnclosingMethod.read(in, cpool, major, minor, length, name);
                    break;
                case BootstrapMethods.NAME:
                    info = BootstrapMethods.read(in, cpool, major, minor, length, name);
                    break;
                case SourceDebugExtension.NAME:
                    info = SourceDebugExtension.read(read(in, length), cpool, major, minor, name);
                    break;
                case AnnotationDefault.NAME:
                    info = AnnotationDefault.read(in, cpool, major, minor, length, name);
                    break;
                case RuntimeInvisibleAnnotations.NAME:
                    info = RuntimeInvisibleAnnotations.read(in, cpool, major, minor, length, name);
                    break;
                case RuntimeVisibleAnnotations.NAME:
                    info = RuntimeVisibleAnnotations.read(in, cpool, major, minor, length, name);
                    break;
                case RuntimeInvisibleParameterAnnotations.NAME:
                    info = RuntimeInvisibleParameterAnnotations.read(in, cpool, major, minor, length, name);
                    break;
                case RuntimeVisibleParameterAnnotations.NAME:
                    info = RuntimeVisibleParameterAnnotations.read(in, cpool, major, minor, length, name);
                    break;
                case NestHost.NAME:
                    info = NestHost.read(in, cpool, major, minor, length, name);
                    break;
                case NestMembers.NAME:
                    info = NestMembers.read(in, cpool, major, minor, length, name);
                    break;
                default:
                    byte[] raw = read(in, length);
                    info = new UnknownAttribute(name, raw);
            }
        }
        return info;
    }

    private static byte[] read(DataInputStream in, int len) throws IOException, ByteCodeException {
        byte[] dest = new byte[len];
        int c;
        int offset = 0;
        while ((c = in.read(dest, offset, len - offset)) != -1 && offset < len) {
            offset += c;
        }
        if (offset != len) {
            throw new ByteCodeException("constant pool: Unexpected end of file");
        }
        return dest;
    }
}
