package teajay.bytecode.attributes;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import teajay.bytecode.U4;
import teajay.bytecode.cpool.*;
import teajay.bytecode.exceptions.ByteCodeException;
import static teajay.bytecode.util.Tools.ccpt;

/**
 * Implementation of the ConstantValue attribute as stated in "The JavaTM Virtual
 * Machine Specification Version 7". Revision: 27.07.2012
 * This Attribute should only appear in the attribute table of a field.
 * @author Michael Kußmann <michi@outx.de>
 */
public final class ConstantValue extends AttributeInfo {
    
    public static enum ValueType {
        Long,
        Float,
        Double,
        Integer,
        String;
    }

    public static final String NAME = "ConstantValue";
    private final CPoolEntry value;
    private final ValueType type;

    private ConstantValue(Utf8Info name, CPoolEntry value) throws ByteCodeException {
        super(name);
        checkName(NAME, name);
        switch (value.getTag()) {
            case Long:
                type = ValueType.Long;
                break;
            case Float:
                type = ValueType.Float;
                break;
            case Double:
                type = ValueType.Double;
                break;
            case Integer:
                type = ValueType.Integer;
                break;
            case String:
                type = ValueType.String;
                break;
            default:
                throw new ByteCodeException("constant pool: Invalid attribute type for ConstantValue " + value.getTag());
        }
        this.value = value;
    }

    private ConstantValue(CPoolEntry value) throws ByteCodeException {
        this(new Utf8Info(NAME), value);
    }
    
    public ConstantValue(StringInfo value) throws ByteCodeException {
        this( (CPoolEntry) value);
    }
       
    public ConstantValue(String value) throws ByteCodeException {
        this(new StringInfo(value));
    }
    
    public ConstantValue(IntegerInfo value) throws ByteCodeException {
        this( (CPoolEntry) value);
    }

    public ConstantValue(int value) throws ByteCodeException {
        this(new IntegerInfo(value));
    }
    
    public ConstantValue(LongInfo value) throws ByteCodeException {
        this( (CPoolEntry) value);
    }

    public ConstantValue(long value) throws ByteCodeException {
        this(new LongInfo(value));
    }

    public ConstantValue(FloatInfo value) throws ByteCodeException {
        this( (CPoolEntry) value);
    }
    
    public ConstantValue(float value) throws ByteCodeException {
        this(new FloatInfo(value));
    }
    
    public ConstantValue(DoubleInfo value) throws ByteCodeException {
        this( (CPoolEntry) value);
    }

    public ConstantValue(double value) throws ByteCodeException {
        this(new DoubleInfo(value));
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        return new U4(2);
    }

    public CPoolEntry getValue() {
        return value;
    }
       
    public ValueType getType() {
        return type;
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        super.add(cpool);
        value.add(cpool);
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        super.write(out, cpool);
        value.writeIndex(out, cpool);
    }

    public static ConstantValue read(DataInputStream in, ConstantPool cpool, int major, int minor, int length, Utf8Info name) throws ByteCodeException, IOException {
        CPoolEntry entry = ccpt(cpool.get(in.readUnsignedShort()), CPoolEntry.class, cpool);
        return new ConstantValue(name, entry);

    }
}
