package teajay.bytecode.attributes;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.*;
import teajay.bytecode.U2;
import teajay.bytecode.U4;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;
import static teajay.bytecode.util.Tools.ccpt;

/**
 * Implementation of the Exceptions attribute as stated in "The JavaTM Virtual
 * Machine Specification Version 7". Revision: 27.07.2012
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class Exceptions extends AttributeInfo {
    public static final String NAME = "Exceptions";

    private final Collection<ClassInfo> exceptions;
    
    public Exceptions() {
        this(Collections.<ClassInfo>emptyList(), new Utf8Info(NAME));
    }
    
    public Exceptions(Collection<ClassInfo> exceptions) {
        this(exceptions, new Utf8Info(NAME));
    }
    
    private Exceptions(Collection<ClassInfo> exceptions, Utf8Info name) {
        super(name);
        checkName(NAME, name);
        this.exceptions = new HashSet<>();
        for (ClassInfo exception : exceptions) {
            this.exceptions.add(exception);
        }
    }
    
    public void addException(ClassInfo exception) {
        exceptions.add(exception);
    }
    
    public boolean removeException(ClassInfo exception) {
        return exceptions.remove(exception);
    }    
    
    public Collection<ClassInfo> getExceptions() {
        return Collections.unmodifiableCollection(exceptions);
    }
    
    public boolean containsException(ClassInfo exception) {
        return exceptions.contains(exception);
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        return new U4(2 + exceptions.size() * 2);
    }
    
    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        super.write(out, cpool);
        new U2(exceptions.size()).write(out, cpool);
        for (ClassInfo exception : exceptions) {
            exception.writeIndex(out, cpool);
        }
    }
    
    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        super.add(cpool);
        for (ClassInfo cl : exceptions) {
            cl.add(cpool);
        }
    }
    
    public static Exceptions read(DataInputStream in, ConstantPool cpool, int major, int minor, int length, Utf8Info name) throws ByteCodeException, IOException {
         int numExceptions = in.readUnsignedShort();
         Exceptions ex = new Exceptions(Collections.<ClassInfo>emptyList(), name);
         for (int i = 0; i < numExceptions; i++) {
             ClassInfo cl = ccpt(cpool.get(in.readUnsignedShort()), ClassInfo.class, cpool);
             ex.addException(cl);
         }
         return ex;
     }
}
