package teajay.bytecode.attributes;

import java.io.*;
import teajay.bytecode.U4;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 * Implementation of the SourceDebugExtension attribute as stated in "The JavaTM
 * Virtual Machine Specification Version 7". Revision: 27.07.2012
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class SourceDebugExtension extends AttributeInfo {

    public final static String NAME = "SourceDebugExtension";
    private final String information;

    public SourceDebugExtension(String information) {
        this(information, new Utf8Info(NAME));
    }

    private SourceDebugExtension(String information, Utf8Info name) {
        super(name);
        checkName(NAME, name);
        this.information = information;
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        return new U4(utflen(information));
    }

    public String getInformation() {
        return information;
    }

    /**
     * @see java.io.DataInputStream
     */
    private static int utflen(String str) {
        int strlen = str.length();
        int utflen = 0;
        for (int i = 0; i < strlen; i++) {
            int c = str.charAt(i);
            if ((c >= 0x0001) && (c <= 0x007F)) {
                utflen++;
            } else if (c > 0x07FF) {
                utflen += 3;
            } else {
                utflen += 2;
            }
        }
        return utflen + 2;
    }

    /**
     * @see java.io.DataInputStream
     */
    private static String readUTF(DataInput in, int utflen) throws IOException {
        byte[] bytearr = new byte[utflen];
        char[] chararr = new char[utflen];

        int c, char2, char3;
        int count = 0;
        int chararr_count = 0;

        in.readFully(bytearr, 0, utflen);

        while (count < utflen) {
            c = (int) bytearr[count] & 0xff;
            if (c > 127) {
                break;
            }
            count++;
            chararr[chararr_count++] = (char) c;
        }

        while (count < utflen) {
            c = (int) bytearr[count] & 0xff;
            switch (c >> 4) {
                case 0:
                case 1:
                case 2:
                case 3:
                case 4:
                case 5:
                case 6:
                case 7:
                    /*
                     * 0xxxxxxx
                     */
                    count++;
                    chararr[chararr_count++] = (char) c;
                    break;
                case 12:
                case 13:
                    /*
                     * 110x xxxx 10xx xxxx
                     */
                    count += 2;
                    if (count > utflen) {
                        throw new UTFDataFormatException(
                                "malformed input: partial character at end");
                    }
                    char2 = (int) bytearr[count - 1];
                    if ((char2 & 0xC0) != 0x80) {
                        throw new UTFDataFormatException(
                                "malformed input around byte " + count);
                    }
                    chararr[chararr_count++] = (char) (((c & 0x1F) << 6)
                            | (char2 & 0x3F));
                    break;
                case 14:
                    /*
                     * 1110 xxxx 10xx xxxx 10xx xxxx
                     */
                    count += 3;
                    if (count > utflen) {
                        throw new UTFDataFormatException(
                                "malformed input: partial character at end");
                    }
                    char2 = (int) bytearr[count - 2];
                    char3 = (int) bytearr[count - 1];
                    if (((char2 & 0xC0) != 0x80) || ((char3 & 0xC0) != 0x80)) {
                        throw new UTFDataFormatException(
                                "malformed input around byte " + (count - 1));
                    }
                    chararr[chararr_count++] = (char) (((c & 0x0F) << 12)
                            | ((char2 & 0x3F) << 6)
                            | (char3 & 0x3F));
                    break;
                default:
                    /*
                     * 10xx xxxx, 1111 xxxx
                     */
                    throw new UTFDataFormatException(
                            "malformed input around byte " + count);
            }
        }
        // The number of chars produced may be less than utflen
        return new String(chararr, 0, chararr_count);
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws ByteCodeException, IOException {
        super.write(out, cpool);
        out.writeUTF(information);
    }

    public static SourceDebugExtension read(byte[] data, ConstantPool cpool, int major, int minor, Utf8Info name) throws ByteCodeException, IOException {
        try (DataInputStream in = new DataInputStream(new ByteArrayInputStream(data))) {
            /*
             * Not sure how to handle the data here as the jvm spec states that
             * the string here may be longer than what java.lang.String can
             * hold. Thus it cannot be in the format that @see
             * #java.io.DataInputStream.readUTF(DataOutput) accepts, but javap
             * fails to read the string if it is not encoded in excatly that
             * way.
             */
            try {
                if (data.length >= 2) {
                    int len = in.readUnsignedShort();
                    in.reset();
                    if (data.length == len + 2) {
                        String information = in.readUTF();
                        return new SourceDebugExtension(information, name);
                    }
                }

                String information = readUTF(in, data.length);
                return new SourceDebugExtension(information, name);
            } catch (UTFDataFormatException ex) {
                //Cannot handle the format.
                return new SourceDebugExtension("", name);
            }
        }
    }
}
