package teajay.bytecode.attributes.annotations;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.IndexWritable;
import teajay.bytecode.U2;
import static teajay.bytecode.attributes.annotations.AnnotationElementType.*;
import static teajay.bytecode.util.Tools.ccpt;
import teajay.bytecode.cpool.*;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.MalformedClassNameException;
import teajay.bytecode.exceptions.UTFOutOfBoundsException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class ElementValue implements ByteCodeWritable, ConstantPoolAddable {

    public static class EnumDescr implements ConstantPoolAddable, IndexWritable {

        public final FieldDescriptor typeName;
        public final Utf8Info constName;

        private EnumDescr(FieldDescriptor typeName, String constName) throws UTFOutOfBoundsException {
            this.typeName = typeName;
            this.constName = new Utf8Info(constName);
        }

        private EnumDescr(FieldDescriptor typeName, Utf8Info constName) {
            this.typeName = typeName;
            this.constName = constName;
        }

        @Override
        public void add(ConstantPool cpool) throws ByteCodeException {
            typeName.add(cpool);
            constName.add(cpool);
        }

        public FieldDescriptor getType() {
            return typeName;
        }

        public Utf8Info getConst() {
            return constName;
        }

        public String getConstString() {
            return constName.getValue();
        }

        @Override
        public void writeIndex(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
            write(out, cpool);
        }

        @Override
        public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
            typeName.writeIndex(out, cpool);
            constName.writeIndex(out, cpool);
        }

        @Override
        public String toString() {
            return "type=" + typeName + ", name=" + constName.getValue();
        }
    }

    public static class ElementValueList implements List<ElementValue>, ConstantPoolAddable, IndexWritable {

        private static final long serialVersionUID = 1L;
        private final List<ElementValue> list;

        private ElementValueList(ElementValue[] arr) {
            this.list = Arrays.asList(arr);
        }

        @Override
        public void add(ConstantPool cpool) throws ByteCodeException {
            for (ElementValue val : this) {
                val.add(cpool);
            }
        }

        @Override
        public void writeIndex(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
            write(out, cpool);
        }

        @Override
        public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
            new U2(this.size()).write(out, cpool);
            for (ElementValue val : this) {
                val.write(out, cpool);
            }
        }

        @Override
        public String toString() {
            StringBuilder buf = new StringBuilder();
            for (ElementValue v : this) {
                buf.append(v);
                buf.append(", ");
            }
            buf.deleteCharAt(buf.length() - 1);
            buf.deleteCharAt(buf.length() - 1);
            return buf.toString();
        }

        @Override
        public int size() {
            return list.size();
        }

        @Override
        public boolean isEmpty() {
            return list.isEmpty();
        }

        @Override
        public boolean contains(Object o) {
            return list.contains(o);
        }

        @Override
        public Iterator<ElementValue> iterator() {
            return list.iterator();
        }

        @Override
        public Object[] toArray() {
            return list.toArray();
        }

        @Override
        public <T> T[] toArray(T[] a) {
            return list.toArray(a);
        }

        @Override
        public boolean add(ElementValue e) {
            throw new UnsupportedOperationException("this list is unmodifiable");
        }

        @Override
        public boolean remove(Object o) {
            throw new UnsupportedOperationException("this list is unmodifiable");
        }

        @Override
        public boolean containsAll(Collection<?> c) {
            return list.containsAll(c);
        }

        @Override
        public boolean addAll(Collection<? extends ElementValue> c) {
            throw new UnsupportedOperationException("this list is unmodifiable");
        }

        @Override
        public boolean addAll(int index, Collection<? extends ElementValue> c) {
            throw new UnsupportedOperationException("this list is unmodifiable");
        }

        @Override
        public boolean removeAll(Collection<?> c) {
            throw new UnsupportedOperationException("this list is unmodifiable");
        }

        @Override
        public boolean retainAll(Collection<?> c) {
            throw new UnsupportedOperationException("this list is unmodifiable");
        }

        @Override
        public void clear() {
            throw new UnsupportedOperationException("this list is unmodifiable");
        }

        @Override
        public ElementValue get(int index) {
            return list.get(index);
        }

        @Override
        public ElementValue set(int index, ElementValue element) {
            throw new UnsupportedOperationException("this list is unmodifiable");
        }

        @Override
        public void add(int index, ElementValue element) {
            throw new UnsupportedOperationException("this list is unmodifiable");
        }

        @Override
        public ElementValue remove(int index) {
            throw new UnsupportedOperationException("this list is unmodifiable");
        }

        @Override
        public int indexOf(Object o) {
            return list.indexOf(o);
        }

        @Override
        public int lastIndexOf(Object o) {
            return list.lastIndexOf(o);
        }

        @Override
        public ListIterator<ElementValue> listIterator() {
            return list.listIterator();
        }

        @Override
        public ListIterator<ElementValue> listIterator(int index) {
            return list.listIterator(index);
        }

        @Override
        public List<ElementValue> subList(int fromIndex, int toIndex) {
            return list.subList(fromIndex, toIndex);
        }
    }
    private final ConstantPoolAddable value;
    private final AnnotationElementType tag;

    /**
     * Creates a new element value entity of type <tt>boolean</tt>. Valid types
     * for retrieval via the {@link #getValue(java.lang.Class)} method are:
     * {@link java.lang.Boolean} and {@link teajay.bytecode.cpool.IntegerInfo}
     *
     * @param value the value
     */
    public ElementValue(boolean value) {
        this.value = new IntegerInfo(value);
        this.tag = Boolean;
    }

    /**
     * Creates a new element value entity of type <tt>byte</tt>. Valid types for
     * retrieval via the {@link #getValue(java.lang.Class)} method are:
     * {@link java.lang.Byte} and {@link teajay.bytecode.cpool.IntegerInfo}
     *
     * @param value the value
     */
    public ElementValue(byte value) {
        this.value = new IntegerInfo(value);
        this.tag = Byte;
    }

    /**
     * Creates a new element value entity of type <tt>short</tt>. Valid types
     * for retrieval via the {@link #getValue(java.lang.Class)} method are:
     * {@link java.lang.Short} and {@link teajay.bytecode.cpool.IntegerInfo}
     *
     * @param value the value
     */
    public ElementValue(short value) {
        this.value = new IntegerInfo(value);
        this.tag = Short;
    }

    /**
     * Creates a new element value entity of type <tt>char</tt>. Valid types for
     * retrieval via the {@link #getValue(java.lang.Class)} method are:
     * {@link java.lang.Character} and {@link teajay.bytecode.cpool.IntegerInfo}
     *
     * @param value the value
     */
    public ElementValue(char value) {
        this.value = new IntegerInfo(value);
        this.tag = Char;
    }

    /**
     * Creates a new element value entity of type <tt>int</tt>. Valid types for
     * retrieval via the {@link #getValue(java.lang.Class)} method are:
     * {@link java.lang.Integer} and {@link teajay.bytecode.cpool.IntegerInfo}
     *
     * @param value the value
     */
    public ElementValue(int value) {
        this.value = new IntegerInfo(value);
        this.tag = Int;
    }

    /**
     * Creates a new element value entity of type <tt>long</tt>. Valid types for
     * retrieval via the {@link #getValue(java.lang.Class)} method are:
     * {@link java.lang.Long} and {@link teajay.bytecode.cpool.LongInfo}
     *
     * @param value the value
     */
    public ElementValue(long value) {
        this.value = new LongInfo(value);
        this.tag = Long;
    }

    /**
     * Creates a new element value entity of type <tt>float</tt>. Valid types
     * for retrieval via the {@link #getValue(java.lang.Class)} method are:
     * {@link java.lang.Float} and {@link teajay.bytecode.cpool.FloatInfo}
     *
     * @param value the value
     */
    public ElementValue(float value) {
        this.value = new FloatInfo(value);
        this.tag = Float;
    }

    /**
     * Creates a new element value entity of type <tt>double</tt>. Valid types
     * for retrieval via the {@link #getValue(java.lang.Class)} method are:
     * {@link java.lang.Double} and {@link teajay.bytecode.cpool.DoubleInfo}
     *
     * @param value the value
     */
    public ElementValue(double value) {
        this.value = new DoubleInfo(value);
        this.tag = Double;
    }

    /**
     * Creates a new element value entity of type string. Valid types for
     * retrieval via the {@link #getValue(java.lang.Class)} method are:
     * {@link java.lang.String} and {@link teajay.bytecode.cpool.Utf8Info}
     *
     * @param value the value
     */
    public ElementValue(String value) throws UTFOutOfBoundsException {
        this.value = new Utf8Info(value);
        this.tag = String;
    }

    /**
     * Creates a new element value entity of type class. Valid types for
     * retrieval via the {@link #getValue(java.lang.Class)} method is:
     * {@link teajay.bytecode.cpool.FieldDescriptor}
     *
     * @param value the value
     */
    public ElementValue(FieldDescriptor value) {
        this.value = value;
        this.tag = Class;
    }

    /**
     * Creates a new element value entity of type annotation. Valid types for
     * retrieval via the {@link #getValue(java.lang.Class)} method is:
     * {@link teajay.bytecode.attributes.annotations.Annotation}
     *
     * @param value the value
     */
    public ElementValue(Annotation value) {
        this.value = value;
        this.tag = Annotation;
    }

    /**
     * Creates a new element value entity of type array. Valid types for
     * retrieval via the {@link #getValue(java.lang.Class)} method are: null
     * null null     {@link teajay.bytecode.attributes.annotations.ElementValue.ElementValueList}
     * {@link java.util.List}
     * {@link java.util.Collection}
     *
     * @param value the value
     */
    public ElementValue(ElementValue... values) {
        this.value = new ElementValueList(values);
        this.tag = Array;
    }

    /**
     * Creates a new element value entity of type enum. Valid types for
     * retrieval via the {@link #getValue(java.lang.Class)} method is:
     * {@link teajay.bytecode.attributes.annotations.ElementValue.EnumDescr}
     *
     * @param value the value
     */
    public ElementValue(Enum<?> value) throws MalformedClassNameException, ByteCodeException {
        this.value = new EnumDescr(FieldDescriptor.fromClass(value.getClass()), value.name());
        this.tag = Enum;
    }

    public ElementValue(AnnotationElementType tag, ConstantPoolAddable value) {
        if (!tag.isCorrectDataType(value.getClass())) {
            throw new IllegalArgumentException("invalid data type " + value.getClass().getName() + " for annotation element type " + tag);
        }
        this.tag = tag;
        this.value = value;
    }

    public <T> T getValue(Class<T> type) throws IllegalArgumentException {
        if (tag.isCorrectDataType(type)
                || (tag == Enum && type.isEnum())) {
            if (value.getClass().equals(type)) {
                return type.cast(value);
            } else {
                switch (tag) {
                    case Boolean:
                        IntegerInfo info = (IntegerInfo) value;
                        return type.cast(java.lang.Boolean.valueOf(info.getIntegerValue() == 0 ? false : true));
                    case Byte:
                        info = (IntegerInfo) value;
                        return type.cast(java.lang.Byte.valueOf((byte) info.getIntegerValue()));
                    case Short:
                        info = (IntegerInfo) value;
                        return type.cast(java.lang.Short.valueOf((short) info.getIntegerValue()));
                    case Char:
                        info = (IntegerInfo) value;
                        return type.cast(java.lang.Character.valueOf((char) info.getIntegerValue()));
                    case Int:
                        info = (IntegerInfo) value;
                        return type.cast(java.lang.Integer.valueOf(info.getIntegerValue()));
                    case Double:
                        DoubleInfo dinfo = (DoubleInfo) value;
                        return type.cast(java.lang.Double.valueOf(dinfo.getDoubleValue()));
                    case Float:
                        FloatInfo finfo = (FloatInfo) value;
                        return type.cast(java.lang.Double.valueOf(finfo.getFloatValue()));
                    case String:
                        Utf8Info sinfo = (Utf8Info) value;
                        return type.cast(sinfo.getValue());
                    case Array:
                        return type.cast(value);
                    case Enum:
                        EnumDescr descr = (EnumDescr) value;
                        try {
                            if (!descr.getType().getClassInfo().getAsClass().equals(type)) {
                                throw new IllegalArgumentException("incompatible enum types "
                                        + descr.getType().getClassInfo().getAsClass().getName() + " <-> " + type.getName());
                            } else {
                                String name = descr.getConstString();
                                for (T a : type.getEnumConstants()) {
                                    Enum<?> e = (Enum<?>) a;
                                    if (e.name().equals(name)) {
                                        return a;
                                    }
                                }
                                throw new IllegalArgumentException("no matching enum constant");
                            }
                        } catch (ClassNotFoundException ex) {
                            throw new IllegalArgumentException("incompatible enum types "
                                        + descr.getType().getClassInfo().getClassName().replace('/','.') + " <-> " + type.getName());
                        }
                    default:
                        throw new IllegalArgumentException("no matching type");
                }
            }
        } else {
            throw new IllegalArgumentException("invalid value type " + type.getName() + " for annotation element type " + tag);
        }
    }

    public AnnotationElementType getTag() {
        return tag;
    }

    @Override
    public String toString() {
        return "ElementValue (" + tag + "): \n" + value;
    }

    int getLength() {
        switch (tag) {
            case Enum:
                return 1 + 2 + 2;
            case Annotation:
                return ((Annotation) value).getLength() + 1;
            case Array:
                int len = 1 + 2;
                ElementValueList vals = (ElementValueList) value;
                for (ElementValue val : vals) {
                    len += val.getLength();
                }
                return len;
            default:
                return 1 + 2;
        }
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        tag.write(out, cpool);
        ((IndexWritable) value).writeIndex(out, cpool);
    }

    @Override
    public void add(ConstantPool pool) throws ByteCodeException {
        value.add(pool);
    }

    public static ElementValue read(DataInputStream in, ConstantPool cpool, int major, int minor) throws ByteCodeException, IOException {
        char tc = (char) in.readUnsignedByte();
        AnnotationElementType tag = AnnotationElementType.parse(tc);

        switch (tag) {
            case Byte:
            case Char:
            case Int:
            case Boolean:
            case Short:
                return new ElementValue(tag, ccpt(cpool.get(in.readUnsignedShort()), IntegerInfo.class, cpool));
            case Double:
                return new ElementValue(tag, ccpt(cpool.get(in.readUnsignedShort()), DoubleInfo.class, cpool));
            case Float:
                return new ElementValue(tag, ccpt(cpool.get(in.readUnsignedShort()), FloatInfo.class, cpool));
            case Long:
                return new ElementValue(tag, ccpt(cpool.get(in.readUnsignedShort()), LongInfo.class, cpool));
            case String:
                return new ElementValue(tag, ccpt(cpool.get(in.readUnsignedShort()), Utf8Info.class, cpool));
            case Class:
                return new ElementValue(tag, ccpt(cpool.get(in.readUnsignedShort()), FieldDescriptor.class, cpool));
            case Enum:
                FieldDescriptor type = ccpt(cpool.get(in.readUnsignedShort()), FieldDescriptor.class, cpool);
                Utf8Info name = ccpt(cpool.get(in.readUnsignedShort()), Utf8Info.class, cpool);

                return new ElementValue(tag, new EnumDescr(type, name));
            case Annotation:
                return new ElementValue(tag, teajay.bytecode.attributes.annotations.Annotation.read(in, cpool, major, minor));
            case Array:
                int numValues = in.readUnsignedShort();
                ElementValue[] values = new ElementValue[numValues];
                for (int i = 0; i < numValues; i++) {
                    values[i] = ElementValue.read(in, cpool, major, minor);
                }
                return new ElementValue(values);
            default:
                throw new ByteCodeException("unknown element value tag: " + tc);
        }
    }
}
