package teajay.bytecode.attributes.annotations;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import teajay.bytecode.U2;
import teajay.bytecode.U4;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
abstract class RuntimeAnnotation extends AttributeInfo implements Iterable<Annotation> {

    private final List<Annotation> annotations;

    protected RuntimeAnnotation(Utf8Info name, Annotation... annotations) {
        super(name);
        this.annotations = new ArrayList<>();
        this.annotations.addAll(Arrays.asList(annotations));
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        int len = 2;
        for (Annotation an : annotations) {
            len += an.getLength();
        }
        return new U4(len);
    }
    
    public void addAnnotation(Annotation annotation) {
        annotations.add(annotation);
    }
    
    public void removeAnnotation(Annotation annotation) {
        annotations.remove(annotation);
    }
    
    @Override
    public String toString() {
        StringBuilder buf = new StringBuilder();
        buf.append(super.getName().getValue());
        buf.append(":\n");
        for (Annotation a : annotations) {
            buf.append(a);
            buf.append('\n');
        }
        return buf.toString();
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        super.write(out, cpool);
        new U2(annotations.size()).write(out, cpool);
        for (Annotation an : annotations) {
            an.write(out, cpool);
        }
    }
    
    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        super.add(cpool);
        for (Annotation an : annotations) {
            an.add(cpool);
        }
    }
    
    protected static Annotation[] read(DataInputStream in, ConstantPool cpool, int major, int minor) throws ByteCodeException, IOException {
        int numAnnotations = in.readUnsignedShort();
        Annotation[] annotations = new Annotation[numAnnotations];
        for (int i = 0; i < numAnnotations; i++) {
            annotations[i] = Annotation.read(in, cpool, major, minor);
        }
        return annotations;
    }

    @Override
    public Iterator<Annotation> iterator() {
        return annotations.iterator();
    }
}
