package teajay.bytecode.attributes.annotations;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import teajay.bytecode.U1;
import teajay.bytecode.U2;
import teajay.bytecode.U4;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
abstract class RuntimeParameterAnnotation extends AttributeInfo implements Iterable<List<Annotation>> {

    private final List<List<Annotation>> annotations;

    protected RuntimeParameterAnnotation(Utf8Info name) {
        super(name);
        this.annotations = new ArrayList<>();
    }
    
    protected RuntimeParameterAnnotation(Utf8Info name, List<List<Annotation>> annotations) {
        super(name);
        this.annotations = annotations;
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        int len = 1;
        for (List<Annotation> lst : annotations) {
            for (Annotation an : lst) {
                len += an.getLength();
            }
            len += 2;
        }
        return new U4(len);
    }

    public void addParameterAnnotations(List<Annotation> annotations) {
        this.annotations.add(annotations);
    }
    
    public void addParameterAnnotations(Annotation... annotations) {
        this.annotations.add(Arrays.asList(annotations));
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        super.write(out, cpool);
        new U1(annotations.size()).write(out, cpool);
        for (List<Annotation> lst : annotations) {
            new U2(lst.size()).write(out, cpool);
            for (Annotation an : lst) {
                an.write(out, cpool);
            }
        }
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        super.add(cpool);
        for (List<Annotation> lst : annotations) {
            for (Annotation an : lst) {
                an.add(cpool);
            }
        }
    }
    
    protected static List<List<Annotation>> read(DataInputStream in, ConstantPool cpool, int major, int minor) throws ByteCodeException, IOException {
        int numParams = in.readUnsignedByte();
        List<List<Annotation>> result = new ArrayList<>(numParams);
        for (int i = 0; i < numParams; i++) {
            int numAnnotations = in.readUnsignedShort();
            List<Annotation> tmp = new ArrayList<>(numAnnotations);
            result.add(tmp);
            for (int r = 0; r < numAnnotations; r++) {
                tmp.add(Annotation.read(in, cpool, major, minor));
            }
        }
        return result;
    }

    @Override
    public Iterator<List<Annotation>> iterator() {
        return annotations.iterator();
    }
}
