package teajay.bytecode.attributes.bootstrap;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.U2;
import teajay.bytecode.cpool.CPoolEntry;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.ConstantPoolAddable;
import teajay.bytecode.cpool.MethodHandleInfo;
import teajay.bytecode.exceptions.ByteCodeException;


/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class BootstrapMethodEntry implements ByteCodeWritable, ConstantPoolAddable {

    private final MethodHandleInfo<?> methodHandle;
    private final CPoolEntry[] arguments;
    private final U2 index;

    BootstrapMethodEntry(MethodHandleInfo<?> methodHandle, CPoolEntry[] arguments, U2 index) throws ByteCodeException {
        this.methodHandle = methodHandle;
        this.arguments = arguments;
        switch (methodHandle.getReferenceKind()) {
            case REF_invokeStatic:
            case REF_newInvokeSpecial:
                break;
            default:
                throw new ByteCodeException("Invalid refence kind for a bootstrap method: " + methodHandle.getReferenceKind());
        }
        this.index = index;
    }
    
    public U2 getIndex() throws ByteCodeException {
        return index;
    }
    
    public List<CPoolEntry>getArguments() {
        return Collections.unmodifiableList(Arrays.asList(arguments));
    }
    
    public MethodHandleInfo<?> getMethodHandle() {
        return methodHandle;
    }
    
    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        methodHandle.writeIndex(out, cpool);
        new U2(arguments.length).write(out, cpool);
        for (CPoolEntry entry : arguments) {
            entry.writeIndex(out, cpool);
        }
    }

    public int getLength() {
        return 2 + 2 + arguments.length * 2;
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        methodHandle.add(cpool);
        for (CPoolEntry entry : arguments) {
            entry.add(cpool);
        }
    }
    
    @Override
    public boolean equals(Object obj) {
        if (obj instanceof BootstrapMethodEntry) {
            BootstrapMethodEntry other = (BootstrapMethodEntry) obj;
            return methodHandle.equals(other.methodHandle) && index.equals(other.index) && Arrays.equals(arguments, other.arguments);
        }
        return false;
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 67 * hash + Objects.hashCode(this.methodHandle);
        hash = 67 * hash + Arrays.deepHashCode(this.arguments);
        hash = 67 * hash + Objects.hashCode(this.index);
        return hash;
    }
}
