package teajay.bytecode.attributes.bootstrap;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import teajay.bytecode.U2;
import teajay.bytecode.U4;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.cpool.CPoolEntry;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.MethodHandleInfo;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;
import static teajay.bytecode.util.Tools.ccpt;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class BootstrapMethods extends AttributeInfo {

    public static final String NAME = "BootstrapMethods";
    private final List<BootstrapMethodEntry> methods = new ArrayList<>();

    public BootstrapMethods() {
        this(new Utf8Info(NAME));
    }

    private BootstrapMethods(Utf8Info name) {
        super(name);
        checkName(NAME, name);
    }

    public BootstrapMethodEntry addBootstrapMethod(MethodHandleInfo<?> handle, CPoolEntry... staticArgs) throws ByteCodeException {
        for (CPoolEntry arg : staticArgs) {
            switch (arg.getTag()) {
                case String:
                case Class:
                case Integer:
                case Long:
                case Float:
                case Double:
                case MethodHandle:
                case MethodType:
                    break;
                default:
                    throw new ByteCodeException("Invalid argument type for a bootstrap method.");
            }
        }
        BootstrapMethodEntry entry = new BootstrapMethodEntry(handle, staticArgs.clone(), new U2(methods.size()));
        methods.add(entry);
        return entry;
    }

    public List<BootstrapMethodEntry> getBootstrapMethods() {
        return Collections.unmodifiableList(methods);
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        int len = 2;
        for (BootstrapMethodEntry entry : methods) {
            len += entry.getLength();
        }
        return new U4(len);
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        super.add(cpool);
        for (BootstrapMethodEntry entry : methods) {
            entry.add(cpool);
        }
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws ByteCodeException, IOException {
        super.write(out, cpool);
        new U2(methods.size()).write(out, cpool);
        for (BootstrapMethodEntry entry : methods) {
            entry.write(out, cpool);
        }
    }

    public static BootstrapMethods read(DataInputStream in, ConstantPool cpool, int major, int minor, int length, Utf8Info name) throws ByteCodeException, IOException {
        int numMethods = in.readUnsignedShort();
        BootstrapMethods attr = new BootstrapMethods(name);
        for (int i = 0; i < numMethods; i++) {
            MethodHandleInfo<?> methodHandle = ccpt(cpool.get(in.readUnsignedShort()), MethodHandleInfo.class, cpool);
            int numBootstrapArgs = in.readUnsignedShort();
            CPoolEntry[] args = new CPoolEntry[numBootstrapArgs];
            for (int r = 0; r < numBootstrapArgs; r++) {
                args[r] = cpool.get(in.readUnsignedShort());
            }
            attr.addBootstrapMethod(methodHandle, args);
        }
        return attr;
    }
}
