package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Collection;
import teajay.bytecode.*;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.ConstantPoolAddable;
import teajay.bytecode.exceptions.ByteCodeAddressOverflow;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public abstract class JVMInstruction implements Operand, ConstantPoolAddable {

    private static long MAX_U4 = (1L << 32) - 1;
    private long index = -1;

    /**
     * Returns the length in byte of this instruction. An invocation of
     * validate() may change the length of an instruction.
     *
     * @return byte-length of this instruction
     * @throws ByteCodeException if the class could not yet decide for a length.
     */
    abstract int getLength() throws ByteCodeException;

    /**
     * This is only the opcode of the virtual instruction. The implementing
     * class may decide to generate equivalent code.
     *
     * @return the virtual opcode
     * @throws ByteCodeException if the class could not yet decide for an
     * opcode.
     */
    public abstract OpCode opcode() throws ByteCodeException;

    /**
     * This are only the operands of the virtual instruction. The implementing
     * class may decide to generate equivalent code.
     *
     * @return list of virtual operands
     * @throws teajay.bytecode.exceptions.ByteCodeException
     */
    protected abstract Collection<? extends Operand> operands() throws ByteCodeException;

    abstract int getLocalModifier() throws ByteCodeException;

    abstract int getPush() throws ByteCodeException;

    abstract int getPop() throws ByteCodeException;

    /**
     * If this method returns false all addresses have to be recalculated and
     * then all instructions have to be revalidated. This has to be done until
     * every instruction is valid.
     *
     * @param cpool the value of cpool
     * @return true only if the instruction is valid
     * @throws ByteCodeException
     */
    abstract boolean validate(ConstantPool cpool) throws ByteCodeException;

    final void setAddress(long index) throws ByteCodeException {
        if (index < 0 || index > MAX_U4) {
            throw new ByteCodeException("Instruction address out of bounds: " + index);
        }
        this.index = index;
    }

    public final boolean hasAddress() {
        return index >= 0;
    }

    @Override
    public final long getAddress(ConstantPool cpool) throws ByteCodeException {
        if (index < 0) {
            throw new ByteCodeException("Trying to access an not yet set address of an vm-instruction.");
        }
        return index;
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        for (Operand op : operands()) {
            if (op instanceof ConstantPoolAddable) {
                ((ConstantPoolAddable) op).add(cpool);
            }
        }
    }

    @Override
    public OpCode.OpSize getDefaultLengthAsOperand() {
        return OpCode.OpSize.U2;
    }

    @Override
    public void writeIndex(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        new U2(getAddress(cpool)).write(out, cpool);
    }

    @Override
    public String toString() {
        try {
            return getAddress(null) + ": " + opcode() + " - " + operands();
        } catch (ByteCodeException ex) {
            try {
                return String.valueOf(opcode() + " - " + operands());
            } catch (ByteCodeException e) {
                return "Unknown instruction: " + e.getMessage();
            }
        }
    }

    S4 diffS4(Operand target, ConstantPool cpool) throws ByteCodeException {
        return new S4(target.getAddress(cpool) - getAddress(cpool));
    }

    S2 diffS2(Operand target, ConstantPool cpool) throws ByteCodeException {
        long diff = target.getAddress(cpool) - getAddress(cpool);
        if (diff < Short.MIN_VALUE || diff > Short.MAX_VALUE) {
            throw new ByteCodeAddressOverflow("requesting a singed 16 bit offset but offset exceeds that range");
        } else {
            return new S2(diff);
        }
    }
    
    @Override
    public final boolean equals(Object obj) {
        return obj == this;
    }

    @Override
    public final int hashCode() {
        return super.hashCode();
    }
}
