package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.*;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.U2;
import teajay.bytecode.U4;
import teajay.bytecode.VersionConstants;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.VMStackUnderflow;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class JVMInstructionList implements ByteCodeWritable, Iterable<JVMInstruction> {

    private final List<JVMInstruction> instructions = new ArrayList<>();
    private final Map<JVMInstruction, Set<Label>> labels = new HashMap<>();
    private final StackMapTable stackMap;
    private int initialLocals = 0;
    private boolean locked = false;
    private int mark = -2;
   
    JVMInstructionList(StackMapTable tbl) {
        this.stackMap = tbl;
    }

    private void putLabels(JVMInstruction instr, Collection<Label> label) throws ByteCodeException {
        Set<Label> set = labels.get(instr);
        if (set == null) {
            set = new HashSet<>(8);
            labels.put(instr, set);
        }
        for (Label l : label) {
            l.setTarget(instr);
        }
        set.addAll(label);
    }

    private void putLabel(JVMInstruction instr, Label label) throws ByteCodeException {
        Set<Label> set = labels.get(instr);
        if (set == null) {
            set = new HashSet<>(8);
            labels.put(instr, set);
        }
        label.setTarget(instr);
        set.add(label);
    }

    private void checkLocked() throws IllegalStateException {
        if (locked) {
            throw new IllegalStateException("the instruction list is locked");
        }
    }

    public JVMInstruction pushInt(int val) throws ByteCodeException {
        return pushInt(null, val);
    }

    public JVMInstruction pushInt(Label label, int val) throws ByteCodeException {
        JVMInstruction instr;
        if (val <= Short.MAX_VALUE && val >= Short.MIN_VALUE) {
            instr = new IConst(val);
        } else {
            instr = new Ldc(val);
        }
        if (label != null) {
            return add(label, instr);
        } else {
            return add(instr);
        }
    }

    public JVMInstruction pushLong(long val) throws ByteCodeException {
        return pushLong(null, val);
    }

    public JVMInstruction pushLong(Label label, long val) throws ByteCodeException {
        JVMInstruction instr;
        if (val == 0L) {
            instr = new LConst0();
        } else if (val == 1L) {
            instr = new LConst1();
        } else {
            instr = new Ldc(val);
        }
        if (label != null) {
            return add(label, instr);
        } else {
            return add(instr);
        }
    }

    public JVMInstruction pushFloat(float val) throws ByteCodeException {
        return pushFloat(null, val);
    }

    public JVMInstruction pushFloat(Label label, float val) throws ByteCodeException {
        JVMInstruction instr;
        if (val == 0f) {
            instr = new FConst0();
        } else if (val == 1f) {
            instr = new FConst1();
        } else if (val == 2f) {
            instr = new FConst2();
        } else {
            instr = new Ldc(val);
        }
        if (label != null) {
            return add(label, instr);
        } else {
            return add(instr);
        }
    }

    public JVMInstruction pushDouble(double val) throws ByteCodeException {
        return pushDouble(null, val);
    }

    public JVMInstruction pushDouble(Label label, double val) throws ByteCodeException {
        JVMInstruction instr;
        if (val == 0d) {
            instr = new DConst0();
        } else if (val == 1d) {
            instr = new DConst1();
        } else {
            instr = new Ldc(val);
        }
        if (label != null) {
            return add(label, instr);
        } else {
            return add(instr);
        }
    }
    
    /**
     * Adds the instruction to the list.
     *
     * @param instr Instruction to be added.
     * @throws RuntimeException if the list is already locked
     */
    public JVMInstruction add(JVMInstruction instr) throws ByteCodeException {
        return add((Label) null, instr);
    }
    
    /**
     * Adds the instruction to the list.
     *
     * @param labels the label to be assigned to this instruction
     * @param instr Instruction to be added.
     * @throws RuntimeException if the list is already locked
     */
    public JVMInstruction add(Collection<Label> labels, JVMInstruction instr) throws ByteCodeException {
        checkLocked();
        if (instructions.size() >= 0xFFFF) {
            throw new RuntimeException("instruction list overflow");
        }
        if (instructions.contains(instr)) {
            throw new ByteCodeException("instruction list already contains the supplied instruction");
        }
        instructions.add(instr);
        if (labels != null) {
            putLabels(instr, labels);
        }
        return instr;
    }

    public JVMInstruction add(Label label, JVMInstruction instr) throws ByteCodeException {
        return add(label != null ? Arrays.asList(label) : null, instr);
    }

    /**
     * After invokating this method the instruction list is locked and no
     * instructions may be added.
     *
     * @return byte-length of the instruction list
     * @throws ByteCodeException
     */
    U4 getLength(ConstantPool cpool) throws ByteCodeException {
        validateCycle(cpool);
        int len = 0;
        for (JVMInstruction instr : instructions) {
            len += instr.getLength();
        }
        if (len > VersionConstants.MAX_CODE_LENGTH) {
            throw new ByteCodeException("instruction list's maximal code length exceeded: " + len);
        }
        return new U4(len);
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        validateCycle(cpool);
        getLength(cpool).write(out, cpool);
        
        for (JVMInstruction instr : instructions) {
            instr.write(out, cpool);
        }
    }

    @Override
    public Iterator<JVMInstruction> iterator() {
        return new Iterator<JVMInstruction>() {
            private final Iterator<JVMInstruction> it = instructions.iterator();
            private JVMInstruction current;

            @Override
            public boolean hasNext() {
                return it.hasNext();
            }

            @Override
            public JVMInstruction next() {
                return current = it.next();
            }

            @Override
            public void remove() {
                if (current != null && !locked) {
                    JVMInstructionList.this.remove(current, false);
                    it.remove();
                } else {
                    throw new IllegalStateException();
                }
            }
        };
    }

    /**
     * Delegates the add(ConstantPool) call to all instructions in the list.
     *
     * @param cpool the ConstantPool
     * @throws ByteCodeException
     */
    public void add(ConstantPool cpool) throws ByteCodeException {
        for (JVMInstruction instr : instructions) {
            instr.add(cpool);
        }
    }

    private void validateCycle(ConstantPool cpool) throws ByteCodeException {
        if (!locked) {
            locked = true;
            long cnt = 0;
            for (JVMInstruction instr : instructions) {
                instr.setAddress(cnt);
                cnt += instr.getLength();
            }
            boolean success;
            do {
                cnt = 0;
                success = true;
                for (JVMInstruction instr : instructions) {
                    instr.setAddress(cnt);
                    success &= instr.validate(cpool);
                    cnt += instr.getLength();
                }
            } while (!success);
        }
    }

    void validate(ConstantPool cpool) throws ByteCodeException {
        validateCycle(cpool);
    }

    /**
     * Sets the initital local variable count. For example non-static Methods do
     * have the implicit this pointer as their first local variable.
     *
     * @param num count of initital locals
     */
    void setInititalLocalsCount(int num) {
        checkLocked();
        this.initialLocals = num;
    }

    /**
     * Calculates the maximum size the operand stack will reach during
     * execution. This method locks the instruction list. If this method fails
     * either the byte code itself or the stack map table is invalid.
     *
     * @param table the stack map table for this code
     * @return the maximum stack size as an unsigned 16bit integer
     * @throws ByteCodeException if validation fails
     * @throws VMStackUnderflow tf the operand stack underflows at some point
     */
    U2 getMaxStackSize(StackMapTable table, ConstantPool cpool) throws ByteCodeException, VMStackUnderflow {
        validateCycle(cpool);
        int max = 0;
        int current = 0;
        if (instructions.size() > 0) {
            for (JVMInstruction instr : instructions) {
                StackMapFrame frame = table.getFrame(instr);
                if (frame != null) {
                    List<VerificationType> currentStack = frame.getStack();
                    current = currentStack.size();
                    for (VerificationType type : currentStack) {
                        //long and double take up two stack items
                        switch (type.getTag()) {
                            case Long:
                            case Double:
                                current++;
                        }
                    }
                }
                current -= instr.getPop();
                if (current < 0) {
                    throw new VMStackUnderflow("Trying to pop from an empty stack: " + instr);
                }
                current += instr.getPush();

                if (current > max) {
                    max = current;
                }
            }
        }

        return new U2(max);
    }

    /**
     * Calculates the max locals count for this method. After invocation of this
     * method the instruction list is locked.
     *
     * @return the maximum size of the local variable table for this method.
     * @throws ByteCodeException if validation of this list fails
     */
    U2 getMaxLocals(ConstantPool cpool) throws ByteCodeException {
        validateCycle(cpool);
        int max = initialLocals;
        for (JVMInstruction instr : instructions) {
            int locals = instr.getLocalModifier();
            if (locals > max) {
                max = locals;
            }
        }
        return new U2(max);
    }

    /**
     * Returns the label that is assigned to <tt>instr</tt>.
     *
     * @param instr the instruction to be looked up
     * @return A Label or null if this list does not contain <tt>instr</tt> or
     * if <tt>instr</tt> does not have a label.
     */
    public Set<Label> getLabelsFor(JVMInstruction instr) {
        Set<Label> set = labels.get(instr);
        if (set != null && !set.isEmpty()) {
            return Collections.unmodifiableSet(set);
        } else {
            return Collections.emptySet();
        }
    }

    public void addLabelFor(JVMInstruction instr, Label label) throws ByteCodeException {
        checkLocked();
        if (label.isBound()) {
            throw new ByteCodeException("cannot rebind already bound label");
        }
        if (!instructions.contains(instr)) {
            throw new ByteCodeException("the supplied instruction is not contained within this instruction list");
        }
        if (getLabelsFor(instr).contains(label)) {
            throw new ByteCodeException("there is a label already pointing to this instruction");
        }
        putLabel(instr, label);
    }

    public void addLabelsFor(JVMInstruction instr, Collection<Label> lst) throws ByteCodeException {
        for (Label label : lst) {
            addLabelFor(instr, label);
        }
    }

    /**
     * Removes all instructions from this list.
     */
    public void clear() {
        checkLocked();
        for (JVMInstruction instr : instructions) {
            removeLabelAndStackMapFrame(instr);
        }
        labels.clear();
        instructions.clear();
        this.mark = -2;
    }

    /**
     * Sets a marking at the current position in the list.
     */
    public void mark() {
        this.mark = instructions.size() - 1;
    }

    /**
     * Removes a marking set by <tt>mark()</tt>.
     */
    public void removeMarking() {
        this.mark = -2;
    }

    /**
     * Reset the list to a marked stage and deletes the marking.
     */
    public void reset() {
        checkLocked();
        if (mark > -2) {
            for (int i = instructions.size() - 1; i > mark; i--) {
                removeLabelAndStackMapFrame(instructions.remove(i));
            }
            mark = -2;
        } else {
            throw new IllegalStateException("no marking present");
        }
    }

    /**
     *
     * @param oldInstr
     * @param newInstr
     * @param keepLabel
     * @throws ByteCodeException
     */
    public void replace(JVMInstruction oldInstr, JVMInstruction newInstr, boolean keepLabel) throws ByteCodeException {
        checkLocked();
        Set<Label> set = removeLabelAndStackMapFrame(oldInstr);
        if (keepLabel && set != null) {
            replace(oldInstr, set, newInstr);
        } else {
            int index = instructions.indexOf(oldInstr);
            if (index > -1) {
                instructions.set(index, newInstr);
            } else {
                throw new IllegalArgumentException("no such instruction");
            }
        }
    }

    public void replace(JVMInstruction oldInstr, JVMInstruction newInstr) throws ByteCodeException {
        replace(oldInstr, newInstr, false);
    }

    public void replace(JVMInstruction oldInstr, Collection<Label> newLabels, JVMInstruction newInstr) throws ByteCodeException {
        checkLocked();
        int index = instructions.indexOf(oldInstr);
        if (index > -1) {
            instructions.set(index, newInstr);
            removeLabelAndStackMapFrame(oldInstr);
            putLabels(newInstr, newLabels);
        } else {
            throw new IllegalArgumentException("no such instruction");
        }
    }

    public void replace(JVMInstruction oldInstr, Label newLabel, JVMInstruction newInstr) throws ByteCodeException {
        replace(oldInstr, Arrays.asList(newLabel), newInstr);
    }

    public JVMInstruction getPredecessor(JVMInstruction instr) {
        int index = instructions.indexOf(instr);
        if (index > 0) {
            return instructions.get(index - 1);
        } else {
            throw new IllegalArgumentException("no such instruction");
        }
    }

    public JVMInstruction getSuccessor(JVMInstruction instr) {
        int index = instructions.indexOf(instr);
        if (index > -1 && index < instructions.size() - 1) {
            return instructions.get(index + 1);
        } else {
            throw new IllegalArgumentException("no such instruction");
        }
    }

    public JVMInstruction getFirst() throws ByteCodeException {
        if (!instructions.isEmpty()) {
            return instructions.get(0);
        } else {
            throw new IllegalArgumentException("no such instruction");
        }
    }

    public JVMInstruction getLast() throws ByteCodeException {
        int size = instructions.size();
        if (size > 0) {
            return instructions.get(size - 1);
        } else {
            throw new IllegalArgumentException("no such instruction");
        }
    }

    private void remove(JVMInstruction instr, boolean reallyDo) {
        checkLocked();
        int index = instructions.indexOf(instr);
        if (index > -1) {
            removeLabelAndStackMapFrame(instr);
            if (index == mark) {
                mark = -2;
            } else if (index < mark) {
                mark--;
            }
            if (reallyDo) {
                instructions.remove(instr);
            }
        } else {
            throw new IllegalArgumentException("no such instruction");
        }
    }

    public void remove(JVMInstruction instr) {
        remove(instr, true);
    }

    private Set<Label> removeLabelAndStackMapFrame(JVMInstruction instr) {
        Set<Label> set = labels.remove(instr);
        if (set != null) {
            for (Label label : set) {
                label.reset();
            }
        }
        stackMap.remove(instr);
        return set;
    }

    public int size() {
        return instructions.size();
    }
}
