package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Collections;
import java.util.List;
import teajay.bytecode.U1;
import teajay.bytecode.U4;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.FrameOffsetDeltaOutOfBounds;

/**
 * If the offset delta exceeds 63 this frame automatically becomes a same_frame_extended.
 * @author Michael Kußmann <michi@outx.de>
 */
public final class SameFrame extends StackMapFrame {

    private SameFrameExtended extended = null;

    SameFrame(JVMInstruction start) throws ByteCodeException {
        super(start);
    }

    @Override
    U4 getLength() throws ByteCodeException {
        if (extended == null) {
            return new U4(1);
        } else {
            return extended.getLength();
        }
    }

    @Override
    U1 getFrameType() throws ByteCodeException {
        int offsetDelta = getOffsetDelta();
        if (offsetDelta < 0) {
            throw new FrameOffsetDeltaOutOfBounds("SameFrame: offset delta out of bounds: " + offsetDelta);
        } else if (offsetDelta > 63) {
            return new U1(SameFrameExtended.FRAME_TYPE);
        } else {
            return new U1(offsetDelta);
        }
    }
      
    @Override
    void validate() throws ByteCodeException {
        int offsetDelta = getOffsetDelta();
        if (offsetDelta > 63 && extended == null) {
            extended = new SameFrameExtended(getStart());
            extended.setOffsetDelta(offsetDelta);
        } else if (offsetDelta > 63) {
            extended.setOffsetDelta(offsetDelta);
        } else {
            extended = null;
        }
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        if (extended != null) {
            extended.add(cpool);
        }
    }

    @Override
    List<VerificationType> getStack() {
        return Collections.emptyList();
    }
    
    @Override
    List<VerificationType> getLocals() throws ByteCodeException {
        return this.getPredecessor().getLocals();
    }
    
    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        if (extended != null) {
            extended.write(out, cpool);
        } else {
            super.write(out, cpool);
        }
    }
}
