package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import teajay.bytecode.U1;
import teajay.bytecode.U2;
import teajay.bytecode.U4;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.util.UnmodifiableIterator;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class StackMapTable extends AttributeInfo implements Iterable<StackMapFrame> {
    public static final String NAME = "StackMapTable";

    private static final StackMapFrameComparator comparator = new StackMapFrameComparator();
    private final List<StackMapFrame> frames;
    private final FullFrame first;
    private boolean locked = false;

    StackMapTable(FullFrame first) {
        super(new Utf8Info(NAME));
        this.frames = new ArrayList<>();
        this.first = first;
    }

    private void add(StackMapFrame frame) throws ByteCodeException {
        if (!locked) {
            if (frame.isAdded()) {
                throw new ByteCodeException("frame has already been added to a stack map table");
            }
            frames.add(frame);
            frame.setAdded(true);
        } else {
            throw new ByteCodeException("Cannot add a stack map frame to a locked stack map table");
        }
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        super.add(cpool);
        for (StackMapFrame frame : frames) {
            frame.add(cpool);
        }
        first.add(cpool);
    }

    private void calcOffsetDeltas() throws ByteCodeException {
        if (frames.size() > 0 && !locked) {
            this.locked = true;
            Collections.sort(frames, comparator);
            int offset = 0;
            StackMapFrame lastFrame = first;
            for (StackMapFrame frame : frames) {
                JVMInstruction start = frame.getStart();
                int offsetDelta = (int) (start.getAddress(null) - offset);
                if (offsetDelta < 0) {
                    throw new ByteCodeException("Bad stack map frame offset: " + offsetDelta);
                }
                frame.setOffsetDelta(offsetDelta);
                offset += offsetDelta + 1;
                frame.setPredecessor(lastFrame);
                lastFrame = frame;
                frame.validate();
            }
        }
    }

    public SameFrame addSameFrame(JVMInstruction start) throws ByteCodeException {
        SameFrame same = new SameFrame(start);
        add(same);
        return same;
    }

    public ChopFrame addChopFrame(JVMInstruction start, int num) throws ByteCodeException {
        ChopFrame chop = new ChopFrame(start, new U1(251 - num));
        add(chop);
        return chop;
    }

    public AppendFrame addAppendFrame(JVMInstruction start) throws ByteCodeException {
        AppendFrame append = new AppendFrame(start);
        add(append);
        return append;
    }

    public FullFrame addFullFrame(JVMInstruction start) throws ByteCodeException {
        FullFrame full = new FullFrame(start);
        add(full);
        return full;
    }

    public SameLocalsOneStackItemFrame addSameLocalsOneStackItemFrame(JVMInstruction start) throws ByteCodeException {
        SameLocalsOneStackItemFrame one = new SameLocalsOneStackItemFrame(start);
        add(one);
        return one;
    }
    
    public void addFrame(StackMapFrame frame, JVMInstruction instruction) throws ByteCodeException {
        if (!frame.isAdded()) {
            frame.setStart(instruction);
            add(frame);
        } else {
            throw new ByteCodeException("frame has already been added to a stack map table");
        }
    }
    
    public StackMapFrame remove(JVMInstruction start) {
        Iterator<StackMapFrame> it = frames.iterator();
        while (it.hasNext()) {
            StackMapFrame frame = it.next();
            if (frame.getStart() == start) {
                it.remove();
                frame.setAdded(false);
                return frame;
            }
        }
        return null;
    }
    
    public void remove(StackMapFrame frame) {
        frames.remove(frame);
        frame.setAdded(false);
    }

    public boolean isEmpty() {
        return frames.isEmpty();
    }

    public StackMapFrame getFrame(JVMInstruction start) {
        for (StackMapFrame frame : frames) {
            if (frame.getStart() == start) {
                return frame;
            }
        }
        return null;
    }

    public FullFrame getFirstFrame() {
        return first;
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        int len = 0;
        for (StackMapFrame frame : frames) {
            len += frame.getLength().getValue();
        }
        return new U4(2 + len);
    }

    void validate() throws ByteCodeException {
        calcOffsetDeltas();
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        validate();
        super.write(out, cpool);
        new U2(frames.size()).write(out, cpool);
        for (StackMapFrame frame : frames) {
            frame.write(out, cpool);
        }
    }

    @Override
    public Iterator<StackMapFrame> iterator() {
        return new UnmodifiableIterator<>(frames.iterator());
    }
}
