package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Collection;
import java.util.Map;
import java.util.Map.Entry;
import java.util.SortedMap;
import java.util.TreeMap;
import teajay.bytecode.OpCode;
import teajay.bytecode.Operand;
import teajay.bytecode.S4;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 * This class implements both of the VM's switch instructions (lookupswitch and
 * tableswitch). The class decides by itself which one is to be used.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class Switch extends JVMInstruction {

    private static final float DEFAULT_MAX_OVERHEAD = 1.15f;
    private float maxOverhead = DEFAULT_MAX_OVERHEAD;
    private Operand defaultTarget;
    /**
     * It is important that this map is sorted, thus TreeMap.
     */
    SortedMap<Integer, Operand> table = new TreeMap<>();
    private int padding = 0;
    private boolean locked = false;
    /**
     * Default behaviour is to use lookupswitch.
     */
    private OpCode opcode = OpCode.lookupswitch;
    /**
     * low and high value for tableswitch.
     */
    private int low = 0;
    private int high = 0;
    /**
     * Stores whether the compactness is yet to be checked.
     */
    private boolean checkCompact = true;

    /**
     * Creates a new Switch instruction.
     */
    public Switch() {
        this.defaultTarget = null;
    }

    /**
     * Creates a new Switch instruction.
     *
     * @param maxOverhead the maximum allowed overhead for a tableswitch
     * instruction.
     */
    public Switch(float maxOverhead) {
        this.defaultTarget = null;
        this.maxOverhead = maxOverhead;
    }

    /**
     * Creates a new Switch instruction.
     *
     * @param defaultTarget The branch target for the default case.
     */
    public Switch(JVMInstruction defaultTarget) {
        this.defaultTarget = defaultTarget;
    }

    /**
     * Creates a new Switch instruction.
     *
     * @param defaultTarget The branch target for the default case.
     */
    public Switch(Label defaultTarget) {
        this.defaultTarget = defaultTarget;
    }

    /**
     * Creates a new Switch instruction.
     *
     * @param defaultTarget The branch target for the default case.
     * @param maxOverhead the maximum allowed overhead for a tableswitch
     * instruction.
     */
    public Switch(Label defaultTarget, float maxOverhead) {
        this.defaultTarget = defaultTarget;
        this.maxOverhead = maxOverhead;
    }

    /**
     * Creates a new Switch instruction.
     *
     * @param defaultTarget The branch target for the default case.
     * @param maxOverhead the maximum allowed overhead for a tableswitch
     * instruction.
     */
    public Switch(JVMInstruction defaultTarget, float maxOverhead) {
        this.defaultTarget = defaultTarget;
        this.maxOverhead = maxOverhead;
    }

    @Override
    int getLength() throws ByteCodeException {
        switch (opcode) {
            case lookupswitch:
                return lookupSize(table) + padding + 1;
            default:
                return tableSize(high, low) + padding + 1;
        }
    }

    private static int lookupSize(Map<?,?> table) {
        return 4 + 4 + table.size() * 8;
    }

    private static int tableSize(int high, int low) {
        return 4 + 4 + 4 + (high - low + 1) * 4;
    }

    @Override
    public OpCode opcode() throws ByteCodeException {
        return opcode;
    }

    @Override
    protected Collection<Operand> operands() {
        return table.values();
    }

    @Override
    int getLocalModifier() {
        return 0;
    }

    @Override
    int getPush() {
        return 0;
    }

    @Override
    int getPop() {
        return 1;
    }

    @Override
    boolean validate(ConstantPool cpool) throws ByteCodeException {
        locked = true;
        boolean ret = true;

        if (checkCompact) {
            this.low = table.firstKey();
            this.high = table.lastKey();
            //decide whether to use lookupswitch or tableswitch
            if (checkCompact(table, high, low)) {
                this.opcode = OpCode.tableswitch;
                ret = false;
            } else {
                ret = true;
            }
            //make sure to not check again
            checkCompact = false;
        }
        //calculate the padding
        long address = getAddress(cpool) + 1; //start of the instruction's body
        if ((address + padding) % 4 != 0) {
            //the table has to be aligned to a an address divisible by 4
            if (address % 4 == 0) {
                padding = 0;
            } else {
                padding = (int) (4 - address % 4);
            }
            return false;
        } else {
            return ret;
        }
    }

    private boolean checkCompact(Map<?,?> table, int high, int low) {
        if (table.size() > 2) {
            int tsize = tableSize(high, low);
            int lsize = lookupSize(table);
            float overhead = (float) tsize / (float) lsize;
            return overhead <= maxOverhead;
        } else {
            return false;
        }
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        locked = true;
        opcode.write(out, cpool);
        switch (opcode) {
            case lookupswitch:
                for (int i = 0; i < padding; i++) {
                    OpCode.nop.write(out, cpool);
                }
                this.diffS4(defaultTarget, cpool).write(out, cpool);
                out.writeInt(table.size());
                for (Entry<Integer, Operand> entry : table.entrySet()) {
                    out.writeInt(entry.getKey());
                    this.diffS4(entry.getValue(), cpool).write(out, cpool);
                }
                break;
            default:
                for (int i = 0; i < padding; i++) {
                    OpCode.nop.write(out, cpool);
                }
                S4 defaultOffset = this.diffS4(defaultTarget, cpool);
                defaultOffset.write(out, cpool);
                out.writeInt(low);
                out.writeInt(high);

                for (int i = low; i <= high; i++) {
                    Operand target = table.get(i);
                    if (target != null) {
                        this.diffS4(target, cpool).write(out, cpool);
                    } else {
                        defaultOffset.write(out, cpool);
                    }
                }
        }
    }

    /**
     * Set the default case.
     *
     * @param label default target
     */
    public void setDefaultCase(Label label) {
        this.defaultTarget = label;
    }

    /**
     * Set the default case.
     *
     * @param defaultTarget default target
     */
    public void setDefaultCase(JVMInstruction defaultTarget) {
        this.defaultTarget = defaultTarget;
    }

    /**
     * Adds a case to this switch instruction.
     *
     * @param key match key
     * @param branchTarget branch target
     */
    public void addCase(int key, JVMInstruction branchTarget) {
        if (!locked) {
            table.put(key, branchTarget);
        } else {
            throw new RuntimeException("Switch is locked.");
        }
    }

    /**
     * Adds a case to this switch instruction.
     *
     * @param key match key
     * @param branchTarget branch target
     */
    public void addCase(int key, Label branchTarget) {
        if (!locked) {
            table.put(key, branchTarget);
        } else {
            throw new RuntimeException("Switch is locked.");
        }
    }
}
