package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Objects;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.OpCode;
import teajay.bytecode.Operand;
import teajay.bytecode.U4;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.ConstantPoolAddable;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class VerificationType implements ByteCodeWritable, ConstantPoolAddable {
    private final static VerificationType top = new VerificationType(VariableInfo.Top);
    private final static VerificationType integer = new VerificationType(VariableInfo.Integer);
    private final static VerificationType _float = new VerificationType(VariableInfo.Float);
    private final static VerificationType _long = new VerificationType(VariableInfo.Long);
    private final static VerificationType _double = new VerificationType(VariableInfo.Double);
    private final static VerificationType _null = new VerificationType(VariableInfo.Null);
    private final static VerificationType uninitializedThis = new VerificationType(VariableInfo.UninitializedThis);

    private final Operand arg;
    private final VariableInfo tag;
    
    private VerificationType(VariableInfo tag, Operand arg) {
        this.tag = tag;
        this.arg = arg;
    }

    VerificationType(VariableInfo tag) {
        this(tag, null);
    }

    VerificationType(ClassInfo className) {
        this(className != null ? VariableInfo.Object : VariableInfo.Null, className);
    }

    VerificationType(JVMInstruction newInstr) throws ByteCodeException {
        this(VariableInfo.Uninitialized, newInstr);
        OpCode opcode = newInstr.opcode();
        if (!(opcode == OpCode.vm_new || opcode == OpCode.anewarray || opcode == OpCode.newarray)) {
            throw new ByteCodeException(VariableInfo.Uninitialized + " has to point to a new operation");
        }
    }
    
    public static VerificationType makeTop() {
        return top;
    }
    
    public static VerificationType makeInteger() {
        return integer;
    }
    
    public static VerificationType makeFloat() {
        return _float;
    }

    public static VerificationType makeDouble() {
        return _double;
    }
    
    public static VerificationType makeLong() {
        return _long;
    }

    public static VerificationType makeNull() {
        return _null;
    }

    public static VerificationType makeUninitializedThis() {
        return uninitializedThis;
    }
    
    public static VerificationType makeObject(ClassInfo type) {
        return new VerificationType(type);
    }
    
    public static VerificationType makeUninitialized(JVMInstruction newInstr) throws ByteCodeException {
        return new VerificationType(newInstr);
    }
    
    public static VerificationType valueOf(FieldDescriptor type) throws ByteCodeException {
            if (type.isPrimitive()) {
                switch (type.getBaseType()) {
                    case Void:
                        throw new ByteCodeException("void is not valid in a stack map frame");
                    case Boolean:
                    case Byte:
                    case Short:
                    case Char:
                    case Int:
                        return makeInteger();
                    case Long:
                        return makeLong();
                    case Float:
                        return makeFloat();
                    case Double:
                        return makeDouble();
                    default:
                        throw new RuntimeException("unreachable statement reached");
                }
            } else if (type.getClassInfo().getClassName().equals("null")) {
                return makeNull();
            } else {
                return makeObject(type.getClassInfo());
            }
    }
    
    public static VerificationType valueOf(ClassInfo type) throws ByteCodeException {
            if (type.isPrimitive()) {
                switch (type.getClassName()) {
                    case "void":
                        throw new ByteCodeException("void is not valid in a stack map frame");
                    case "boolean":
                    case "byte":
                    case "short":
                    case "char":
                    case "int":
                        return makeInteger();
                    case "long":
                        return makeLong();
                    case "float":
                        return makeFloat();
                    case "double":
                        return makeDouble();
                    case "null":
                        return makeNull();
                    default:
                        throw new RuntimeException("unreachable statement reached");
                }
            } else {
                return makeObject(type);
            }
    }
    
    public static VerificationType valueOf(Class<?> cl) throws ByteCodeException {
        if (!cl.isPrimitive()) {
            return makeObject(ClassInfo.fromClass(cl));
        } else if (cl.isPrimitive()) {
            if (cl.equals(void.class)) {
                throw new ByteCodeException("void is not valid in a stack map frame");
            } else if (cl.equals(boolean.class)) {
                return makeInteger();
            } else if (cl.equals(byte.class)) {
                return makeInteger();
            } else if (cl.equals(short.class)) {
                return makeInteger();
            } else if (cl.equals(char.class)) {
                return makeInteger();
            } else if (cl.equals(int.class)) {
                return makeInteger();
            } else if (cl.equals(float.class)) {
                return makeFloat();
            } else if (cl.equals(long.class)) {
                return makeLong();
            } else if (cl.equals(double.class)) {
                return makeDouble();
            }
        }
        throw new RuntimeException("unreachable statement reached!");
    }
    
    public VariableInfo getTag() {
        return tag;
    }

    public Operand getOperand() {
        return arg;
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        tag.write(out, cpool);
        if (arg != null) {
            arg.writeIndex(out, cpool);
        }
    }

    public U4 getLength() {
        return tag.getLength();
    }

    @Override
    public String toString() {
        if (arg != null) {
            return tag + ": " + arg;
        } else {
            return tag.toString();
        }
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        if (this.getTag() == VariableInfo.Object) {
            ((ConstantPoolAddable) arg).add(cpool);
        }
    }

    @Override
    public int hashCode() {
        int hash = 3;
        hash = 97 * hash + Objects.hashCode(this.arg);
        hash = 97 * hash + (this.tag != null ? this.tag.hashCode() : 0);
        return hash;
    }
    
    @Override
    public boolean equals(Object obj) {
        if (obj instanceof VerificationType) {
            VerificationType other = (VerificationType) obj;
            if (tag == other.tag) {
                if (arg == null || other.arg == null) {
                    return arg == other.arg;
                } else {
                    return arg.equals(other.arg);
                }
            }
        }
        return false;
    }
}
