package teajay.bytecode.attributes.nest;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import teajay.bytecode.U2;
import teajay.bytecode.U4;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;
import static teajay.bytecode.util.Tools.ccpt;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class NestMembers extends AttributeInfo {
    public static final String NAME = "NestMembers";
    private final Collection<ClassInfo> classes;
    
    public NestMembers() {
        this(Collections.<ClassInfo>emptyList(), new Utf8Info(NAME));
    }
    
    public NestMembers(Collection<ClassInfo> classes) {
        this(classes, new Utf8Info(NAME));
    }
    
    private NestMembers(Collection<ClassInfo> classes, Utf8Info name) {
        super(name);
        this.classes = new HashSet<>();
        for (ClassInfo exception : classes) {
            this.classes.add(exception);
        }
    }
    
    public void addNestMember(ClassInfo info) {
        classes.add(info);
    }
    
    public boolean removeNestMember(ClassInfo info) {
        return classes.remove(info);
    }
    
    public Collection<ClassInfo> getNestMembers() {
        return Collections.unmodifiableCollection(classes);
    }
    
    public boolean containsNestMember(ClassInfo info) {
        return classes.contains(info);
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        return new U4(2 + classes.size() * 2);
    }
    
    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        super.write(out, cpool);
        new U2(classes.size()).write(out, cpool);
        for (ClassInfo exception : classes) {
            exception.writeIndex(out, cpool);
        }
    }
    
    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        super.add(cpool);
        for (ClassInfo cl : classes) {
            cl.add(cpool);
        }
    }
    
    public static NestMembers read(DataInputStream in, ConstantPool cpool, int major, int minor, int length, Utf8Info name) throws ByteCodeException, IOException {
         int numClasses = in.readUnsignedShort();
         NestMembers ex = new NestMembers(Collections.<ClassInfo>emptyList(), name);
         for (int i = 0; i < numClasses; i++) {
             ClassInfo cl = ccpt(cpool.get(in.readUnsignedShort()), ClassInfo.class, cpool);
             ex.addNestMember(cl);
         }
         return ex;
    }
}
