package teajay.bytecode.cpool;

import teajay.bytecode.BaseType;
import teajay.bytecode.ClassFile;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.MalformedClassNameException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class FieldDescriptor extends Utf8Info {

    private final BaseType type;
    private final int dimension;
    private final ClassInfo classInfo;

    FieldDescriptor(BaseType type, int dimension, ClassInfo classInfo, String val) {
        super(val);
        this.type = type;
        this.dimension = dimension;
        this.classInfo = classInfo;
    }

    /**
     * Creates a descriptor for an array of the base type
     * <code>type</code>.
     *
     * @param type the base type of the array
     * @param dimension the dimension of the array (must not be greater than
     * 255)
     */
    FieldDescriptor(FieldDescriptor type, int dimension) {
        super(buildArrayString(type.getValue(), dimension));
        this.type = BaseType.Array;
        dimension += type.getDimension();
        this.dimension = dimension;
        if (dimension <= 0 || dimension > 255) {
            throw new IllegalArgumentException("Cannot generate Arrays of a dimension less than 1 or greater than 255.");
        }
        try {
            this.classInfo = new ClassInfo(getValue());
        } catch (MalformedClassNameException ex) {
            throw new Error(ex);
        }
    }

    private static String buildArrayString(String append, int dimension) {
        StringBuilder buf = new StringBuilder(append.length() + dimension);
        for (int i = dimension; i > 0; i--) {
            buf.append('[');
        }
        buf.append(append);
        return buf.toString();
    }

    /**
     * Creates a descriptor for a reference to the class
     * <code>className</code>.
     *
     * @param className name of the class
     */
    FieldDescriptor(ClassInfo className) {
        super(className.isPrimitive() ? className.getPrimitiveChar() : "L" + className.getClassName() + ";");
        String name = className.getClassName();
        if (name.length() > 0 && name.charAt(0) == '[') {
            throw new IllegalArgumentException("cannot create array types with this constructor");
        }
        this.type = BaseType.Reference;
        this.dimension = 0;
        this.classInfo = className;
    }

    FieldDescriptor(String className) throws MalformedClassNameException {
        this(new ClassInfo(className));
    }

    FieldDescriptor(Class<?> type) throws MalformedClassNameException {
        this(ClassInfo.fromClass(type));
        if (type.isPrimitive() || type.isArray()) {
            throw new IllegalArgumentException("this constructor cannot be used to create field descriptors of primitive types");
        }
    }

    /**
     * Creates a descriptor for a jvm base type. This constructor cannot create
     * reference or array types.
     *
     * @param type type to be created
     */
    FieldDescriptor(BaseType type) {
        super(type.getValue());
        try {
            switch (type) {
                case Void:
                    this.classInfo = new ClassInfo(void.class.getName());
                    break;
                case Boolean:
                    this.classInfo = new ClassInfo(boolean.class.getName());
                    break;
                case Byte:
                    this.classInfo = new ClassInfo(byte.class.getName());
                    break;
                case Short:
                    this.classInfo = new ClassInfo(short.class.getName());
                    break;
                case Int:
                    this.classInfo = new ClassInfo(int.class.getName());
                    break;
                case Long:
                    this.classInfo = new ClassInfo(long.class.getName());
                    break;
                case Float:
                    this.classInfo = new ClassInfo(float.class.getName());
                    break;
                case Double:
                    this.classInfo = new ClassInfo(double.class.getName());
                    break;
                case Char:
                    this.classInfo = new ClassInfo(char.class.getName());
                    break;
                default:
                    throw new IllegalArgumentException("cannot generate references or arrays with this constructor");
            }
            this.type = type;
            this.dimension = 0;
        } catch (MalformedClassNameException ex) {
            throw new Error("FieldDescriptor: internal error " + ex.getMessage());
        }
    }

    public static FieldDescriptor fromBaseType(BaseType type) {
        return new FieldDescriptor(type);
    }

    public static FieldDescriptor makeArray(FieldDescriptor type, int dimension) {
        return new FieldDescriptor(type, dimension);
    }

    public static FieldDescriptor fromClassFile(ClassFile cFile) {
        return FieldDescriptor.fromClassInfo(cFile.getThis());
    }

    public static FieldDescriptor fromString(String className) {
        try {
            return FieldDescriptor.fromClassInfo(new ClassInfo(className.replace('.', '/')));
        } catch (MalformedClassNameException ex) {
            throw new Error(ex);
        }
    }

    public static FieldDescriptor fromClassInfo(ClassInfo cl) {
        if (!cl.isPrimitive() && !cl.isArray()) {
            return new FieldDescriptor(cl);
        } else if (cl.isPrimitive()) {
            String name = cl.getClassName();
            if (name.equals(void.class.getName())) {
                return new FieldDescriptor(BaseType.Void);
            } else if (name.equals(boolean.class.getName())) {
                return new FieldDescriptor(BaseType.Boolean);
            } else if (name.equals(byte.class.getName())) {
                return new FieldDescriptor(BaseType.Byte);
            } else if (name.equals(short.class.getName())) {
                return new FieldDescriptor(BaseType.Short);
            } else if (name.equals(char.class.getName())) {
                return new FieldDescriptor(BaseType.Char);
            } else if (name.equals(int.class.getName())) {
                return new FieldDescriptor(BaseType.Int);
            } else if (name.equals(float.class.getName())) {
                return new FieldDescriptor(BaseType.Float);
            } else if (name.equals(long.class.getName())) {
                return new FieldDescriptor(BaseType.Long);
            } else if (name.equals(double.class.getName())) {
                return new FieldDescriptor(BaseType.Double);
            } else {
                throw new RuntimeException("unreachable statement reached!");
            }
        } else if (cl.isArray()) {
            FieldDescriptor fd = parseFieldDescriptor(cl.getClassName());
            if (fd != null) {
                return fd;
            } else {
                throw new RuntimeException("encountered invalid class descriptor format: " + cl.getClassName());
            }
        }
        throw new RuntimeException("unreachable statement reached");
    }

    public static FieldDescriptor fromClass(Class<?> cl) {
        return FieldDescriptor.fromString(cl.getName());
    }

    public static FieldDescriptor getVoid() {
        return new FieldDescriptor(BaseType.Void);
    }
    
    public static FieldDescriptor getString() {
        return FieldDescriptor.fromClass(String.class);
    }
    
    public static FieldDescriptor getObject() {
        return FieldDescriptor.fromClass(Object.class);
    }
    
    public static FieldDescriptor getInt() {
        return new FieldDescriptor(BaseType.Int);
    }

   
    @Override
    public String toString() {
        return this.getValue();
    }

    public BaseType getBaseType() {
        return this.type;
    }

    /**
     *
     * @return @deprecated use #getBaseType() instead
     */
    @Deprecated
    public BaseType getLiteral() {
        return getBaseType();
    }

    public int getDimension() {
        return this.dimension;
    }

    public boolean isArray() {
        return this.getDimension() > 0;
    }

    public boolean isReference() {
        return this.type == BaseType.Reference;
    }

    public boolean isVoid() {
        return this.type == BaseType.Void;
    }

    public boolean isLongOrDouble() {
        return this.type == BaseType.Double || this.type == BaseType.Long;
    }
    
    public boolean isObject() {
        return classInfo.getClassName().equals("java/lang/Object");
    }

    /**
     * Returns whether this field descriptor describes a primitive. This method
     * returns true for void.
     *
     * @return true <pre><=></pre> this field descriptor describes a primitive
     * (including void)
     */
    public boolean isPrimitive() {
        switch (this.type) {
            case Reference:
            case Array:
                return false;
            default:
                return true;
        }
    }

    /**
     * Returns the ClassInfo object for this field.
     *
     * @return class info for this field
     */
    public ClassInfo getClassInfo() {
        return classInfo;
    }
    /**
     * 
     * @return 
     */
    public FieldDescriptor getLeafElementType() {
        if (isArray()) {
            String val = getValue();
            int i1 = val.lastIndexOf("[") + 1;
            int i2 = val.lastIndexOf(";");
            if (val.charAt(i1) == 'L') {
                return FieldDescriptor.fromString(val.substring(i1 + 1, i2));
            } else {
                try {
                    return FieldDescriptor.fromBaseType(BaseType.parse(val.charAt(i1)));
                } catch (ByteCodeException ex) {
                    throw new Error(ex);
                }
            }
        } else {
            throw new IllegalStateException("not an array");
        }
    }
    
    public FieldDescriptor getElementType() {
        if (isArray()) {
            String val = getValue();
            int i2 = val.lastIndexOf(";");
            switch (val.charAt(1)) {
                case 'L':
                    return FieldDescriptor.fromString(val.substring(2, i2));
                case '[':
                    return FieldDescriptor.fromString(val.substring(1));
                default:
                    try {
                        return FieldDescriptor.fromBaseType(BaseType.parse(val.charAt(1)));
                    } catch (ByteCodeException ex) {
                        throw new Error(ex);
                    }
            }
        } else {
            throw new IllegalStateException("not an array");
        }
    }

    private static String parseReference(String utf8, int i) {
        StringBuilder b = new StringBuilder(utf8.length() - i);
        char next;
        for (; (next = utf8.charAt(i)) != ';'; i++) {
            if (next == '.') {
                return null;
            }
            b.append(next);
        }
        i++;
        if (i != utf8.length()) {
            return null;
        }
        return b.toString();
    }

    protected static FieldDescriptor parseFieldDescriptor(String utf8) {
        try {
            BaseType type = BaseType.parseNoThrow(utf8.charAt(0));
            if (type == null) {
                return null;
            }
            switch (type) {
                case Array:
                    int i;
                    for (i = 0; utf8.charAt(i) == '['; i++) {
                        if (i > 255) {
                            return null;
                        }
                    }
                    int dimension = i;
                    char next = utf8.charAt(i++);
                    if (next == 'L') {
                        String ref = parseReference(utf8, i);
                        if (ref != null) {
                            return new FieldDescriptor(new FieldDescriptor(ref), dimension);
                        } else {
                            return null;
                        }
                    } else {
                        BaseType arrayType = BaseType.parseNoThrow(next);
                        if (arrayType != null && i == utf8.length() && arrayType != BaseType.Void) {
                            return new FieldDescriptor(new FieldDescriptor(arrayType), dimension);
                        } else {
                            return null;
                        }
                    }
                case Reference:
                    String ref = parseReference(utf8, 1);
                    if (ref != null) {
                        return new FieldDescriptor(ref);
                    } else {
                        return null;
                    }
                default:
                    if (utf8.length() == 1) {
                        return new FieldDescriptor(type);
                    } else {
                        return null;
                    }
            }
        } catch (ByteCodeException | IndexOutOfBoundsException ex) {
            return null;
        }
    }
}
