package teajay.bytecode.cpool;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import teajay.bytecode.BaseType;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.UTFOutOfBoundsException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class MethodDescriptor extends Utf8Info {

    static class MParser {

        private int i = 0;
        private String utf8;
        private final StringBuilder buf = new StringBuilder(128);

        private MParser(String utf8) {
            this.utf8 = utf8;
        }

        MParser() {
        }

        private void reset(String input) {
            this.utf8 = input;
        }

        private MethodDescriptor parse() {
            this.i = 0;
            try {
                //Check necessary condition for this beeing a method descriptor
                if (utf8.charAt(i++) == '(') {
                    List<FieldDescriptor> params = new LinkedList<>();

                    while (utf8.charAt(i) != ')') { //parse argument list
                        FieldDescriptor descr = parseFieldDescriptor();
                        if (descr == null) {
                            return null;
                        }
                        params.add(descr);
                    }
                    i++; //skip )
                    //parse the return value
                    FieldDescriptor returnType = parseFieldDescriptor();
                    /*
                     * If there is still data in this string it cannot be a method
                     * descriptor.
                     */
                    if (returnType != null && i == utf8.length()) {
                        return new MethodDescriptor(returnType, params);
                    } else {
                        return null;
                    }
                } else {
                    return null;
                }
            } catch (IndexOutOfBoundsException ex) {
                //not a method descriptor
                return null;
            }
        }

        private FieldDescriptor parseFieldDescriptor() {
            try {
                BaseType type = BaseType.parseNoThrow(utf8.charAt(i++));
                if (type == null) {
                    return null;
                }
                switch (type) {
                    case Array:
                        int dimension = 1; //honor the first [
                        for (; utf8.charAt(i) == '['; i++) {
                            dimension++;
                            if (dimension > 255) {
                                return null;
                            }
                        }
                        char next = utf8.charAt(i++);
                        if (next == 'L') {
                            String ref = parseReference();
                            if (ref != null) {
                                return new FieldDescriptor(new FieldDescriptor(ref), dimension);
                            } else {
                                return null;
                            }
                        } else {
                            BaseType arrayType = BaseType.parseNoThrow(next);
                            if (arrayType != null && arrayType != BaseType.Void) {
                                return new FieldDescriptor(new FieldDescriptor(arrayType), dimension);
                            } else {
                                return null;
                            }
                        }
                    case Reference:
                        String ref = parseReference();
                        if (ref != null) {
                            return new FieldDescriptor(ref);
                        } else {
                            return null;
                        }
                    default:
                        return new FieldDescriptor(type);
                }
            } catch (ByteCodeException | IndexOutOfBoundsException ex) {
                return null;
            }
        }

        private String parseReference() throws ByteCodeException {
            buf.setLength(0);
            char next;
            for (; (next = utf8.charAt(i)) != ';'; i++) {
                if (next == '.') {
                    return null;
                }
                buf.append(next);
            }
            i++; //skip ;
            return buf.toString();
        }
    }
    private final FieldDescriptor returnType;
    private final List<FieldDescriptor> parameters;

    public MethodDescriptor(FieldDescriptor returnType, FieldDescriptor... parameters) throws UTFOutOfBoundsException {
        super(buildInternalString(returnType, parameters));
        this.returnType = returnType;
        this.parameters = Collections.unmodifiableList(Arrays.asList(parameters.clone()));
    }

    public MethodDescriptor(FieldDescriptor returnType, List<FieldDescriptor> parameters) throws UTFOutOfBoundsException {
        super(buildInternalString(returnType, parameters));
        this.returnType = returnType;
        this.parameters = Collections.unmodifiableList(new ArrayList<>(parameters));
    }

    private static String buildInternalString(FieldDescriptor returnType, List<FieldDescriptor> parameters) {
        StringBuilder buf = new StringBuilder();
        buf.append('(');
        for (FieldDescriptor p : parameters) {
            buf.append(p);
        }
        buf.append(')');
        buf.append(returnType);
        return buf.toString();
    }

    private static String buildInternalString(FieldDescriptor returnType, FieldDescriptor[] parameters) {
        StringBuilder buf = new StringBuilder();
        buf.append('(');
        for (FieldDescriptor p : parameters) {
            buf.append(p);
        }
        buf.append(')');
        buf.append(returnType);
        return buf.toString();
    }

    public FieldDescriptor getReturnType() {
        return this.returnType;
    }

    public List<FieldDescriptor> getParameters() {
        return parameters;
    }

    public int getParameterCount() {
        return parameters.size();
    }

    @Override
    public String toString() {
        return this.getValue();
    }

    static MethodDescriptor parseMethodDescriptor(String utf8) {
        return new MParser(utf8).parse();
    }

    static MethodDescriptor parseMethodDescriptor(String utf8, MParser parser) {
        parser.reset(utf8);
        return parser.parse();
    }
}
