package teajay.bytecode.methods;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.*;
import teajay.bytecode.AccessFlagsIO;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.U2;
import teajay.bytecode.VisibilityAccessFlags;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.attributes.AttributesReader;
import teajay.bytecode.attributes.code.CodeAttribute;
import teajay.bytecode.attributes.code.FullFrame;
import teajay.bytecode.attributes.code.VerificationType;
import teajay.bytecode.cpool.*;
import teajay.bytecode.exceptions.ByteCodeException;
import static teajay.bytecode.util.Tools.ccpt;
import static teajay.bytecode.util.Tools.checkUnqualifiedMethodName;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class MethodInfo implements ByteCodeWritable, ConstantPoolAddable {

    private final Collection<MethodAccessFlag> accessFlags;
    private final Utf8Info name;
    private final MethodDescriptor descriptor;
    private final Map<String, AttributeInfo> attributes;
    private final CodeAttribute code;
    private final ClassInfo myClass;

    /**
     * Create a new MethodInfo object. Not supplying any of the access flags
     * PRIVATE, PUBLIC or PROTECTED will make the method package private.
     *
     * @param name method name
     * @param descriptor method descriptor
     * @param accessFlags list of access flags for the method
     */
    private MethodInfo(Utf8Info name, MethodDescriptor descriptor, ClassInfo myClass, MethodAccessFlag... accessFlags) throws ByteCodeException {
        this(name, descriptor, myClass, Arrays.asList(accessFlags), false);
    }
    
    /**
     * Create a new MethodInfo object. Not supplying any of the access flags
     * PRIVATE, PUBLIC or PROTECTED will make the method package private.
     *
     * @param name method name
     * @param descriptor method descriptor
     * @param accessFlags list of access flags for the method
     */
    private MethodInfo(Utf8Info name, MethodDescriptor descriptor, ClassInfo myClass, Collection<MethodAccessFlag> accessFlags, boolean noCode) throws ByteCodeException {
        this.name = name;
        /*
         * The Oracle JVM does not enforce method names to be unqualified names.
         * (but the spec. does)
         */
        if (!checkUnqualifiedMethodName(name)) {
            throw new ByteCodeException("MethodInfo: not a unqualified method name " + name.getValue());
        }
        this.descriptor = descriptor;
        this.attributes = new HashMap<>();
        accessFlags = VisibilityAccessFlags.handlePackagePrivate(accessFlags, MethodAccessFlag.class);
        this.accessFlags = Collections.unmodifiableCollection(accessFlags);
        this.myClass = myClass;
        if (!noCode && !accessFlags.contains(MethodAccessFlag.ABSTRACT) && !accessFlags.contains(MethodAccessFlag.NATIVE)) {
            this.code = new CodeAttribute(makeFirstFrame(name.getValue(), descriptor, myClass, accessFlags.contains(MethodAccessFlag.STATIC)));
            this.addAttribute(code);
        } else {
            this.code = null;
        }
    }

    private static FullFrame makeFirstFrame(String name, MethodDescriptor descriptor, ClassInfo myClass, boolean isStatic) throws ByteCodeException {
        FullFrame firstFrame = new FullFrame();
        if (!isStatic) {
            if (name.equals("<init>")) {
                firstFrame.addLocal(VerificationType.makeUninitializedThis());
            } else {
                firstFrame.addLocal(VerificationType.makeObject(myClass));
            }
        }
        List<FieldDescriptor> params = descriptor.getParameters();
        firstFrame.addLocalTypes(params);
        return firstFrame;
    }

    /**
     * Create a new MethodInfo object. Not supplying any of the access flags
     * PRIVATE, PUBLIC or PROTECTED will make the method package private.
     *
     * @param name method name
     * @param descriptor method descriptor
     * @param myClass
     * @param accessFlags list of access flags for the method
     * @throws teajay.bytecode.exceptions.ByteCodeException
     */
    public MethodInfo(String name, MethodDescriptor descriptor, ClassInfo myClass, MethodAccessFlag... accessFlags) throws ByteCodeException {
        this(new Utf8Info(name), descriptor, myClass, accessFlags);
    }
    
    public MethodInfo(String name, MethodDescriptor descriptor, ClassInfo myClass, Collection<MethodAccessFlag> accessFlags) throws ByteCodeException {
        this(new Utf8Info(name), descriptor, myClass, accessFlags, false);
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        name.add(cpool);
        descriptor.add(cpool);
        for (AttributeInfo info : attributes.values()) {
            info.add(cpool);
        }
    }

    public void addAttribute(AttributeInfo info) {
        attributes.put(info.getName().getValue(), info);
    }

    public AttributeInfo getAttribute(String name) {
        return attributes.get(name);
    }

    public void removeAttribute(String name) {
        if (name.equals(CodeAttribute.NAME) && attributes.containsKey(CodeAttribute.NAME)) {
            throw new IllegalArgumentException("Cannot remove code attribute");
        } else {
            attributes.remove(name);
        }
    }

    public CodeAttribute getCode() {
        return code;
    }

    public Collection<AttributeInfo> getAttributes() {
        return Collections.unmodifiableCollection(attributes.values());
    }

    public Collection<MethodAccessFlag> getAccessFlags() {
        return accessFlags;
    }

    public Utf8Info getName() {
        return name;
    }

    public ClassInfo getDefiningClass() {
        return myClass;
    }

    public MethodDescriptor getDescriptor() {
        return descriptor;
    }

    public boolean isStatic() {
        return accessFlags.contains(MethodAccessFlag.STATIC);
    }

    public boolean isAbstract() {
        return accessFlags.contains(MethodAccessFlag.ABSTRACT);
    }

    public boolean isPublic() {
        return accessFlags.contains(MethodAccessFlag.PUBLIC);
    }

    public boolean isPrivate() {
        return accessFlags.contains(MethodAccessFlag.PRIVATE);
    }

    public boolean isProtected() {
        return accessFlags.contains(MethodAccessFlag.PROTECTED);
    }

    public boolean isPackagePrivate() {
        return accessFlags.contains(MethodAccessFlag.PACKAGE_PRIVATE);
    }

    public boolean isFinal() {
        return accessFlags.contains(MethodAccessFlag.FINAL);
    }

    public boolean isSynthetic() {
        return accessFlags.contains(MethodAccessFlag.SYNTHETIC);
    }

    public boolean isVarArgs() {
        return accessFlags.contains(MethodAccessFlag.VARARGS);
    }

    public boolean isNative() {
        return accessFlags.contains(MethodAccessFlag.NATIVE);
    }

    public boolean isSynchronized() {
        return accessFlags.contains(MethodAccessFlag.SYNCHRONIZED);
    }

    public boolean isStrict() {
        return accessFlags.contains(MethodAccessFlag.STRICT);
    }

    public boolean isBridge() {
        return accessFlags.contains(MethodAccessFlag.BRIDGE);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj instanceof MethodInfo) {
            MethodInfo info = (MethodInfo) obj;
            
            return info.descriptor.equals(descriptor) && info.name.equals(name);
        } else {
            return false;
        }
    }

    @Override
    public int hashCode() {
        int hash = 3;
        hash = 53 * hash + Objects.hashCode(this.name);
        hash = 53 * hash + Objects.hashCode(this.descriptor);
        return hash;
    }

    @Override
    public String toString() {
        return name.getValue() + " " + descriptor.getValue();
    }

    @Override
    public final void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        AccessFlagsIO.write(out, accessFlags);
        name.getCIndex(cpool).write(out, cpool);
        descriptor.getCIndex(cpool).write(out, cpool);
        new U2(attributes.size()).write(out, cpool);
        for (AttributeInfo attr : attributes.values()) {
            attr.write(out, cpool);
        }
        if (code != null && code.getInstructionList().size() <= 0) {
            throw new ByteCodeException("empty instruction list in method " + myClass.getClassName().replace('/', '.') + "." + name.getValue() + descriptor);
        }
    }

    public static MethodInfo read(DataInputStream in, ConstantPool cpool, int major, int minor, ClassInfo className) throws ByteCodeException, IOException {
        Collection<MethodAccessFlag> flags = AccessFlagsIO.parseMethodAccessFlags(in.readUnsignedShort());
        Utf8Info name = ccpt(cpool.get(in.readUnsignedShort()), Utf8Info.class, cpool);
        MethodDescriptor descriptor = ccpt(cpool.get(in.readUnsignedShort()), MethodDescriptor.class, cpool);
        MethodInfo method = new MethodInfo(name, descriptor, className, flags, true);

        int attributesCount = in.readUnsignedShort();
        
        for (int i = 0; i < attributesCount; i++) {
            AttributeInfo info = AttributesReader.read(in, cpool, major, minor);
            if (info != null) {
                if (info.getName().getValue().equals("Code")) {
                    //TODO: Special handling for the code attribute and StackMapFrame
                    method.addAttribute(info);
                } else {
                    method.addAttribute(info);
                }
            }
        }
        return method;
    }

    public MethodRefInfo makeMethodRef() {
        return new MethodRefInfo(myClass, new NameAndTypeInfo<>(name, descriptor));
    }
    
    public MethodRefInfo makeMethodRef(ClassInfo newClass) {
        return new MethodRefInfo(newClass, new NameAndTypeInfo<>(name, descriptor));
    }

    
    public MethodInfo makeFresh(ClassInfo newClass, Collection<MethodAccessFlag> newFlags) throws ByteCodeException {
        return new MethodInfo(name, descriptor, newClass, newFlags, false);
    }
}
