package teajay.bytecode.util;

import java.util.regex.Matcher;
import java.util.regex.Pattern;
import teajay.bytecode.cpool.CPoolEntry;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.U2Index;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class Tools {

    private final static Pattern unqualifiedName = Pattern.compile("[^\\.^;^\\[^/]*");
    private final static Pattern unqualifiedMethodName = Pattern.compile("[^\\.^;^\\[^/^<^>]*");

    public static boolean checkUnqualifiedName(Utf8Info name) {
        return checkUnqualifiedName(name.getValue());
    }
    
    public static boolean checkUnqualifiedClassName(Utf8Info name) {
        return checkUnqualifiedClassName(name.getValue());
    }
    
    public static boolean checkUnqualifiedClassName(String name) {
        if (name.length() == 0) {
            return true;
        }
        String[] idents = name.split("/");
        Matcher m = unqualifiedName.matcher(idents[0]);
        if (!m.matches()) {
            return false;
        }
        for (int i = 1; i < idents.length; i++) {
            m.reset(idents[i]);
            if (!m.matches()) {
                return false;
            }
        }
        return true;
    }

    public static boolean checkUnqualifiedName(String name) {
        Matcher m = unqualifiedName.matcher(name);
        return m.matches();
    }
    
    public static boolean checkUnqualifiedMethodName(Utf8Info name) {
        String val = name.getValue();
        if (!val.equals("<init>") && !val.equals("<clinit>")) {
            return checkUnqualifiedMethodName(val);
        } else {
            return true;
        }
    }
    
    public static boolean checkUnqualifiedMethodName(String name) {
        Matcher m = unqualifiedMethodName.matcher(name);
        return m.matches();
    }

    public static boolean checkBinaryName(String name) {
        return !name.contains(".");
    }

    public static String toBinaryName(String name) throws ByteCodeException {
        if (name.contains("/")) {
            throw new ByteCodeException("Illegal character \"/\" in class name: " + name);
        } else {
            return name.replace('.', '/');
        }
    }

    public static boolean contains(Object[] arr, Object obj) {
        for (Object item : arr) {
            if (item.equals(obj)) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * This method checks for <tt>entry</tt> to be of class <tt>check</tt> and
     * throws a ByteCodeException if that is not the case.
     *
     * @param <T> the type to check for
     * @param entry the entry to be checked
     * @param check the class the entry should have
     * @param cpool the value of cpool
     * @return <tt>entry</tt> itself
     * @throws ByteCodeException if the check fails
     */
    @SuppressWarnings("unchecked")
    public static <T extends CPoolEntry> T ccpt(CPoolEntry entry, Class<T> check, ConstantPool cpool) throws ByteCodeException {
        if (check.isInstance(entry)) {
            return (T) entry;
        }
        U2Index cIndex = entry.getCIndex(cpool);
        if (cIndex != null) {
            throw new ByteCodeException("constant pool: expecting a " + check.getSimpleName() + " but found a " + entry.getClass().getSimpleName() + " at constant pool index " + cIndex.getValue());
        } else {
            throw new ByteCodeException("constant pool: expecting a " + check.getSimpleName() + " but found a " + entry.getClass().getSimpleName());
        }
    }
}
