package teajay.generics.writer;

import teajay.generics.parser.Type;
import teajay.generics.parser.Type.Visitor;

/**
 * Encodes a Type-Object into a Signature string.
 * @author Michael Kußmann <michi@outx.de>
 */
public final class SignatureEncoder implements Visitor<StringBuilder, StringBuilder> {
    
    private static final SignatureEncoder instance = new SignatureEncoder();
   
    private SignatureEncoder() {
    }

    public static String encode(Type t) {
        return t.accept(instance, new StringBuilder()).toString();
    }

    @Override
    public StringBuilder visitSimpleType(Type.SimpleType type, StringBuilder p) {
        p.append(type.internalName);
        return p;
    }

    @Override
    public StringBuilder visitArrayType(Type.ArrayType type, StringBuilder p) {
        p.append('[');
        type.elemType.accept(this, p);
        return p;
    }

    @Override
    public StringBuilder visitMethodType(Type.MethodType type, StringBuilder p) {
        if (type.typeParamTypes != null && type.typeParamTypes.size() > 0) {
            p.append('<');
            for (Type t : type.typeParamTypes) {
                t.accept(this, p);
            }    
            p.append('>');
        }
        p.append('(');
        if (type.paramTypes != null) {
            for (Type t : type.paramTypes) {
                t.accept(this, p);
            }
        }
        p.append(')');
        type.returnType.accept(this, p);
        if (type.throwsTypes != null) {
            for (Type t : type.throwsTypes) {
                p.append('^');
                t.accept(this, p);
            }
        }
        return p;
    }

    @Override
    public StringBuilder visitClassSigType(Type.ClassSigType type, StringBuilder p) {
        p.append('<');
        if (type.typeParamTypes != null && type.typeParamTypes.size() > 0) {
            for (Type t : type.typeParamTypes) {
                t.accept(this, p);
            }
        }
        p.append('>');
        type.superclassType.accept(this, p);
        if (type.superinterfaceTypes != null && type.superinterfaceTypes.size() > 0) {
            for (Type t : type.superinterfaceTypes) {
                t.accept(this, p);
            }
        }
        return p;
    }

    @Override
    public StringBuilder visitClassType(Type.ClassType type, StringBuilder p) {
        if (type.outerType != null) {
            type.outerType.accept(this, p);
            p.deleteCharAt(p.length() - 1);
            p.append('.');
        } else {
            p.append('L');
        }
        p.append(type.name);
        if (type.typeArgs != null && type.typeArgs.size() > 0) {
            p.append('<');
            for (Type t : type.typeArgs) {
                t.accept(this, p);
            }
            p.append('>');
        }
        p.append(';');
        return p;
    }

    @Override
    public StringBuilder visitTypeParamType(Type.TypeParamType type, StringBuilder p) {
        p.append(type.name);
        if (type.classBound != null) {
            p.append(':');
            type.classBound.accept(this, p);
        }
        if (type.interfaceBounds != null) {
            if (type.classBound == null) {
                p.append(':');
            }
            for (Type t : type.interfaceBounds) {
                p.append(':');
                t.accept(this, p);
            }
        }

        return p;
    }

    @Override
    public StringBuilder visitWildcardType(Type.WildcardType type, StringBuilder p) {
        switch (type.kind) {
            case EXTENDS:
                p.append('+');
                break;
            case SUPER:
                p.append('-');
                break;
            case UNBOUNDED:
                p.append('*');
                break;
            default:
                throw new AssertionError(type.kind);
        }
        if (type.boundType != null) {
            type.boundType.accept(this, p);
        }
        return p;
    }
}
