/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package lexer;

import exceptions.LexerException;
import java.io.*;
import java.nio.charset.Charset;
import java.util.LinkedList;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class LookAheadLexer implements LexerWrapper, Closeable {

    private final Lexer lex;
    private final LinkedList<Symbol> lookaheadBuffer = new LinkedList<>();
    private final LinkedList<Symbol> templateBuffer = new LinkedList<>();
    private int offset = -1;
    private int numTemplateRequests = 0;
    private Symbol currentToken;
    private Symbol markedToken;

    public LookAheadLexer(InputStream in) {
        this(in, null);
    }

    public LookAheadLexer(Reader reader) {
        this(reader, null);
    }

    public LookAheadLexer(InputStream in, String sourceFile) {
        this.lex = new Lexer(in);
        lex.setSourceFile(sourceFile);
    }

    public LookAheadLexer(Reader reader, String sourceFile) {
        this.lex = new Lexer(reader);
        lex.setSourceFile(sourceFile);
    }

    public LookAheadLexer(String file) throws FileNotFoundException {
        this.lex = new Lexer(new BufferedReader(new FileReader(file), 32768 + 32));
        lex.setSourceFile(file);
    }

    public LookAheadLexer(File file) throws FileNotFoundException {
        this.lex = new Lexer(new BufferedReader(new FileReader(file), 32768 + 32));
        lex.setSourceFile(file.getPath());
    }

    public LookAheadLexer(File file, Charset charset) throws FileNotFoundException {
        this.lex = new Lexer(new BufferedReader(new InputStreamReader(new FileInputStream(file), charset), 32768 + 32));
        lex.setSourceFile(file.getPath());
    }

    public LookAheadLexer(String file, Charset charset) throws FileNotFoundException {
        this.lex = new Lexer(new BufferedReader(new InputStreamReader(new FileInputStream(file), charset), 32768 + 32));
        lex.setSourceFile(file);
    }

    @Override
    public Symbol yylex() throws IOException, LexerException {
        if (!templateBuffer.isEmpty()) {
            return currentToken = templateBuffer.removeFirst();
        } else if (offset >= lookaheadBuffer.size()) {
            currentToken = lex.yylex();
            lookaheadBuffer.add(currentToken);
            offset++;
        } else if (offset > -1) {
            currentToken = lookaheadBuffer.get(offset++);
        } else if (!lookaheadBuffer.isEmpty()) {
            currentToken = lookaheadBuffer.removeFirst();
        } else {
            currentToken = lex.yylex();
        }
        if (checkTemplateState() && fillTemplateBuffer(currentToken)) {
            currentToken = templateBuffer.removeFirst();
        }
        return currentToken;
    }

    private boolean fillTemplateBuffer(Symbol symbol) {
        int sym = symbol.sym;
        int line = symbol.line;
        int col = symbol.column;
        if (sym == Sym.OP_RIGHT_SHIFT) {
            templateBuffer.add(new Symbol(Sym.OP_GREATER, line, col++));
            templateBuffer.add(new Symbol(Sym.OP_GREATER, line, col));
            return true;
        } else if (sym == Sym.OP_RIGHT_SHIFT_SET) {
            templateBuffer.add(new Symbol(Sym.OP_GREATER, line, col++));
            templateBuffer.add(new Symbol(Sym.OP_GREATER, line, col++));
            templateBuffer.add(new Symbol(Sym.OP_EQ, line, col));
            return true;
        } else if (sym == Sym.OP_RIGHT_RIGHT_SHIFT) {
            templateBuffer.add(new Symbol(Sym.OP_GREATER, line, col++));
            templateBuffer.add(new Symbol(Sym.OP_GREATER, line, col++));
            templateBuffer.add(new Symbol(Sym.OP_GREATER, line, col));
            return true;
        } else if (sym == Sym.OP_RIGHT_RIGHT_SHIFT_SET) {
            templateBuffer.add(new Symbol(Sym.OP_GREATER, line, col++));
            templateBuffer.add(new Symbol(Sym.OP_GREATER, line, col++));
            templateBuffer.add(new Symbol(Sym.OP_GREATER, line, col++));
            templateBuffer.add(new Symbol(Sym.OP_EQ, line, col));
            return true;
        } else if (sym == Sym.OP_GEQ) {
            templateBuffer.add(new Symbol(Sym.OP_GREATER, line, col++));
            templateBuffer.add(new Symbol(Sym.OP_EQ, line, col));
            return true;
        }
        return false;
    }

    @Override
    public void mark() {
        if (hasMarking()) {
            for (int i = 0; i < offset; i++) {
                lookaheadBuffer.removeFirst();
            }
        }
        this.markedToken = currentToken;
        this.offset = 0;
    }

    @Override
    public void unmark() {
        if (hasMarking()) {
            for (int i = 0; i < offset; i++) {
                lookaheadBuffer.removeFirst();
            }
            this.offset = -1;
            this.markedToken = null;
        } else {
            throw new IllegalStateException("call to unmark() without previous mark()");
        }
    }

    @Override
    public void reset() {
        if (!hasMarking()) {
            throw new IllegalStateException("reset without previous mark");
        }
        templateBuffer.clear();
        this.currentToken = markedToken;
        this.offset = -1;
    }

    @Override
    public boolean lookahead(int rule, int... rules) {
        return lookahead(currentToken, rule, rules);
    }

    @Override
    public boolean lookahead(Rule rule) {
        this.mark();
        boolean res = rule.checkRule(this);
        this.reset();
        return res;
    }

    @Override
    public boolean lookahead(Symbol symbol, int rule, int... rules) {
        if (symbol.sym == rule) {
            if (rules.length > 0) {
                this.mark();
                for (int i = 0; i < rules.length; i++) {
                    try {
                        Symbol s = this.yylex();
                        if (s.sym != rules[i]) {
                            this.reset();
                            return false;
                        }
                    } catch (IOException ex) {
                        this.reset();
                        return false;
                    }
                }
                this.reset();
            }
            return true;
        }
        return false;
    }

    @Override
    public void setTemplateState() {
        numTemplateRequests++;
    }

    @Override
    public void resetTemplateState() {
        if (numTemplateRequests > 0) {
            numTemplateRequests--;
            if (numTemplateRequests <= 0) {
                templateBuffer.clear();
            }
        } else {
            throw new IllegalStateException("resetTemplateState(): Not in template state");
        }
    }

    public boolean checkTemplateState() {
        return numTemplateRequests > 0;
    }

    public boolean hasMarking() {
        return offset > -1;
    }

    @Override
    public int yyline() {
        return currentToken.line;
    }

    @Override
    public int yycolumn() {
        return currentToken.column;
    }

    @Override
    public String yytext() {
        return currentToken.value;
    }

    @Override
    public Symbol getCurrentToken() {
        return currentToken;
    }

    @Override
    public void close() throws IOException {
        lex.yyclose();
    }

    @Override
    public String toString() {
        return "current token: " + currentToken;
    }
}
