/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package minicompiler;

import codegeneration.environment.Stage3Environment;
import exceptions.CompilerException;
import java.io.BufferedOutputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.Collection;
import java.util.Deque;
import java.util.LinkedList;
import java.util.Properties;
import teajay.bytecode.ClassFile;
import teajay.bytecode.attributes.bootstrap.BootstrapMethods;
import teajay.bytecode.attributes.code.CodeAttribute;
import teajay.bytecode.attributes.code.JVMInstructionList;
import teajay.bytecode.attributes.code.Return;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.methods.MethodInfo;
import teajay.util.QIDTools;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 *
 * @author Michael (Burzan | Kußmann)
 */
public class CompileManager {

    private static CompileManager instance = null;
    private static final String FILE_SEP = System.getProperty("file.separator");
    private static final boolean isNotUnixStyleFileSeparator = !FILE_SEP.equals("/");
    private Deque<File> stage = new LinkedList<>();
    private File currentFile;

    public static CompileManager instance() {
        if (instance == null) {
            instance = new CompileManager();
        }
        return instance;
    }
    private String fileName;

    public void addSourceFile(final File s) throws CompilerException {
        if (s.getName().toLowerCase().endsWith(".tj")) {
            if (!s.canRead()) {
                throw new CompilerException("file not found: " + s);
            }
            stage.push(s);
        } else {
            throw new CompilerException(s + ": not a TeaJay source file");
        }
    }

    public File popFile() {
        if (!this.stage.isEmpty()) {
            return this.currentFile = this.stage.pop();
        }
        return null;
    }

    public boolean hasFiles() {
        return !this.stage.isEmpty();
    }

    public void writeClassFile(ClassFile classFile, Stage3Environment stage3Env) throws CompilerException {
        //XXX fix constructorless types
        Collection<MethodInfo> inits = classFile.getMethods("$init$");
        if (inits != null) {
            for (MethodInfo info : inits) {
                CodeAttribute code = info.getCode();
                if (code != null) {
                    JVMInstructionList lst = code.getInstructionList();
                    if (lst.size() == 0 && info.getDescriptor().getReturnType().isVoid()) {
                        try {
                            lst.add(new Return());
                        } catch (ByteCodeException ex) {
                            ex.printStackTrace(System.err);
                            System.exit(-1);
                        }
                    }
                }
            }
        }
        //XXX
        //Fix: Java 8 does not like empty BootstrapMethods attributes
        BootstrapMethods ms = (BootstrapMethods) classFile.getAttribute(BootstrapMethods.NAME);
        if (ms != null && ms.getBootstrapMethods().isEmpty()) {
            classFile.removeAttribute(BootstrapMethods.NAME);
        }
        //END Fix Java8
        DataOutputStream out = null;
        File outFile = null;
        try {
            File outDir = new File(outputDirectory());
            String packageName = QIDTools.packageName(classFile.getThis().getClassName());
            if (packageName != null && isNotUnixStyleFileSeparator) {
                packageName = packageName.replace("/", FILE_SEP);
            }
            outDir = new File(outDir, packageName != null ? packageName : "");
            if (!outDir.exists() && !outDir.mkdirs()) {
                throw new IOException("unable to create output directory: " + outDir);
            }
            String simpleName = classFile.getThis().getSimpleName();
            outFile = new File(outDir, simpleName + ".class");
            out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(outFile)));
            classFile.write(out);
            TeaJayAttribute attr = (TeaJayAttribute) classFile.getAttribute(TeaJayAttribute.NAME);
            if (attr != null) {
                if (attr.getType() == TeaJayAttribute.Type.Impl) {
                    Properties p = new Properties();
                    p.setProperty(classFile.getSuper().getClassName().replace('/', '.'), simpleName);

                    try (FileWriter wr = new FileWriter(new File(outDir, simpleName + ".tjt"))) {
                        p.store(wr, "TeaJay type file");
                    }
                }
            }
        } catch (IOException ex) {
            System.err.println("I/O error while writing class file: " + ex.getMessage());
            System.exit(1);
        } catch (ByteCodeException ex) {
            System.err.println("Fatal code generation in file " + outFile + " error: " + ex.getMessage());
            ex.printStackTrace(System.err);
            System.exit(-1);
        } finally {
            if (out != null) {
                try {
                    out.close();
                } catch (IOException ex) {
                }
            }
        }
    }

    public String outputDirectory() {
        String dir = System.getProperty("tjc_d");
        if (dir == null) {
            return ".";
        }
        return dir;
    }

    public String classPath() {
        String cp = System.getProperty("tjc_cp");
        if (cp == null) {
            return ".";
        }
        return cp;
    }

    public String sourcePath() {
        String sp = System.getProperty("tjc_sourcepath");
        if (sp == null) {
            return ".";
        }
        return sp;
    }

    public Charset encoding() {
        String enc = System.getProperty("tjc_encoding");
        if (enc == null) {
            return Charset.defaultCharset();
        }
        return Charset.forName(enc);
    }

    private CompileManager() {
    }

    public File getCurrentFile() {
        if (currentFile == null) {
            throw new IllegalStateException("there is no current file yet");
        }
        return this.currentFile;
    }

    public String getCurrentFilename() {
        return this.fileName;
    }

    public void setCurrentFilename(String fielName) {
        this.fileName = fielName;
    }
}
