/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package stage1;

import abstractsyntaxtree.BasicTypeNode;
import abstractsyntaxtree.ClosureTypeDecNode;
import abstractsyntaxtree.EdgeClampsOpenCloseHelperNode;
import abstractsyntaxtree.IdentifierNode;
import abstractsyntaxtree.IdentifierSuffixNode;
import abstractsyntaxtree.MethodeDeclaratorRestNode;
import abstractsyntaxtree.QualifiedIdentifierListNode;
import abstractsyntaxtree.QualifiedIdentifierNode;
import abstractsyntaxtree.ReferenceTypeNode;
import abstractsyntaxtree.SyntaxNode;
import codegeneration.Visitor;
import codegeneration.abstractvisitors.AbstractTypeVisitor;
import codegeneration.environment.Stage1Environment;
import exceptions.CompilerException;
import java.io.IOException;
import java.util.List;
import minicompiler.WorkingNode;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.runtime.exceptions.ResolveException;
import teajay.util.io.ClassFinder;
import teajay.util.io.Input;
import teajay.util.io.TJSourceFinder;

/**
 *
 * @author Michael Burzan
 * Visit the TypeNodes in AST. If the type not exists he added him to the Compiler
 */
class Stage1TypeVisitor extends AbstractTypeVisitor {

    private final Stage1Environment stage1Env;
    private final List<String> identifierList;

    public Stage1TypeVisitor(Stage1Environment stage1Env, List<String> identifierList) {
        this.stage1Env = stage1Env;
        this.identifierList = identifierList;
    }
    
    //Find types that do not occur as TypeNodes
    @Override
    public void begin(IdentifierSuffixNode node) throws CompilerException {
        identifierList.add(node.getIdentifier());
        List<String> tmp = node.getIdentifierList();
        if (tmp != null && tmp.size() > 1) {
            identifierList.addAll(tmp.subList(1, tmp.size()));
        }
        process(node, false);
    }

    @Override
    public void begin(BasicTypeNode node) throws CompilerException {
    }

    @Override
    public void visit(BasicTypeNode node) throws CompilerException {
    }

    @Override
    public void end(BasicTypeNode node) throws CompilerException {
    }

    @Override
    public void begin(ReferenceTypeNode node) throws CompilerException {
        this.identifierList.add(node.getName());
    }

    @Override
    public Visitor visit(ReferenceTypeNode node) throws CompilerException {
        return new Stage1ReferenceTypeVisitor(this.stage1Env, this.identifierList);
    }

    @Override
    public void end(ReferenceTypeNode node) throws CompilerException {
        process(node, true);
    }
    //Fix that the throws list is ignored in type resolving
    @Override
    public Visitor visit(QualifiedIdentifierListNode node) throws CompilerException {
        return new Visitor() {
            @Override
            public Visitor visit(QualifiedIdentifierNode node) {
                for (IdentifierNode id : node.getList()) {
                    Stage1TypeVisitor.this.identifierList.add(id.getName());
                }
                return null;
            }
        };
    }
    @Override
    public void end(QualifiedIdentifierListNode node) throws CompilerException {
        process(node, false);
    }
    
    private void process(SyntaxNode node, boolean error) throws CompilerException {
        //Make Identifier.Identifier  to Identifier/Identifier 
        try {
            StringBuilder identifier = new StringBuilder();
            for (int i = 0; i < identifierList.size(); i++) {
                identifier.append(identifierList.get(i));
                if (i != (identifierList.size() - 1)) {
                    identifier.append('/');
                }
            }
            String ident = identifier.toString();
            //identifier is not a generic parameter 
            if (!stage1Env.isGenericParameter(ident)) {
                //get the full qualified id from identifier
                if (stage1Env.getFacade().getFullQualifiedId(ident) == null) {
                    //if not exist, check exist path to identifier
                    if (ClassFinder.instance().find(ident) == null) {
                        String packageName = stage1Env.getPackageName();
                        if (packageName == null) {
                            packageName = "";
                        } else {
                            packageName += "/";
                        }
                        Input find = TJSourceFinder.instance().find(packageName + ident);
                        if (find != null) {
                            stage1Env.getFacade().addNameFromImportNoCheck(packageName + ident);
                            stage1Env.addWorkingNode(new WorkingNode(stage1Env.getFacade(), packageName + ident, find));
                        } else {
                            find = TJSourceFinder.instance().find(ident);
                            if (find != null) {
                                stage1Env.getFacade().addNameFromImportNoCheck(ident);
                                stage1Env.addWorkingNode(new WorkingNode(stage1Env.getFacade(), ident, find));
                            } else if (error) {
                                throw new CompilerException("Type " + ident + " not found", node.getSymbol());
                            }
                        }
                    } else {
                        stage1Env.getFacade().addNameFromImportNoCheck(ident);
                        Input find = TJSourceFinder.instance().find(ident);
                        if (find != null) {
                            stage1Env.addWorkingNode(new WorkingNode(stage1Env.getFacade(), ident, find));
                        }
                    }
                } else {
                    String qid = stage1Env.getFacade().getFullQualifiedId(ident);
                    Input find = TJSourceFinder.instance().find(qid);
                    if (find != null) {
                        stage1Env.addWorkingNode(new WorkingNode(stage1Env.getFacade(), qid, find));
                    }
                }
            }
            identifierList.clear();
        } catch (ByteCodeException | ResolveException | IOException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
    }

    @Override
    public void begin(EdgeClampsOpenCloseHelperNode node) throws CompilerException {
    }

    @Override
    public void visit(EdgeClampsOpenCloseHelperNode node) throws CompilerException {
    }

    @Override
    public void end(EdgeClampsOpenCloseHelperNode node) throws CompilerException {
    }

    @Override
    public void begin(ClosureTypeDecNode node) throws CompilerException {
    }

    @Override
    public Visitor visit(ClosureTypeDecNode node) throws CompilerException {
        return null;
    }

    @Override
    public void end(ClosureTypeDecNode node) throws CompilerException {
    }

    @Override
    public String getTypeName() throws CompilerException {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }
}
