/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package stage2;

import abstractsyntaxtree.EnumDeclarationNode;
import abstractsyntaxtree.ModifierNode;
import abstractsyntaxtree.NormalClassDeclarationNode;
import codegeneration.Visitor;
import codegeneration.abstractvisitors.AbstractClassDeclarationVisitor;
import codegeneration.environment.Stage2Environment;
import exceptions.CompilerException;
import java.util.List;
import teajay.bytecode.BaseType;
import teajay.bytecode.ClassFile;
import teajay.bytecode.attributes.code.AAStore;
import teajay.bytecode.attributes.code.ALoad;
import teajay.bytecode.attributes.code.ANewArray;
import teajay.bytecode.attributes.code.CheckCast;
import teajay.bytecode.attributes.code.Dup;
import teajay.bytecode.attributes.code.GetStatic;
import teajay.bytecode.attributes.code.InvokeSpecial;
import teajay.bytecode.attributes.code.InvokeStatic;
import teajay.bytecode.attributes.code.InvokeVirtual;
import teajay.bytecode.attributes.code.JVMInstructionList;
import teajay.bytecode.attributes.code.Ldc;
import teajay.bytecode.attributes.code.New;
import teajay.bytecode.attributes.code.PutStatic;
import teajay.bytecode.attributes.code.Return;
import teajay.bytecode.attributes.code.ReturnType;
import teajay.bytecode.attributes.signature.Signature;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.cpool.FieldRefInfo;
import teajay.bytecode.cpool.MethodDescriptor;
import teajay.bytecode.cpool.MethodRefInfo;
import teajay.bytecode.cpool.NameAndTypeInfo;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.fields.FieldAccessFlag;
import teajay.bytecode.fields.FieldInfo;
import teajay.bytecode.methods.MethodAccessFlag;
import teajay.bytecode.methods.MethodInfo;
import teajay.lang.TJComparable;

/**
 *
 * @author Michael Kußmann
 */
class Stage2ClassDeclarationVisitor extends AbstractClassDeclarationVisitor {

    private final Stage2Environment stage2Env;
    private final List<ModifierNode.Modifier> modifierList;
    private Stage2EnumDeclarationVisitor enumDeclVisitor;

    public Stage2ClassDeclarationVisitor(Stage2Environment stage2Env, List<ModifierNode.Modifier> modifierList) {
        this.stage2Env = stage2Env;
        this.modifierList = modifierList;
    }

    @Override
    public void begin(NormalClassDeclarationNode node) throws CompilerException {
    }

    @Override
    public Visitor visit(NormalClassDeclarationNode node) throws CompilerException {
        return new Stage2NormalClassDeclarationVisitor(stage2Env, modifierList, node.getClassName());
    }

    @Override
    public void end(NormalClassDeclarationNode node) throws CompilerException {
    }

    @Override
    public Object getClassContainer() throws CompilerException {
        return null;
    }

    @Override
    public void begin(EnumDeclarationNode node) throws CompilerException {
        try {
            stage2Env.getClassFile().addField(new FieldInfo("$VALUES", FieldDescriptor.makeArray(FieldDescriptor.fromClassFile(stage2Env.getClassFile()), 1),
                    FieldAccessFlag.PRIVATE, FieldAccessFlag.STATIC, FieldAccessFlag.FINAL, FieldAccessFlag.SYNTHETIC));
        } catch (ByteCodeException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
    }

    @Override
    public Visitor visit(EnumDeclarationNode node) throws CompilerException {
        return enumDeclVisitor = new Stage2EnumDeclarationVisitor(stage2Env, modifierList);
    }

    @Override
    public void end(EnumDeclarationNode node) throws CompilerException {
        try {
            ClassFile classFile = stage2Env.getClassFile();
            List<String> enumConstants = enumDeclVisitor.getContainer();
            for (String name : enumConstants) {
                classFile.addField(new FieldInfo(name, FieldDescriptor.fromClassFile(classFile),
                        FieldAccessFlag.PUBLIC, FieldAccessFlag.STATIC, FieldAccessFlag.FINAL, FieldAccessFlag.ENUM));
            }
            //values()
            String enumName = stage2Env.getClassFile().getThis().getClassName();
            {
                MethodInfo values = new MethodInfo("values",
                        new MethodDescriptor(FieldDescriptor.makeArray(FieldDescriptor.fromClassFile(classFile), 1)),
                        classFile.getThis(), MethodAccessFlag.PUBLIC, MethodAccessFlag.STATIC);
                classFile.addMethod(values);
                
                JVMInstructionList lst = values.getCode().getInstructionList();
                FieldRefInfo ref = new FieldRefInfo(classFile.getThis(),
                        new NameAndTypeInfo<>("$VALUES", FieldDescriptor.makeArray(FieldDescriptor.fromClassFile(classFile), 1)));
                lst.add(new GetStatic(ref));
                ClassInfo array = FieldDescriptor.makeArray(FieldDescriptor.fromClassFile(classFile), 1).getClassInfo();
                MethodRefInfo mref = new MethodRefInfo(array, new NameAndTypeInfo<>("clone",
                        new MethodDescriptor(FieldDescriptor.fromClass(Object.class))));
                lst.add(new InvokeVirtual(mref));
                lst.add(new CheckCast(array));
                lst.add(new Return(ReturnType.Reference));
                // List<Enum> values() 
                MethodInfo listValues = new MethodInfo("getConstants",
                        new MethodDescriptor(FieldDescriptor.fromClass(List.class)),
                        classFile.getThis(), MethodAccessFlag.PUBLIC, MethodAccessFlag.STATIC);
                classFile.addMethod(listValues);
                
                listValues.addAttribute(new Signature("()Lteajay/lang/List<L"+enumName+";>;"));
                
                lst = listValues.getCode().getInstructionList();
                MethodRefInfo toList = MethodRefInfo.make(ClassInfo.fromClass(teajay.util.ArrayUtils.class), "toList",
                        FieldDescriptor.fromClass(teajay.lang.List.class), FieldDescriptor.fromClass(Object[].class));
                lst.add(new InvokeStatic(values.makeMethodRef()));
                lst.add(new InvokeStatic(toList));
                lst.add(new Return(ReturnType.Reference));
            } //end values(), getConstants()
            
            { //TJComparable
                MethodInfo tjComp = new MethodInfo("tjCompareTo",
                        new MethodDescriptor(FieldDescriptor.fromString("teajay/lang/Number"), FieldDescriptor.getObject()),
                        classFile.getThis(), MethodAccessFlag.PUBLIC);
                tjComp.addAttribute(new Signature("(L" + enumName + ";)Lteajay/lang/Number;"));
                classFile.addMethod(tjComp);
                JVMInstructionList lst = tjComp.getCode().getInstructionList();
                MethodRefInfo compareTo = MethodRefInfo.make(classFile.getThis(), "compareTo", FieldDescriptor.getInt(),
                        FieldDescriptor.fromClass(Enum.class));
                lst.add(new ALoad(0));
                lst.add(new ALoad(1));
                lst.add(new CheckCast(classFile.getThis()));
                lst.add(new InvokeVirtual(compareTo));
                lst.add(new InvokeStatic(MethodRefInfo.make("teajay/lang/Number", "$valueOf", "teajay/lang/Number", "int")));
                lst.add(new Return(ReturnType.Reference));
            }
            //end TJComparable

            //valueOf(String)
            {
                MethodInfo values = new MethodInfo("valueOf",
                        new MethodDescriptor(FieldDescriptor.fromClassFile(classFile), FieldDescriptor.fromClass(String.class)),
                        classFile.getThis(), MethodAccessFlag.PUBLIC, MethodAccessFlag.STATIC);
                classFile.addMethod(values);
                JVMInstructionList lst = values.getCode().getInstructionList();
                lst.add(new Ldc(classFile.getThis()));
                lst.add(new ALoad(0));
                MethodRefInfo mref = new MethodRefInfo(ClassInfo.fromClass(Enum.class),
                        new NameAndTypeInfo<>("valueOf", new MethodDescriptor(FieldDescriptor.fromClass(Enum.class),
                        FieldDescriptor.fromClass(Class.class), FieldDescriptor.fromClass(String.class))));
                lst.add(new InvokeStatic(mref));
                lst.add(new CheckCast(classFile.getThis()));
                lst.add(new Return(ReturnType.Reference));
            }
            //end valueOf(String)
            {
                MethodInfo clinit = new MethodInfo("<clinit>",
                        new MethodDescriptor(FieldDescriptor.getVoid()),
                        classFile.getThis(), MethodAccessFlag.STATIC);
                classFile.addMethod(clinit);
                JVMInstructionList lst = clinit.getCode().getInstructionList();

                //MethodDescriptor baseConstr = new MethodDescriptor(FieldDescriptor.getVoid(),
                //        FieldDescriptor.fromClass(String.class), FieldDescriptor.fromBaseType(BaseType.Int));
                //NameAndTypeInfo nameInfo = new NameAndTypeInfo("<init>", baseConstr);
                //MethodRefInfo init = new MethodRefInfo(this.classFile.getThis(), nameInfo);
                MethodRefInfo init = MethodRefInfo.make(classFile.getThis(), "<init>", FieldDescriptor.getVoid(),
                        FieldDescriptor.fromClass(String.class), FieldDescriptor.fromBaseType(BaseType.Int));

                int i = 0;
                for (String name : enumConstants) {
                    lst.add(new New(classFile.getThis()));
                    lst.add(new Dup());
                    lst.add(new Ldc(name));
                    lst.pushInt(i);
                    lst.add(new InvokeSpecial(init));
                    FieldRefInfo finfo = new FieldRefInfo(classFile.getThis(),
                            new NameAndTypeInfo<>(name, FieldDescriptor.fromClassFile(classFile)));
                    lst.add(new PutStatic(finfo));
                    i++;
                }
                lst.pushInt(i);
                lst.add(new ANewArray(classFile.getThis()));
                lst.add(new Dup());
                i = 0;
                for (String name : enumConstants) {
                    lst.pushInt(i);
                    FieldRefInfo finfo = new FieldRefInfo(classFile.getThis(),
                            new NameAndTypeInfo<>(name, FieldDescriptor.fromClassFile(classFile)));
                    lst.add(new GetStatic(finfo));
                    lst.add(new AAStore());
                    i++;
                    if (i < enumConstants.size()) {
                        lst.add(new Dup());
                    }
                }
                FieldRefInfo ref = new FieldRefInfo(classFile.getThis(),
                        new NameAndTypeInfo<>("$VALUES", FieldDescriptor.makeArray(FieldDescriptor.fromClassFile(classFile), 1)));
                lst.add(new PutStatic(ref));
                lst.add(new Return());
            }
            // genDefaultConstructor(true);
        } catch (ByteCodeException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }

    }
}
