/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package stage2;

import abstractsyntaxtree.ClassBodyNode;
import abstractsyntaxtree.SyntaxNode;
import abstractsyntaxtree.TypeNode;
import abstractsyntaxtree.TypeParametersNode;
import codegeneration.Visitor;
import codegeneration.abstractvisitors.AbstractTjImplDeclarationVisitor;
import codegeneration.environment.Stage2Environment;
import exceptions.CompilerException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import stage2.utils.ConstructorBuilder;
import teajay.bytecode.ClassFile;
import teajay.bytecode.attributes.code.ALoad;
import teajay.bytecode.attributes.code.CheckCast;
import teajay.bytecode.attributes.code.DLoad;
import teajay.bytecode.attributes.code.Dup;
import teajay.bytecode.attributes.code.FLoad;
import teajay.bytecode.attributes.code.GetField;
import teajay.bytecode.attributes.code.Goto;
import teajay.bytecode.attributes.code.ILoad;
import teajay.bytecode.attributes.code.IfNull;
import teajay.bytecode.attributes.code.InvokeSpecial;
import teajay.bytecode.attributes.code.InvokeVirtual;
import teajay.bytecode.attributes.code.JVMInstructionList;
import teajay.bytecode.attributes.code.LLoad;
import teajay.bytecode.attributes.code.Label;
import teajay.bytecode.attributes.code.NOP;
import teajay.bytecode.attributes.code.Pop;
import teajay.bytecode.attributes.code.Return;
import teajay.bytecode.attributes.code.ReturnType;
import teajay.bytecode.attributes.code.StackMapTable;
import teajay.bytecode.attributes.code.VerificationType;
import teajay.bytecode.attributes.signature.Signature;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.cpool.FieldRefInfo;
import teajay.bytecode.cpool.MethodDescriptor;
import teajay.bytecode.cpool.MethodRefInfo;
import teajay.bytecode.cpool.NameAndTypeInfo;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.methods.MethodAccessFlag;
import teajay.bytecode.methods.MethodInfo;
import teajay.generics.check.TypeException;
import teajay.generics.check.TypeUtils;
import teajay.generics.parser.SignatureParser;
import teajay.generics.parser.Type;
import teajay.generics.writer.SignatureEncoder;
import teajay.runtime.exceptions.ResolveException;
import teajay.util.io.ClassFileRef;
import teajay.util.io.ClassFinder;
import teajay.util.tjattribute.TJBridge;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 *
 * @author Michael (Burzan | Kußmann)
 */
class Stage2TjImplDeclarationVistor extends AbstractTjImplDeclarationVisitor {

    private final String implName;
    private final Stage2Environment stage2Env;
    private Stage2TypeVisitor stage2Implements;
    private Stage2TypeParametersVisitor stage2Generic;
    private Stage2TjImplClassBodyVisitor stage2tjImplBody;

    public Stage2TjImplDeclarationVistor(Stage2Environment stage2Env, String implName) {
        this.stage2Env = stage2Env;
        this.implName = implName;
    }

    @Override
    public void begin(TypeNode node) throws CompilerException {
    }

    @Override
    public void begin(TypeParametersNode node) throws CompilerException {
    }

    @Override
    public void begin(ClassBodyNode node) throws CompilerException {
        ClassFile classFile = this.stage2Env.getClassFile();
        Type.ClassSigType sigType;
        //get signature of generic 
        if (stage2Generic != null) {
            sigType = new Type.ClassSigType(stage2Generic.getTypeParams(), stage2Implements.getType(), null);
        } else {
            //Observe #1
            sigType = new Type.ClassSigType(null, stage2Implements.getType(), null);
        }
        try {
            try {
                //check generic signature from typedef and typeimpl 
                // exact match
                // the signature from typedef is detected in TypeUtils
                if (!TypeUtils.signatureEquals(sigType)) {
                    throw new CompilerException("generic signature missmatch", node.getSymbol());
                }
            } catch (TypeException ex) {
                throw new CompilerException(ex.getMessage(), node.getSymbol());
            }
            //add signature as attribute
            classFile.addAttribute(new Signature(SignatureEncoder.encode(sigType)));
            // set typedef as super class to typeimpl
            classFile.setSuper(((Type.ClassType) stage2Implements.getType()).name);
        } catch (ByteCodeException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
    }

    @Override
    public Visitor visit(TypeNode node) throws CompilerException {
        return stage2Implements = new Stage2TypeVisitor(stage2Env);
    }

    @Override
    public Visitor visit(TypeParametersNode node) throws CompilerException {
        stage2Env.addGenericFrame();
        stage2Env.getTypeUtils().addFrame();
        return stage2Generic = new Stage2TypeParametersVisitor(stage2Env);
    }

    @Override
    public Visitor visit(ClassBodyNode node) throws CompilerException {
        return stage2tjImplBody = new Stage2TjImplClassBodyVisitor(stage2Env);
    }

    @Override
    public void end(TypeNode node) throws CompilerException {
    }

    @Override
    public void end(TypeParametersNode node) throws CompilerException {
        node.setGeneric(this.stage2Generic.getTypeParams());
    }

    private void addBridgeMethods(ClassFile typedef, SyntaxNode node) throws CompilerException {
        try {
            Collection<List<MethodInfo>> tmpMethods = typedef.getMethods();

            Collection<List<MethodInfo>> methods = new ArrayList<>();
            methods.addAll(tmpMethods);
//            for (ClassInfo iface : typedef.getInterfaces()) {
//                ClassFileRef ref = ClassFinder.instance().find(iface.getClassName());
//                if (ref == null) {
//                    throw new CompilerException("Type " + iface.getClassName() + " not found", node.getSymbol());
//                }
//                Collection<List<MethodInfo>> ifaceMethods = ref.get().getMethods();
//                methods.addAll(ifaceMethods);
//            }

            for (List<MethodInfo> mTemp : methods) {
                List<MethodInfo> lInfo = new ArrayList<>();
                lInfo.addAll(mTemp);
                for (MethodInfo method : lInfo) {
                    if (!method.isBridge()) {
                        continue;
                    }
                    TJBridge bridgeAttribute = (TJBridge) method.getAttribute(TJBridge.NAME);
                    String signature = bridgeAttribute.getSignature();

                    Type.MethodType mType = (Type.MethodType) SignatureParser.parse(signature);

                    Collection<MethodAccessFlag> tmp = method.getAccessFlags();
                    Collection<MethodAccessFlag> accessFlags = new LinkedList<>();
                    accessFlags.addAll(tmp);
                    accessFlags.add(MethodAccessFlag.BRIDGE);
                    accessFlags.remove(MethodAccessFlag.ABSTRACT);
                    MethodInfo bridgeMethod = new MethodInfo(method.getName().getValue(),
                            method.getDescriptor(),
                            stage2Env.getClassFile().getThis(), accessFlags);
                    if (!stage2Env.getClassFile().hasMethod(bridgeMethod.getName().getValue(), bridgeMethod.getDescriptor())) {
                        stage2Env.getClassFile().addMethod(bridgeMethod);

                        JVMInstructionList ins = bridgeMethod.getCode().getInstructionList();

                        ins.add(new ALoad(0));
                        //ins.add(new GetField(FieldRefInfo.make(stage2Env.getClassFile().getThis().getClassName(),
                        //        "this", typedef.getThis().getClassName())));
                        List<FieldDescriptor> parameters = bridgeMethod.getDescriptor().getParameters();
                        for (int i = 0; i < parameters.size(); i++) {
                            ins.add(new ALoad(i + 1));
                        }
                        MethodDescriptor makeMethodDescriptor = stage2Env.getTypeUtils().makeMethodDescriptor(mType);
                        ins.add(new InvokeVirtual(new MethodRefInfo(stage2Env.getClassFile().getThis(),
                                new NameAndTypeInfo<>(bridgeMethod.getName().getValue(), makeMethodDescriptor))));
                        if (makeMethodDescriptor.getReturnType().isVoid()) {
                            ins.add(new Return());
                        } else {
                            ins.add(new Return(ReturnType.Reference));
                        }
                    }

                }


            }
        } catch (ByteCodeException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }

    }

    private void checkMethods(ClassFile typedef, boolean checkConstructor, SyntaxNode node) throws CompilerException {
        Collection<List<MethodInfo>> typeDefMethods = typedef.getMethods();
        ClassFile typeImpl = stage2Env.getClassFile();
        for (List<MethodInfo> list : typeDefMethods) {

            for (MethodInfo info : list) {
                if (info.getName().getValue().equals("<clinit>") /*|| info.isBridge()*/) {
                    break;
                }
                TeaJayAttribute attr = (TeaJayAttribute) info.getAttribute(TeaJayAttribute.NAME);

                if (info.isPrivate() || (attr != null && attr.getAccessType() == TeaJayAttribute.AccessType.Hidden)) {
                    continue;
                }
                if (info.getName().getValue().equals("<init>") && !checkConstructor) {
                    break;
                }
                if (info.getName().getValue().equals("<init>") && info.isProtected()) {
                    break;
                }
                List<MethodInfo> methods;
                //if (!info.getName().getValue().equals("<init>")) {
                    methods = typeImpl.getMethods(info.getName().getValue());
                //} else {
                //    methods = typeImpl.getMethods("$init$");
                //}

                if (methods == null) {
                    throw new CompilerException("Missing method " + info.getName().getValue() + " in typeimpl", node.getSymbol());
                }
                boolean exists = false;
                for (MethodInfo i : methods) {
                    MethodDescriptor descr;
                    if (i.getName().getValue().equals("<init>")) {
                        List<FieldDescriptor> newParams = new LinkedList<>();
                        newParams.addAll(i.getDescriptor().getParameters().subList(1, i.getDescriptor().getParameterCount()));
                        descr = new MethodDescriptor(i.getDescriptor().getReturnType(), newParams);
                    } else {
                        descr = i.getDescriptor();
                    }
                    if (descr.equals(info.getDescriptor())) {                     
                        exists = i.isStatic() == info.isStatic() && i.isPublic() == info.isPublic();
                        break;
                    }
                }
                if (!exists) {
                    throw new CompilerException("Missing method " + info + " in typeimpl", node.getSymbol());
                }
            }
        }
    }

    private static boolean containsPrimitivesExceptVoid(MethodDescriptor mdesc) {
        for (FieldDescriptor descr : mdesc.getParameters()) {
            if (descr.isPrimitive() && !descr.isVoid()) {
                return true;
            }
        }
        FieldDescriptor returnType = mdesc.getReturnType();
        if (returnType.isPrimitive() && !returnType.isVoid()) {
            return true;
        }
        return false;
    }

    private void writeStubs(ClassFile typedef, boolean onlyBridge, SyntaxNode node) throws CompilerException, ByteCodeException {
        writeStubs(typedef, onlyBridge, node, new HashSet<ClassInfo>());
    }

    private void writeStubs(ClassFile typedef, boolean onlyBridge, SyntaxNode node, Set<ClassInfo> visited) throws CompilerException, ByteCodeException {
        Collection<List<MethodInfo>> typeDefMethods = new ArrayList<>();
        typeDefMethods.addAll(typedef.getMethods());
        for (ClassInfo iface : typedef.getInterfaces()) {
            try {
                ClassFile get = ClassFinder.instance().find(iface.getClassName()).get();
                if (!visited.contains(iface)) {
                    visited.add(iface);
                    writeStubs(get, true, node, visited);
                }
            } catch (IOException | ResolveException ex) {
                ex.printStackTrace(System.err);
                System.exit(-1);
            }
        }
        ClassFile typeImpl = stage2Env.getClassFile();
        for (List<MethodInfo> list : typeDefMethods) {

            for (MethodInfo info : list) {
                if (info.getName().getValue().equals("<init>")) {
                    break;
                } else if (info.isStatic() || info.isFinal() || info.isPrivate() || info.isProtected() || info.isPackagePrivate()) {
                    continue;
                } else {
                }

                Collection<MethodInfo> methods = typeImpl.getMethods(info.getName().getValue());
                boolean exists = false;
                if (methods != null) {
                    for (MethodInfo minfo : methods) {
                        if (minfo.getDescriptor().equals(info.getDescriptor())) {
                            String str1;
                            String str2;
                            /*if (sig1 == null || sig2 == null) {
                             Type.MethodType m1 = TypeUtils.getMethodSignature(minfo);
                             Type.MethodType m2 = TypeUtils.getMethodSignature(info);
                             str1 = SignatureEncoder.encode(m1);
                             str2 = SignatureEncoder.encode(m2);
                             } else {
                             str1 = sig1.getSignature();
                             str2 = sig2.getSignature();
                             }*/
                            if (!minfo.isBridge() && !info.isBridge()) {
                                Type.MethodType methodSignatureByEnv1 = stage2Env.getTypeUtils().getMethodSignatureByEnv(minfo, stage2Env.getTypeUtils().createThis(stage2Env.getClassFile().getThis().getClassName()), null);

                                Type.MethodType methodSignatureByEnv2 = stage2Env.getTypeUtils().getMethodSignatureByEnv(info, stage2Env.getTypeUtils().createThis(stage2Env.getClassFile().getThis().getClassName()), null);
                                //Ignore throws list
                                methodSignatureByEnv1.throwsTypes.clear();
                                methodSignatureByEnv2.throwsTypes.clear();
                                
                                str1 = SignatureEncoder.encode(methodSignatureByEnv1);
                                str2 = SignatureEncoder.encode(methodSignatureByEnv2);

                                if (!str1.equals(str2)) {
                                    throw new CompilerException("Incompatible signatures:" + "METHODNAME:=" + minfo + " " + str1 + " != " + str2, node.getSymbol());
                                }
                            }
                            exists = true;
                            break;
                        }
                    }
                }
                if (!exists && !info.isFinal()) {
                    Collection<MethodAccessFlag> flags = new HashSet<>();
                    flags.addAll(info.getAccessFlags());
                    flags.add(MethodAccessFlag.SYNTHETIC);
                    flags.remove(MethodAccessFlag.ABSTRACT);
                    flags.remove(MethodAccessFlag.NATIVE);
                    MethodInfo impl = info.makeFresh(typeImpl.getThis(), flags);
                    impl.addAttribute(new Signature(SignatureEncoder.encode(TypeUtils.getMethodSignature(info))));

                    Type.MethodType methodSignatureByEnv = stage2Env.getTypeUtils().getMethodSignatureByEnv(info, stage2Env.getTypeUtils().createThis(typeImpl.getThis().getClassName()), null);
                    MethodDescriptor makeMethodDescriptor = stage2Env.getTypeUtils().makeMethodDescriptor(methodSignatureByEnv);
                    List<MethodInfo> possibleMethods = typeImpl.getMethods(info.getName().getValue());
                    boolean bridge = false;
                    MethodInfo bridgeCall = null;
                    if (possibleMethods != null) {
                        for (MethodInfo p : possibleMethods) {
                            if (p.getDescriptor().equals(makeMethodDescriptor)) {
                                bridge = true;
                                bridgeCall = p;
                                break;
                            }
                        }
                    }

                    if (bridge) {
                        flags.add(MethodAccessFlag.BRIDGE);
                        flags.remove(MethodAccessFlag.SYNTHETIC);
                        impl = info.makeFresh(typeImpl.getThis(), flags);
                    }
                    if ((!onlyBridge || (onlyBridge && bridge)) && !typeImpl.hasMethod(impl.getName().getValue(), impl.getDescriptor())) {
                        typeImpl.addMethod(impl);
                    }
                    JVMInstructionList lst = impl.getCode().getInstructionList();


                    lst.add(new ALoad(0));

                    //Case 1: Speziellerer Rückgabewert und gleiche Parameter Liste
                    //Case 2: Generic Kompatibilität prüfen auf Rückgabewert und Parameters. Nach erasure, exact Match

                    if (!bridge) {
                        ClassFileRef realTypedef = ClassFinder.instance().find(stage2Env.getClassFile().getSuper().getClassName());
                        if (realTypedef == null) {
                            throw new CompilerException("Class not found: " + stage2Env.getClassFile().getSuper().getClassName(), node.getSymbol());
                        }
                        try {
                            FieldRefInfo finfo = FieldRefInfo.make(typeImpl.getThis(), "super", FieldDescriptor.fromClassInfo(realTypedef.get().getSuper()));
                            Label jmpOver = new Label();
                            Label jmpOver2 = new Label();
                            lst.add(new GetField(finfo));
                            lst.add(new Dup());
                            lst.add(new IfNull(jmpOver));
                            {
                                MethodDescriptor descr = impl.getDescriptor();
                                List<FieldDescriptor> parameters = descr.getParameters();
                                writeArgs(parameters, makeMethodDescriptor, lst, bridge);
                                lst.add(new InvokeVirtual(bridgeCall != null ? bridgeCall.makeMethodRef() : info.makeMethodRef()));
                                lst.add(new Goto(jmpOver2));
                            }
                            StackMapTable stackMapTable = impl.getCode().getStackMapTable();
                            {
                                stackMapTable.addSameLocalsOneStackItemFrame(lst.add(jmpOver, new Pop())).addStack(VerificationType.valueOf(finfo.getNameAndType().getDescriptor()));
                                lst.add(new ALoad(0));
                                MethodDescriptor descr = impl.getDescriptor();
                                List<FieldDescriptor> parameters = descr.getParameters();
                                writeArgs(parameters, makeMethodDescriptor, lst, bridge);
                                lst.add(new InvokeSpecial(bridgeCall != null ? bridgeCall.makeMethodRef() : info.makeMethodRef()));
                            }
                            if (!impl.getDescriptor().getReturnType().isVoid()) {
                                stackMapTable.addSameLocalsOneStackItemFrame(lst.add(jmpOver2, new NOP())).addStack(VerificationType.valueOf(impl.getDescriptor().getReturnType()));
                            } else {
                                stackMapTable.addSameFrame(lst.add(jmpOver2, new NOP()));
                            }
                        } catch (IOException | ResolveException ex) {
                            ex.printStackTrace(System.err);
                            System.exit(-1);
                        }
                    } else {
                        MethodDescriptor descr = impl.getDescriptor();
                        List<FieldDescriptor> parameters = descr.getParameters();
                        writeArgs(parameters, makeMethodDescriptor, lst, bridge);
                        lst.add(new InvokeVirtual(bridgeCall != null ? bridgeCall.makeMethodRef() : info.makeMethodRef()));
                    }

                    if (info.getDescriptor().getReturnType().isVoid()) {
                        lst.add(new Return());
                    } else {
                        if (bridge) {
                            if (!makeMethodDescriptor.getReturnType().isPrimitive()) {
                                lst.add(new CheckCast(makeMethodDescriptor.getReturnType().getClassInfo()));
                            }
                        }
                        if (info.getDescriptor().getReturnType().isPrimitive()) {
                            switch (info.getDescriptor().getReturnType().getBaseType()) {
                                case Boolean:
                                case Byte:
                                case Short:
                                case Char:
                                case Int:
                                    lst.add(new Return(ReturnType.Integer));
                                    break;
                                case Float:
                                    lst.add(new Return(ReturnType.Float));
                                    break;
                                case Double:
                                    lst.add(new Return(ReturnType.Double));
                                    break;
                                case Long:
                                    lst.add(new Return(ReturnType.Long));
                                    break;
                                case Void:
                                case Reference:
                                case Array:
                                default:
                                    throw new AssertionError(info.getDescriptor().getReturnType().getBaseType().name());
                            }
                        } else {
                            lst.add(new Return(ReturnType.Reference));
                        }
                    }
                }
            }
        }
    }

    private static void writeArgs(List<FieldDescriptor> parameters, MethodDescriptor makeMethodDescriptor, JVMInstructionList lst, boolean bridge) throws ByteCodeException {
        int longOrDoubleCount = 0;
        for (int i = 1; i <= parameters.size(); i++) {
            if (parameters.get(i - 1).isPrimitive()) {
                switch (parameters.get(i - 1).getBaseType()) {
                    case Boolean:
                    case Byte:
                    case Short:
                    case Char:
                    case Int:
                        lst.add(new ILoad(i + longOrDoubleCount));
                        break;
                    case Float:
                        lst.add(new FLoad(i + longOrDoubleCount));
                        break;
                    case Double:
                        lst.add(new DLoad(i + longOrDoubleCount++));
                        break;
                    case Long:
                        lst.add(new LLoad(i + longOrDoubleCount++));
                        break;
                    case Reference:
                    case Array:
                    case Void:
                    default:
                        throw new AssertionError(parameters.get(i - 1).getBaseType().name());

                }
            } else {
                lst.add(new ALoad(i + longOrDoubleCount));
            }
            if (bridge) {
                FieldDescriptor param = makeMethodDescriptor.getParameters().get(i - 1);
                if (!param.isPrimitive()) {
                    lst.add(new CheckCast(param.getClassInfo()));
                }
            }
        }
    }

    private void interfaceCheck(ClassFile typedef, SyntaxNode node) throws ByteCodeException, CompilerException, IOException, ResolveException {
        for (ClassInfo iface : typedef.getInterfaces()) {
            ClassFile file = ClassFinder.instance().find(iface.getClassName()).get();
            checkMethods(file, false, node);
            for (ClassInfo info : file.getInterfaces()) {
                interfaceCheck(ClassFinder.instance().find(info.getClassName()).get(), node);
            }
        }
    }

    @Override
    public void end(ClassBodyNode node) throws CompilerException {
        try {
            ClassFile typeDefClass = ClassFinder.instance().find(stage2Env.getClassFile().getSuper().getClassName()).get();
            ClassFile typeImpl = stage2Env.getClassFile();
            addBridgeMethods(typeDefClass, node);

            ClassFile iter = ClassFinder.instance().find(typeDefClass.getSuper().getClassName()).get();
            while (!iter.getThis().getClassName().equals("java/lang/Object")) {
                writeStubs(iter, false, node);
                iter = ClassFinder.instance().find(iter.getSuper().getClassName()).get();
            }
            for (ClassInfo iface : typeDefClass.getInterfaces()) {
                try {
                    ClassFile get = ClassFinder.instance().find(iface.getClassName()).get();
                    writeStubs(get, true, node);
                } catch (IOException | ResolveException ex) {
                    ex.printStackTrace(System.err);
                    System.exit(-1);
                }
            }
            checkMethods(typeDefClass, true, node);
            interfaceCheck(typeDefClass, node);

            if (typeImpl.getMethods("<init>") == null) {

                Type.MethodType constructor = new Type.MethodType(Arrays.asList((Type) new Type.ClassType(null, typeDefClass.getThis().getClassName(), null)),
                        new Type.SimpleType("void"));
                MethodDescriptor makeMethodDescriptor = stage2Env.getTypeUtils().makeMethodDescriptor(constructor);
                MethodInfo constructorInfo = new MethodInfo("<init>", makeMethodDescriptor,
                        typeImpl.getThis(), MethodAccessFlag.PRIVATE);

                ConstructorBuilder.build(constructorInfo, stage2Env);

                typeImpl.addMethod(constructorInfo);


                constructor = new Type.MethodType(Collections.EMPTY_LIST, new Type.SimpleType("void"));
                makeMethodDescriptor = stage2Env.getTypeUtils().makeMethodDescriptor(constructor);
                constructorInfo = new MethodInfo("$init$", makeMethodDescriptor,
                        typeImpl.getThis(), MethodAccessFlag.PRIVATE);
                typeImpl.addMethod(constructorInfo);
            }


        } catch (ByteCodeException | IOException | ResolveException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }

    }
}