/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package stage3;

import abstractsyntaxtree.ArgumentsNode;
import abstractsyntaxtree.BlockNode;
import abstractsyntaxtree.ClosureEnvNode;
import abstractsyntaxtree.ClosureReturnNode;
import abstractsyntaxtree.ExpressionNode;
import abstractsyntaxtree.FormalParametersNode;
import abstractsyntaxtree.InfixOpNode;
import codegeneration.Visitor;
import codegeneration.abstractvisitors.AbstractClosureCreatorVisitor;
import codegeneration.environment.ExpressionEnvironment;
import codegeneration.environment.Stage0Environment;
import codegeneration.environment.Stage1Environment;
import codegeneration.environment.Stage2Environment;
import codegeneration.environment.Stage3Environment;
import codegeneration.instruction.InstructionCreator;
import codegeneration.instruction.MethodInvokeCreator;
import datastructs.Pair;
import exceptions.CompilerException;
import java.io.IOException;
import java.lang.invoke.MethodHandles;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import minicompiler.CompileManager;
import stage2.Stage2FormalParametersVisitor;
import stage3.states.BlockState;
import stage3.states.MethodState;
import symbol.LocalTable;
import teajay.bytecode.ClassAccessFlag;
import teajay.bytecode.ClassFile;
import teajay.bytecode.InnerClassAccessFlag;
import teajay.bytecode.attributes.code.AALoad;
import teajay.bytecode.attributes.code.ALoad;
import teajay.bytecode.attributes.code.AStore;
import teajay.bytecode.attributes.code.CheckCast;
import teajay.bytecode.attributes.code.Dup;
import teajay.bytecode.attributes.code.InvokeSpecial;
import teajay.bytecode.attributes.code.JVMInstruction;
import teajay.bytecode.attributes.code.JVMInstructionList;
import teajay.bytecode.attributes.code.Label;
import teajay.bytecode.attributes.code.NOP;
import teajay.bytecode.attributes.code.New;
import teajay.bytecode.attributes.code.PutField;
import teajay.bytecode.attributes.code.Return;
import teajay.bytecode.attributes.innerclass.InnerClasses;
import teajay.bytecode.attributes.signature.Signature;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.cpool.FieldRefInfo;
import teajay.bytecode.cpool.MethodDescriptor;
import teajay.bytecode.cpool.MethodRefInfo;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.fields.FieldAccessFlag;
import teajay.bytecode.fields.FieldInfo;
import teajay.bytecode.methods.MethodAccessFlag;
import teajay.bytecode.methods.MethodInfo;
import teajay.generics.check.TypeException;
import teajay.generics.parser.Type;
import teajay.generics.writer.SignatureEncoder;
import teajay.runtime.exceptions.ResolveException;
import teajay.util.io.ClassFinder;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 *
 * @author Michael (Burzan | Kußmann)
 */
class Stage3ClosureCreatorVisitor extends AbstractClosureCreatorVisitor {

    private final Stage3Environment stage3Env;
    private Stage2Environment stage2Env;
    private Stage3ClosureReturnVisitor closureReturn;
    private Stage3ClosureEnvVisitor closureEnv;
    private Stage2FormalParametersVisitor formalParameters;
    private Stage3Environment cEnv;
    private Type.ClassType cType;
    private ClassFile closure;

    public Stage3ClosureCreatorVisitor(Stage3Environment stage3Env) {
        this.stage3Env = stage3Env;
        this.stage2Env = new Stage2Environment(stage3Env.getStage1());
        this.stage2Env.setTypeUtils(stage3Env.getTypeUtils());
    }

    @Override
    public void begin(ClosureEnvNode node) throws CompilerException {
    }

    @Override
    public void begin(ClosureReturnNode node) throws CompilerException {
    }

    @Override
    public void begin(FormalParametersNode node) throws CompilerException {
    }

    @Override
    public void begin(ArgumentsNode node) throws CompilerException {
    }

    @Override
    public Visitor visit(ClosureEnvNode node) throws CompilerException {
        return closureEnv = new Stage3ClosureEnvVisitor(stage3Env);
    }

    @Override
    public Visitor visit(ClosureReturnNode node) throws CompilerException {
        return closureReturn = new Stage3ClosureReturnVisitor(stage3Env);
    }

    @Override
    public Visitor visit(FormalParametersNode node) throws CompilerException {
        stage2Env.setTypeVariables(stage3Env.getTypeVariables());
        return formalParameters = new Stage2FormalParametersVisitor(stage2Env);
    }

    @Override
    public Visitor visit(ArgumentsNode node) throws CompilerException {
        return new Stage3ArgumentsVisitor(stage3Env);
    }

    @Override
    public void end(ClosureEnvNode node) throws CompilerException {
    }

    @Override
    public void end(ClosureReturnNode node) throws CompilerException {
    }

    @Override
    public void end(FormalParametersNode node) throws CompilerException {
        try {
            stage2Env.resetGenericEnv(); //<< reset generic enviroment look visit
            Stage0Environment env0 = new Stage0Environment();

            ClassFile outer = stage3Env.getClassFile();


            String name = stage3Env.createClosureName();
            closure = new ClassFile(name, "teajay/lang/Closure", ClassAccessFlag.SYNTHETIC, ClassAccessFlag.FINAL, ClassAccessFlag.SUPER);
            String innerName = name.substring(name.lastIndexOf("$") + 1);
            InnerClasses closureInner = new InnerClasses();
            closureInner.addInnerClass(closure.getThis(), outer.getThis(), innerName, InnerClassAccessFlag.PRIVATE, InnerClassAccessFlag.FINAL, InnerClassAccessFlag.STATIC, InnerClassAccessFlag.SYNTHETIC);
            closure.addAttribute(closureInner);
            closure.addAttribute(new TeaJayAttribute(TeaJayAttribute.Type.Closure));
            env0.setClassFile(closure);
            env0.setSymbolFacade(stage3Env.getSymbolFacade());
            Stage1Environment env1 = new Stage1Environment(env0);
            env1.setFacade(stage3Env.getSymbolFacade());
            Stage2Environment env2 = new Stage2Environment(env1);
            env2.setTypeUtils(stage3Env.getTypeUtils());


            InnerClasses outerInner = (InnerClasses) outer.getAttribute(InnerClasses.NAME);
            if (outerInner == null) {
                outerInner = new InnerClasses();
                outer.addAttribute(outerInner);
            }
            outerInner.addInnerClass(closure.getThis(), outer.getThis(), innerName, InnerClassAccessFlag.PRIVATE, InnerClassAccessFlag.FINAL, InnerClassAccessFlag.STATIC, InnerClassAccessFlag.SYNTHETIC);

            MethodDescriptor desc = new MethodDescriptor(FieldDescriptor.getObject(), FieldDescriptor.makeArray(FieldDescriptor.getObject(), 1));

            MethodInfo info = new MethodInfo("execute", desc, closure.getThis(), MethodAccessFlag.PUBLIC, MethodAccessFlag.SYNTHETIC, MethodAccessFlag.VARARGS);
            env2.pushMethodInfo(info);
            cEnv = new Stage3Environment(env2);
            cEnv.setTypeVariables(stage3Env.getTypeVariables());

            closure.addMethod(info);
            List<Type> list = new ArrayList<>();

            cType = new Type.ClassType(null, "teajay/lang/Closure", list);
            Type retType;
            boolean isVoid = false;
            if (this.closureReturn == null) {
                //retType = new Type.SimpleType("void");
                retType = new Type.ClassType(null, "java/lang/Void", null);
                isVoid = true;
            } else {
                retType = this.closureReturn.getType();
                if (retType instanceof Type.ClassType) {
                    Type.ClassType classType = (Type.ClassType) retType;
                    if (classType.name.equals("java/lang/Void")) {
                        isVoid = true;
                    }
                }
            }

            list.add(retType);
            List<Pair<Type, String>> fieldContainer = formalParameters.getFieldContainer();

            for (Pair<Type, String> p : fieldContainer) {
                if (p.value1 instanceof Type.ArrayType) {
                    throw new CompilerException("VarArg closures not supported yet.", node.getSymbol());
                }
                list.add(p.value1);
            }

            Type.MethodType methodSig = new Type.MethodType(list.subList(1, list.size()), retType);
            info.addAttribute(new Signature(SignatureEncoder.encode(methodSig)));


            FieldDescriptor[] envArray;
            String[] fieldNames;
            ExpressionNode[] expressionNodes;
            List<Type> typesList = new LinkedList<>();
            if (closureEnv == null) {
                envArray = new FieldDescriptor[0];
                fieldNames = new String[0];
                expressionNodes = new ExpressionNode[0];
            } else {
                Set<Map.Entry<String, Pair<Type, ExpressionNode>>> entrySet = closureEnv.getEnvironmentList().entrySet();
                envArray = new FieldDescriptor[entrySet.size()];
                fieldNames = new String[entrySet.size()];
                expressionNodes = new ExpressionNode[entrySet.size()];
                int i = 0;

                for (Map.Entry<String, Pair<Type, ExpressionNode>> entry : entrySet) {
                    envArray[i] = cEnv.getTypeUtils().makeFieldDescriptor(entry.getValue().value1);
                    typesList.add(entry.getValue().value1);
                    FieldInfo fieldInfo = new FieldInfo(entry.getKey(), envArray[i], FieldAccessFlag.PRIVATE, FieldAccessFlag.SYNTHETIC);
                    closure.addField(fieldInfo);
                    fieldInfo.addAttribute(new Signature(SignatureEncoder.encode(entry.getValue().value1)));
                    expressionNodes[i] = entry.getValue().value2;
                    fieldNames[i] = entry.getKey();
                    i++;
                }
            }




            MethodDescriptor constrDesc = new MethodDescriptor(FieldDescriptor.getVoid(), envArray);
            MethodInfo constrInfo = new MethodInfo("<init>", constrDesc, closure.getThis(), MethodAccessFlag.PUBLIC, MethodAccessFlag.SYNTHETIC);
            Type.MethodType methodSignature = new Type.MethodType(typesList, new Type.SimpleType("void"));
            constrInfo.addAttribute(new Signature(SignatureEncoder.encode(methodSignature)));
            closure.addMethod(constrInfo);
            JVMInstructionList lst = constrInfo.getCode().getInstructionList();
            lst.add(new ALoad(0));
            lst.add(new InvokeSpecial(MethodRefInfo.make("teajay/lang/Closure", "<init>", "void")));
            for (int i = 0; i < envArray.length; i++) {
                lst.add(new ALoad(0));
                lst.add(new ALoad(i + 1));
                lst.add(new PutField(FieldRefInfo.make(closure.getThis(), fieldNames[i], envArray[i])));

            }
            lst.add(new Return());

            ExpressionEnvironment env = stage3Env.getExpressionEnvironment();

            Iterator<? extends Type> paramTypes = methodSignature.paramTypes.iterator();
            Iterator<ExpressionNode> expressionIt = Arrays.asList(expressionNodes).iterator();

            while (paramTypes.hasNext() && expressionIt.hasNext()) {
                ExpressionNode expr = expressionIt.next();
                Deque<InfixOpNode.InfixOps> stack = new LinkedList<>();
                ExpressionEnvironment argEnv = new ExpressionEnvironment(stage3Env);
                stage3Env.setExpressionEnvironment(argEnv);
                expr.accept(new Stage3ExpressionVisitor(stage3Env, stack, false));
                InstructionCreator.PutOperatorInstructions(argEnv, stack, node);
                if (argEnv.isShortCircuit()) {
                    Label gotoLabel = new Label("gotoLabel");

                    argEnv.writeShortCircuit(gotoLabel);
                    BlockState hasState = stage3Env.hasState(BlockState.class);
                    hasState.addLabel(gotoLabel);

                    argEnv.insertCodeWithVerificationStack(new NOP(), argEnv.getOperandStack());
                }
                Type next = paramTypes.next();
                try {
                    stage3Env.getTypeUtils().checkAssignability(next, argEnv.typeOfResult());
                } catch (TypeException ex) {
                    throw new CompilerException("Required: " + next + ", found: " + argEnv.typeOfResult(), node.getSymbol());
                }
            }
            MethodDescriptor modifiedDescr = new MethodDescriptor(FieldDescriptor.fromClassFile(cEnv.getClassFile()),
                    constrInfo.makeMethodRef().getNameAndType().getDescriptor().getParameters());
            //Force flush of stack map frame
            stage3Env.getExpressionEnvironment().insertCode(new NOP());
            stage3Env.invokeDynamic(env.getMethod(), modifiedDescr, cEnv.getClassFile().getThis());

            env.setResultType(cType);

            ClassFinder.instance().addClassFile(closure);

            cEnv.setCurrentMethodExpectedType(retType);
            //cEnv.removeMethodInfo();
            cEnv.setLocalTable(new LocalTable(info, cEnv.getTypeUtils()));
            MethodState state = new MethodState();
            state.setClosure(cType);
            state.setClosureIsVoid(isVoid);
            cEnv.initState(state);
            cEnv.changeState(new BlockState(), node);
            //cEnv.setThis(cType);
            cEnv.createThis();


            LocalTable localTable = cEnv.getLocalTable();
            localTable.addFrame();
            localTable.addEntry("@1", cEnv.getThis());
            localTable.addEntry("@2", new Type.ArrayType(new Type.ClassType(null, "java/lang/Object", null)));
            JVMInstructionList ins = info.getCode().getInstructionList();
            for (int i = 0; i < fieldContainer.size(); i++) {
                Pair<Type, String> param = fieldContainer.get(i);
                ins.add(new ALoad(1)); // Array on the stack
                ins.pushInt(i);
                ins.add(new AALoad()); // Access to array element at position i
                FieldDescriptor makeFieldDescriptor = cEnv.getTypeUtils().makeFieldDescriptor(param.value1);
                ins.add(new CheckCast(makeFieldDescriptor.getClassInfo()));
                localTable.addEntry(param.value2, param.value1);
                int index = localTable.getEntry(param.value2).getIndex();
                ins.add(new AStore(index));
            }
            stage3Env.getExpressionEnvironment().setResultType(cType);

        } catch (ByteCodeException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
    }

    @Override
    public void end(ArgumentsNode node) throws CompilerException {
        try {
            MethodInvokeCreator.CallContext context = new MethodInvokeCreator.CallContext(stage3Env, cType, "execute", 1);
            MethodInvokeCreator.createMethodCall(context, stage3Env.getExpressionEnvironment(), "execute", Collections.EMPTY_LIST, node);
        } catch (ByteCodeException | IOException | ResolveException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
    }

    @Override
    public void begin(BlockNode node) throws CompilerException {
    }

    @Override
    public Visitor visit(BlockNode node) throws CompilerException {
        cEnv.setExpressionEnvironment(new ExpressionEnvironment(cEnv));
        return new Stage3BlockVisitor(cEnv, true);
    }

    @Override
    public void end(BlockNode node) throws CompilerException {

        CompileManager.instance().writeClassFile(closure, null);

    }
}
