/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.generics.check;

import java.util.ArrayList;
import java.util.List;
import teajay.generics.parser.Type;
import static teajay.generics.parser.Type.WildcardType.Kind.EXTENDS;
import static teajay.generics.parser.Type.WildcardType.Kind.SUPER;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
class ReplaceVisitor implements Type.Visitor<Type, Void> {

    private final Context context;
    private final boolean unerased;
    private final boolean ignoreUnknownTypeVariables;

    public ReplaceVisitor(Context context) {
        this.context = context;
        this.unerased = false;
        this.ignoreUnknownTypeVariables = false;
    }

    public ReplaceVisitor(Context context, boolean unerased) {
        this.context = context;
        this.unerased = unerased;
        this.ignoreUnknownTypeVariables = false;
    }

    public ReplaceVisitor(Context context, boolean unerased, boolean ignoreUnknownTypeVariables) {
        this.context = context;
        this.unerased = unerased;
        this.ignoreUnknownTypeVariables = ignoreUnknownTypeVariables;
    }

    @Override
    public Type visitSimpleType(Type.SimpleType type, Void p) {
        if (type.isPrimitiveType()) {
            return type;
        } else {
            Type t;
            if (unerased) {
                //t = context.getUnerasedType(type.name);
                Type.TypeParamType unerased = context.getUnerasedType(type.name);
                Type classBound = unerased.classBound;
                if (classBound instanceof Type.ClassType) {
                    Type.ClassType classType = (Type.ClassType) classBound;
                    if (classType.name.equals("java/lang/Object")) {
                        if (unerased.interfaceBounds != null && !unerased.interfaceBounds.isEmpty()) {
                            classBound = unerased.interfaceBounds.get(0);
                        }
                    }
                }
                t = classBound;
                if (t == null) {
                    if (ignoreUnknownTypeVariables) {
                        return type;
                    } else {
                        throw new TypeException("Unknown type variable " + type);
                    }
                }
                t = t.accept(this, null);
            } else {
                try {
                    t = context.getErasedType(type.name);
                } catch (TypeException ex) {
                    if (ignoreUnknownTypeVariables) {
                        t = type;
                    } else {
                        throw ex;
                    }
                }
            }
            if (t == null) {
                if (ignoreUnknownTypeVariables) {
                    return type;
                } else {
                    throw new TypeException("Unknown type variable " + type);
                }
            }
            return t;
        }
    }

    @Override
    public Type visitArrayType(Type.ArrayType type, Void p) {
        return new Type.ArrayType(type.elemType.accept(this, null));
    }

    @Override
    public Type visitMethodType(Type.MethodType type, Void p) {
        List<Type.TypeParamType> typeParamTypes = new ArrayList<>();
        if (type.typeParamTypes != null) {
            for (Type.TypeParamType t : type.typeParamTypes) {
                typeParamTypes.add((Type.TypeParamType) t.accept(this, null));
            }
        }
        List<Type> paramTypes = new ArrayList<>();
        if (type.paramTypes != null) {
            for (Type t : type.paramTypes) {
                paramTypes.add(t.accept(this, null));
            }
        }
        Type returnType = type.returnType.accept(this, null);
        List<Type> throwsTypes = new ArrayList<>();
        if (type.throwsTypes != null) {
            for (Type t : type.throwsTypes) {
                throwsTypes.add(t.accept(this, null));
            }
        }
        return new Type.MethodType(typeParamTypes, paramTypes, returnType, throwsTypes);
    }

    @Override
    public Type visitClassSigType(Type.ClassSigType type, Void p) {
        List<Type.TypeParamType> typeParamTypes = new ArrayList<>();
        if (type.typeParamTypes != null) {
            for (Type.TypeParamType t : type.typeParamTypes) {
                typeParamTypes.add((Type.TypeParamType) t.accept(this, null));
            }
        }
        Type superclassType = type.superclassType.accept(this, null);
        List<Type> superInterfaces = new ArrayList<>();
        if (type.superinterfaceTypes != null) {
            for (Type t : type.superinterfaceTypes) {
                superInterfaces.add(t.accept(this, null));
            }
        }
        return new Type.ClassSigType(typeParamTypes, superclassType, superInterfaces);
    }

    @Override
    public Type visitClassType(Type.ClassType type, Void p) {
        Type.ClassType outerType = null;
        if (type.outerType != null) {
            outerType = (Type.ClassType) type.outerType.accept(this, null);
        }
        List<Type> typeArgs = new ArrayList<>();
        if (type.typeArgs != null) {
            for (Type t : type.typeArgs) {
                typeArgs.add(t.accept(this, null));
            }
        }
        return new Type.ClassType(outerType, type.name, typeArgs);
    }

    @Override
    public Type visitTypeParamType(Type.TypeParamType type, Void p) {
        Type classBound = null;
        if (type.classBound != null) {
            classBound = type.classBound.accept(this, null);
        }
        List<Type> interfaceBounds = new ArrayList<>();
        if (type.interfaceBounds != null) {
            for (Type t : type.interfaceBounds) {
                interfaceBounds.add(t.accept(this, null));
            }
        }
        return new Type.TypeParamType(type.name, classBound, interfaceBounds);
    }

    @Override
    public Type visitWildcardType(Type.WildcardType type, Void p) {
        switch (type.kind) {
            case EXTENDS:
            case SUPER:
                return new Type.WildcardType(type.kind, type.boundType.accept(this, null));
            default:
                return type;
        }
    }
}
