/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.generics.check;

import abstractsyntaxtree.SyntaxNode;
import datastructs.Pair;
import exceptions.CompilerException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import lexer.Symbol;
import symbol.DerivationTree;
import teajay.bytecode.ClassFile;
import teajay.bytecode.attributes.Exceptions;
import teajay.bytecode.attributes.signature.Signature;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.cpool.MethodDescriptor;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.fields.FieldInfo;
import teajay.bytecode.methods.MethodInfo;
import teajay.generics.parser.SignatureParser;
import teajay.generics.parser.Type;
import teajay.runtime.exceptions.ResolveException;
import teajay.util.io.ClassFileRef;
import teajay.util.io.ClassFinder;

/**
 * Implementation of TeaJay's generic type system. This is a prototype
 * implementation!
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class TypeUtils {

    private static final ImplCheckVisitor implCheck = new ImplCheckVisitor();
    static final Type.ClassType object = new Type.ClassType(null, "java/lang/Object", null);
    private final Context context;

    public TypeUtils() {
        this.context = new Context();
    }

    public void removeFrame() {
        context.removeFrame();
    }

    public void addFrame() {
        context.addFrame();
    }

    public void addTypeToContext(Type t1) {
        t1.accept(new EraseVisitor(context), null);
    }

    private Type.ClassType flatten(Type.ClassType target, Type.ClassType type) throws TypeException {
        Type.ClassType result = (Type.ClassType) type.accept(new SignatureDescenderVisitor(this, new Context(), true), target);
        if (result == null) {
            throw new TypeException("Incompatible types: " + target + " <-> " + type);
        }
        return result;
    }

    public void checkAssignability(Type t1, Type t2) throws TypeException {
        checkAssignability(t1, t2, true);
    }

    private boolean isClosure(Type t) {
        if (t instanceof Type.ClassType) {
            Type.ClassType classType = (Type.ClassType) t;
            return classType.name.equals("teajay/lang/Closure");
        } else {
            return false;
        }
    }

    public void checkAssignability(Type t1, Type t2, boolean first) throws TypeException {
        if (t1 == null || t2 == null) {
            throw new NullPointerException(t1 == null ? "t1" : "t2");
        }

        if (t2 instanceof Type.ClassType) {
            if (t1 instanceof Type.SimpleType) {
                Type.SimpleType simpleType = (Type.SimpleType) t1;
                if (simpleType.isPrimitiveType()) {
                    throw new TypeException("Incompatiable types: " + t1 + " - " + t2);
                }
            }
            Type.ClassType classType = (Type.ClassType) t2;
            if (classType.name.equals("null")) {
                return; //OK
            }
        }

        if (t1 instanceof Type.WildcardType || t2 instanceof Type.WildcardType) {
            t1.accept(new FlatAssignabilityVisitor(this, first), t2);
        } else if (t1 instanceof Type.ClassType && t2 instanceof Type.ClassType) {
            Type.ClassType c1 = (Type.ClassType) t1;
            Type.ClassType c2 = (Type.ClassType) t2;
            ClassFileRef a = ClassFinder.instance().find(c1.name);
            ClassFileRef b = ClassFinder.instance().find(c2.name);
            if (a == null || b == null) {
                throw new TypeException("Class not found: " + (a == null ? c1.name : c2.name));
            }
            try {
                if (!DerivationTree.instance().isAssignable(a, b)) {
                    throw new TypeException("Incompatible types: " + t1 + " - " + t2);
                }
                if (c1.name.equals("java/lang/Object")) {
                } else {

                    Type.ClassType flat;
                    if (isClosure(c1) && isClosure(c2)) {
                        flat = c2;
                        List<Type> newTypeArgs = new ArrayList<>();
                        Iterator<Type> it = c1.typeArgs.iterator();
                        if (it.hasNext()) {
                            //Return value
                            newTypeArgs.add(new Type.WildcardType(Type.WildcardType.Kind.EXTENDS, it.next()));
                        }
                        while (it.hasNext()) {
                            //Parameter
                            newTypeArgs.add(new Type.WildcardType(Type.WildcardType.Kind.SUPER, it.next()));
                        }
                        c1 = new Type.ClassType(c1.outerType, c1.name, newTypeArgs);
                    } else {
                        flat = flatten(c1, c2);
                    }
                    //TypeChecker.checkFlatType(flat, c1);
                    c1.accept(new FlatAssignabilityVisitor(this, first), flat);
                }
            } catch (ByteCodeException | IOException | ResolveException ex) {
                throw new TypeException(ex);
            }
        } else if (t1 instanceof Type.SimpleType && t2 instanceof Type.SimpleType) {
            Type.SimpleType c1 = (Type.SimpleType) t1;
            Type.SimpleType c2 = (Type.SimpleType) t2;
            if (!c1.name.equals(c2.name)) {
                throw new TypeException("Incompatible types: " + t1 + " - " + t2);
            }
        } else if (t1 instanceof Type.ClassType && t2 instanceof Type.SimpleType) {
            if (((Type.SimpleType) t2).isPrimitiveType()) {
                throw new TypeException("Incompatible types: " + t1 + " - " + t2);
            }
            Type t = context.getErasedType(((Type.SimpleType) t2).name);
            checkAssignability(t1, t);
        } else if (t2 instanceof Type.ArrayType) {
            Type.ArrayType a2 = (Type.ArrayType) t2;
            if (!t1.isObject()) {
                if (t1 instanceof Type.ArrayType) {
                    Type.ArrayType a1 = (Type.ArrayType) t1;
                    Pair<Type, Integer> p1 = getElementTypeAndDimension(a1);
                    Pair<Type, Integer> p2 = getElementTypeAndDimension(a2);
                    if (p1.value1.isObject() && p2.value1.isObject()) {
                        if (p1.value2 > p2.value2) {
                            throw new TypeException("Incompatible types: " + t1 + " - " + t2);
                        }
                    } else if (p1.value2.intValue() == p2.value2.intValue()) {
                        checkAssignability(p1.value1, p2.value1);
                    } else {
                        throw new TypeException("Incompatible types: " + t1 + " - " + t2);
                    }
                } else {
                    throw new TypeException("Incompatible types: " + t1 + " - " + t2);
                }
            }
        } else {
            throw new TypeException("Incompatible types: " + t1 + " - " + t2);
        }
    }

    private static Pair<Type, Integer> getElementTypeAndDimension(Type.ArrayType t) {
        Type tmp = t.elemType;
        int dim = 1;
        while (tmp != null && tmp instanceof Type.ArrayType) {
            tmp = ((Type.ArrayType) tmp).elemType;
            dim++;
        }
        return new Pair<>(tmp, dim);
    }

    public void checkCallability(Type t1, List<? extends Type> params) throws TypeException {
        if (params == null || t1 == null) {
            throw new NullPointerException(params == null ? "params" : "t1");
        }
        if (t1 instanceof Type.MethodType) {
            Type.MethodType method = (Type.MethodType) t1;
            if (method.paramTypes != null) {
                if (method.paramTypes.size() == params.size()) {
                    Iterator<? extends Type> ita = method.paramTypes.iterator();
                    Iterator<? extends Type> itb = params.iterator();
                    while (ita.hasNext()) {
                        checkAssignability(ita.next(), itb.next());
                    }
                } else {
                    throw new TypeException("Method parameter count missmatch");
                }
            } else if (!params.isEmpty()) {
                throw new TypeException("Method parameter count missmatch");
            }
        } else {
            throw new TypeException("expecting method type");
        }
    }

    public void checkClassSignature(Type t1, Type t2) throws TypeException {
        if (t1 instanceof Type.ClassType && t2 instanceof Type.ClassSigType) {
            Type.ClassSigType type = (Type.ClassSigType) t2;
            Type.ClassType c = (Type.ClassType) t1;
            if (c.name.equals("teajay/lang/Closure")) {
                return;
            }
            if (type.typeParamTypes != null && c.typeArgs != null) {
                if (type.typeParamTypes.size() != c.typeArgs.size()) {
                    throw new TypeException("Type argument count missmatch");
                }
                Iterator<Type.TypeParamType> ita = type.typeParamTypes.iterator();
                Iterator<Type> itb = c.typeArgs.iterator();
                while (ita.hasNext() && itb.hasNext()) {
                    Type.TypeParamType ta = ita.next();
                    Type tb = itb.next();
                    checkAssignability(ta.classBound == null ? object : ta.classBound, tb);
                    if (ta.interfaceBounds != null) {
                        for (Type t : ta.interfaceBounds) {
                            checkAssignability(t, tb);
                        }
                    }
                }
            } else if (type.typeParamTypes == null && c.typeArgs == null) {
            } else if (type.typeParamTypes == null && !c.typeArgs.isEmpty()) {
                throw new TypeException("Type argument count missmatch");
            } else if (c.typeArgs == null && !type.typeParamTypes.isEmpty()) {
                throw new TypeException("Type argument count missmatch");
            }
        } else {
            throw new TypeException("expecting class type and class signature");
        }
    }

    public void checkClassSignature(Type t1) {
        if (t1 instanceof Type.ClassType) {
            Type t2 = getClassSignature((Type.ClassType) t1);
            checkClassSignature(t1, t2);
        } else {
            throw new TypeException("Can only check class types");
        }
    }

    /**
     * This method also generates the current context information!!!
     *
     * @param t1
     * @return
     */
    public Type eraseGenerics(Type t1) throws TypeException {
        return t1.accept(new EraseVisitor(new Context(context)), null);
    }

    public Type.ClassType createThis(String qid) {
        Type.ClassSigType sig = getClassSignature(qid);
        List<Type> typeArgs = new ArrayList<>();
        if (sig.typeParamTypes != null) {
            for (Type.TypeParamType t : sig.typeParamTypes) {
                Type.SimpleType st = new Type.SimpleType(t.name);
                typeArgs.add(st);
            }
        }

        return new Type.ClassType(null, qid, typeArgs);
    }

    public Type.ClassType createErasedThis(String qid) {
        Type.ClassSigType sig = getClassSignature(qid);
        List<Type> typeArgs = new ArrayList<>();
        if (sig.typeParamTypes != null) {
            for (Type.TypeParamType t : sig.typeParamTypes) {
                Type.SimpleType st = new Type.SimpleType(t.name);
                typeArgs.add(st.accept(new ReplaceVisitor(context, true), null));
            }
        }

        return new Type.ClassType(null, qid, typeArgs);
    }

    public MethodDescriptor makeMethodDescriptor(Type t) throws TypeException {
        if (t == null) {
            throw new NullPointerException("t");
        }
        if (!(t instanceof Type.MethodType)) {
            throw new TypeException("expecting method type");
        }
        Type mtype = eraseGenerics(t);
        DescriptorVisitor v = new DescriptorVisitor();
        mtype.accept(v, null);
        return v.getMethodDescriptor();
    }

    public FieldDescriptor makeFieldDescriptor(Type t) throws TypeException {
        if (t == null) {
            throw new NullPointerException("t");
        }
        if (t instanceof Type.ArrayType || t instanceof Type.ClassType || t instanceof Type.SimpleType || t instanceof Type.WildcardType) {
            Type type = eraseGenerics(t);
            DescriptorVisitor v = new DescriptorVisitor();

            return type.accept(v, null);
        } else {
            throw new TypeException("Type not suitable for a field: " + t.getClass().getSimpleName());
        }
    }

    public void checkSignatureCompatibility(Type t1) throws TypeException {
        checkSignatureCompatibility(t1, new Context(context), true);
    }

    private void checkSignatureCompatibility(Type t1, Context localContext, boolean check) throws TypeException {
        if (t1 instanceof Type.ClassType) {
            Type.ClassType ctype = (Type.ClassType) t1;
            if (ctype.name.equals("teajay/lang/Closure")) {
                //Closures can have any signature
                return;
            }
            Type.ClassSigType thisSig = getClassSignature(ctype);
            if (ctype.typeArgs != null && thisSig.typeParamTypes != null) {
                if (ctype.typeArgs.size() != thisSig.typeParamTypes.size()) {
                    throw new TypeException("Type argument count missmatch: " + ctype + " <-> " + thisSig);
                }
                Type.ClassType outerType = ctype.outerType;
                while (outerType != null && !outerType.name.equals("java/lang/Object")) {
                    checkSignatureCompatibility(outerType, localContext, true);
                    outerType = outerType.outerType;
                }
                Iterator<Type> ita = ctype.typeArgs.iterator();
                Iterator<Type.TypeParamType> itb = thisSig.typeParamTypes.iterator();
                ReplaceVisitor rvisitor = new ReplaceVisitor(localContext);
                while (ita.hasNext()) {
                    Type.TypeParamType typeParamType = itb.next();
                    Type atype = ita.next().accept(rvisitor, null);
                    if (check) {
                        if (typeParamType.classBound != null) {
                            Type classBound = typeParamType.classBound.accept(rvisitor, null);
                            if (atype instanceof Type.WildcardType && ((Type.WildcardType) atype).kind == Type.WildcardType.Kind.UNBOUNDED) {
                                //OK
                            } else {
                                checkAssignability(classBound, atype);
                                checkSignatureCompatibility(classBound, localContext, true);
                            }
                        }
                        if (typeParamType.interfaceBounds != null) {
                            for (Type t : typeParamType.interfaceBounds) {
                                Type accept = t.accept(rvisitor, null);
                                if (atype instanceof Type.WildcardType && ((Type.WildcardType) atype).kind == Type.WildcardType.Kind.UNBOUNDED) {
                                    //OK
                                } else {
                                    checkAssignability(accept, atype);
                                    checkSignatureCompatibility(accept, localContext, true);
                                }
                            }
                        }
                    }
                    localContext.put(typeParamType.name, atype, typeParamType);
                }
            } else if (ctype.typeArgs != null && !ctype.typeArgs.isEmpty()) {
                throw new TypeException("Type argument count missmatch: " + ctype + " <-> " + thisSig);
            } else if (thisSig.typeParamTypes != null && !thisSig.typeParamTypes.isEmpty()) {
                throw new TypeException("Type argument count missmatch: " + ctype + " <-> " + thisSig);
            }
        } else if (t1 instanceof Type.SimpleType) {
        } else if (t1 instanceof Type.ArrayType) {
            Type.ArrayType arrayType = (Type.ArrayType) t1;
            checkSignatureCompatibility(arrayType.elemType);
        } else {
            throw new TypeException("expecting a class, array or simple type");
        }
    }

    public static Type.MethodType getMethodSignature(MethodInfo info) throws TypeException {
        Signature sig = null;
        if (info.getAttribute(Signature.NAME) != null) {
            sig = (Signature) info.getAttribute(Signature.NAME);
        }

        if (sig != null) {
            Type.MethodType mtype = (Type.MethodType) SignatureParser.parse(sig.getSignature());
            Exceptions ex = (Exceptions) info.getAttribute(Exceptions.NAME);
            List<Type> throwsTypes = null;
            if (mtype.throwsTypes == null || mtype.throwsTypes.isEmpty()) {
                if (ex != null) {
                    throwsTypes = new ArrayList<>();
                    for (ClassInfo c : ex.getExceptions()) {
                        throwsTypes.add(new Type.ClassType(null, c.getClassName(), null));
                    }
                }
            } else {
                throwsTypes = new LinkedList<>();
                throwsTypes.addAll(mtype.throwsTypes);
            }

            return new Type.MethodType(mtype.typeParamTypes, mtype.paramTypes, mtype.returnType, throwsTypes);
        } else {
            Type returnType = makeType(info.getDescriptor().getReturnType());
            List<Type> paramTypes = new ArrayList<>();
            for (FieldDescriptor descr : info.getDescriptor().getParameters()) {
                paramTypes.add(makeType(descr));
            }
            List<Type> throwsTypes = null;
            Exceptions ex = (Exceptions) info.getAttribute(Exceptions.NAME);
            if (ex != null) {
                throwsTypes = new ArrayList<>();
                for (ClassInfo c : ex.getExceptions()) {
                    throwsTypes.add(new Type.ClassType(null, c.getClassName(), null));
                }
            }
            return new Type.MethodType(null, paramTypes, returnType, throwsTypes);
        }
    }

    public Type getFieldType(FieldInfo info) {
        Signature sig = (Signature) info.getAttribute(Signature.NAME);
        if (sig == null) {
            return makeType(info.getDescriptor());
        } else {
            return SignatureParser.parse(sig.getSignature());
        }
    }

    public static boolean signatureEquals(Type.ClassSigType t1) {
        Type.ClassType superClass = (Type.ClassType) t1.superclassType;
        return signatureEquals(t1, getClassSignature(superClass.name));
    }

    private static boolean signatureEquals(Type.ClassSigType t1, Type.ClassSigType t2) {
        if (t1 == null || t2 == null) {
            throw new NullPointerException(t1 == null ? "t1" : "t2");
        }
        if (t1.typeParamTypes != null && t2.typeParamTypes != null) {
            if (t1.typeParamTypes.size() != t2.typeParamTypes.size()) {
                return false;
            }
            Iterator<Type.TypeParamType> ita = t1.typeParamTypes.iterator();
            Iterator<Type.TypeParamType> itb = t2.typeParamTypes.iterator();
            List<String> names = new LinkedList<>();

            while (ita.hasNext() && itb.hasNext()) {
                Type.TypeParamType ta = ita.next();
                Type.TypeParamType tb = itb.next();

                names.add(ta.name);

                if (!ta.toString().equals(tb.toString())) {
                    return false;
                }
            }
            if (t1.superclassType != null && t2.superclassType != null) {
                if (t1.superclassType instanceof Type.ClassType) {
                    Type.ClassType s1 = (Type.ClassType) t1.superclassType;
                    if (s1.typeArgs != null && (s1.typeArgs.size() == names.size())) {
                        Iterator<String> i1 = names.iterator();
                        Iterator<Type> i2 = s1.typeArgs.iterator();
                        while (i1.hasNext() && i2.hasNext()) {
                            Type t = i2.next();
                            if (t instanceof Type.SimpleType) {
                                if (!((Type.SimpleType) t).name.equals(i1.next())) {
                                    return false;
                                }
                            } else {
                                return false;
                            }
                        }
                    } else {
                        return false;
                    }
                } else {
                    return false;
                }
            } else {
                return false;
            }
        } else if (t1.typeParamTypes == null && t2.typeParamTypes == null) {
            //OK
        } else if (t1.typeParamTypes == null && !t2.typeParamTypes.isEmpty()) {
            return false;
        } else if (t2.typeParamTypes == null && !t1.typeParamTypes.isEmpty()) {
            return false;
        }

        return true;
    }

    public static Type makeType(FieldDescriptor descr) {
        Type result;
        if (descr.isArray()) {
            FieldDescriptor elementType = descr.getLeafElementType();
            if (elementType.isPrimitive()) {
                result = new Type.SimpleType(elementType.getClassInfo().getClassName());
            } else {
                result = new Type.ClassType(null, elementType.getClassInfo().getClassName(), null);
            }
            result = new Type.ArrayType(result);
            descr = descr.getElementType();
            while (descr.isArray()) {
                result = new Type.ArrayType(result);
                descr = descr.getElementType();
            }
        } else if (descr.isPrimitive()) {
            result = new Type.SimpleType(descr.getClassInfo().getClassName());
        } else {
            result = new Type.ClassType(null, descr.getClassInfo().getClassName(), null);
        }
        return result;
    }

    public static Type.ClassSigType getClassSignature(String qid) throws TypeException {
        ClassFileRef ref = ClassFinder.instance().find(qid);
        if (ref == null) {
            throw new TypeException("Class not found: " + qid);
        }
        try {
            ClassFile classFile = ref.get();
            Signature sig = (Signature) classFile.getAttribute(Signature.NAME);
            Type.ClassSigType sigType = null;
            if (sig != null) {
                Type tmp = SignatureParser.parse(sig.getSignature());
                if (tmp instanceof Type.ClassSigType) {
                    sigType = (Type.ClassSigType) tmp;
                } else if (tmp instanceof Type.ClassType) {
                    List<Type> interfaces = new ArrayList<>();
                    for (ClassInfo info : classFile.getInterfaces()) {
                        interfaces.add(new Type.ClassType(null, info.getClassName(), null));
                    }
                    sigType = new Type.ClassSigType(null, tmp, interfaces);
                } else {
                    throw new TypeException("Illegal class signature");
                }
            } else {
                ClassInfo superClass = classFile.getSuper();
                String superclass = null;
                if (superClass != null) {
                    superclass = superClass.getClassName();
                }
                List<Type> interfaces = new ArrayList<>();
                for (ClassInfo info : classFile.getInterfaces()) {
                    interfaces.add(new Type.ClassType(null, info.getClassName(), null));
                }
                sigType = new Type.ClassSigType(null, new Type.ClassType(null, superclass, null), interfaces);
            }
            return sigType;
        } catch (ByteCodeException | IOException | ResolveException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
            return null;
        }
    }

    public static Type.ClassSigType getClassSignature(Type.ClassType type) throws TypeException {
        String qid = type.name;
        return getClassSignature(qid);
    }

    public Type.MethodType getMethodSignatureByEnv(MethodInfo method, Type.ClassType env, Type.ClassType endPoint, List<Type> generics) throws TypeException {
        if (generics == null) {
            generics = Collections.emptyList();
        }
        Context ctx = new Context();
        Type.MethodType mtype = getMethodSignature(method);
        if (env.typeArgs == null || env.typeArgs.isEmpty()) {
            //Fill env with default class bound
            Type.ClassSigType sig = getClassSignature(endPoint);
            if (sig.typeParamTypes != null) {
                for (Type.TypeParamType t : sig.typeParamTypes) {
                    ctx.put(t.name, t.classBound, null);
                }
            }
        }
        SignatureDescenderVisitor sd = new SignatureDescenderVisitor(this, ctx, true);
        env.accept(sd, endPoint);
        ctx = sd.tempContext;

        ReplaceVisitor rvisitor = new ReplaceVisitor(ctx, false, true);

        List<Type> paramTypes = new ArrayList<>();
        if (mtype.paramTypes != null) {
            for (Type type : mtype.paramTypes) {
                paramTypes.add(type.accept(rvisitor, null));
            }
        }
        Type returnType = mtype.returnType.accept(rvisitor, null);
        List<Type> throwsTypes = new ArrayList<>();
        if (mtype.throwsTypes != null) {
            for (Type type : mtype.throwsTypes) {
                throwsTypes.add(type.accept(rvisitor, null));
            }
        }
        /*List<Type.TypeParamType> typeParamTypes = new ArrayList<>();
         if (mtype.typeParamTypes != null) {
         for (Type type : mtype.typeParamTypes) {
         typeParamTypes.add((Type.TypeParamType) type.accept(rvisitor, null));
         }
         }*/
        Context tmp = new Context();
        if (mtype.typeParamTypes != null && generics.size() == mtype.typeParamTypes.size()) {
            Iterator<Type> ita = generics.iterator();
            Iterator<? extends Type.TypeParamType> itb = mtype.typeParamTypes.iterator();

            while (ita.hasNext() && itb.hasNext()) {
                Type.TypeParamType typeParamType = itb.next();
                Type atype = ita.next();

                tmp.put(typeParamType.name, atype, null);
                ctx.put(typeParamType.name, atype, null);

                if (typeParamType.classBound != null) {
                    Type classBound = typeParamType.classBound.accept(rvisitor, null);

                    checkAssignability(classBound, atype);
                }
                if (typeParamType.interfaceBounds != null) {
                    for (Type tt : typeParamType.interfaceBounds) {
                        Type accept = tt.accept(rvisitor, null);
                        checkAssignability(accept, atype);
                    }
                }
            }
        } else if (mtype.typeParamTypes == null && generics.isEmpty()) {
            //ok
        } else {
            throw new TypeException("Type argument count missmatch");
        }

        mtype = new Type.MethodType(null, paramTypes, returnType, throwsTypes);
        mtype = (Type.MethodType) mtype.accept(new ReplaceVisitor(tmp, false, true), null);

        return mtype;
    }

    public static void checkTypes(Type type, Type.ClassType self, Symbol sym) throws CompilerException {
        checkTypes(type, self.name, sym);
    }

    public static void checkTypes(Type type, String self, Symbol sym) throws CompilerException {
        implCheck.init(self);
        try {
            type.accept(implCheck, null);
        } catch (TypeException ex) {
            throw new CompilerException(ex.getMessage(), sym);
        } finally {
            implCheck.reset();
        }
    }

    public Type.MethodType getMethodSignatureByEnv(MethodInfo method, Type.ClassType env, List<Type> generics) throws TypeException {
        try {
            ClassFileRef ownerRef = ClassFinder.instance().find(method.getDefiningClass().getClassName());
            if (ownerRef == null) {
                throw new TypeException("Class not found: " + method.getDefiningClass().getClassName());
            }
            ClassFile owner = ownerRef.get();
            Type.ClassType endPoint = null;
            if (!owner.getThis().getClassName().equals("teajay/lang/Closure")) {
//                MethodDescriptor descr = method.getDescriptor();
//                String name = method.getName().getValue();
//                ClassFileRef ref = ClassFinder.instance().find(env.name);
//
//                top:
//                while (ref != null) {
//                    ClassFile file = ref.get();
//                    Collection<MethodInfo> methods = file.getMethods(name);
//                    if (methods != null) {
//                        for (MethodInfo info : methods) {
//                            if (!info.isBridge()) {
//                                if (info.getDescriptor().equals(descr)) {
//                                    //System.err.println("INFO: " + info.getA);
//                                    //System.err.println("FLAGS: " + info.getAccessFlags());
//                                    //System.err.println("DESCR: " + descr);
//                                    endPoint = new Type.ClassType(null, file.getThis().getClassName(), null);
//                                    break top;
//                                }
//                            }
//                        }
//                    }
//                    if (file.getSuper() == null) {
//                        ref = null;
//                    } else {
//                        ref = ClassFinder.instance().find(file.getSuper().getClassName());
//                    }
//                }
                endPoint = new Type.ClassType(null, method.getDefiningClass().getClassName(), null);
                ClassFileRef envRef = ClassFinder.instance().find(env.name);
                if (!DerivationTree.instance().isAssignable(ownerRef, envRef)) {
                    throw new TypeException("The environment has to be assignable to the method owner: " + ownerRef.getQID() + " - " + envRef.getQID());
                }
            } else {
                endPoint = new Type.ClassType(null, owner.getThis().getClassName(), null);
            }
            /*if (endPoint == null) {
             throw new TypeException("no such method: " + method);
             }*/
            return getMethodSignatureByEnv(method, env, endPoint, generics);
        } catch (ByteCodeException | IOException | ResolveException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
        return null;
    }

    public Type.ClassType getUnerasedType(String typeVariable) {
        Type.TypeParamType param = context.getUnerasedType(typeVariable);
        if (param == null) {
            throw new TypeException("unknown type variable " + typeVariable);
        }
        Type.ClassType res = (Type.ClassType) param.classBound;
        if ((res == null || res.isObject()) && (param.interfaceBounds != null && !param.interfaceBounds.isEmpty())) {
            res = (Type.ClassType) param.interfaceBounds.get(0);
        }
        return res;
    }

    public Type.TypeParamType getTypeParamType(String typeVariable) {
        return context.getUnerasedType(typeVariable);
    }

    FieldDescriptor toJavaArray(Type t1) {
        if (t1 instanceof Type.ClassType && ((Type.ClassType) t1).name.equals("teajay/lang/Array")) {
            Type.ClassType type = ((Type.ClassType) t1);
            if (type.outerType != null) {
                System.err.println("Warning: Outer types not allowed for Array");
            }
            Type arg = object;
            if (type.typeArgs != null && type.typeArgs.size() == 1) {
                arg = type.typeArgs.get(0);
            } else if (type.typeArgs != null && !type.typeArgs.isEmpty()) {
                throw new TypeException("cannot convert to array: " + t1);
            }
            return arg.accept(new TeaJayArrayVisitor(this), null);
        } else {
            throw new TypeException("cannot convert to array: " + t1);
        }
    }

    Type toJavaArrayType(Type t1) {
        if (t1 instanceof Type.ClassType && ((Type.ClassType) t1).name.equals("teajay/lang/Array")) {
            Type.ClassType type = ((Type.ClassType) t1);
            if (type.outerType != null) {
                System.err.println("Warning: Outer types not allowed for Array");
            }
            Type arg = object;
            if (type.typeArgs != null && type.typeArgs.size() == 1) {
                arg = type.typeArgs.get(0);
            } else if (type.typeArgs != null && !type.typeArgs.isEmpty()) {
                throw new TypeException("cannot convert to array: " + t1);
            }
            return arg.accept(new JavaArrayVisitor(context), null);
        } else {
            throw new TypeException("cannot convert to array: " + t1);
        }
    }

    Type toTJArrayType(FieldDescriptor array) {
        if (array == null) {
            throw new NullPointerException("array");
        }
        if (!array.isArray()) {
            throw new TypeException("expecting array descriptor");
        }
        Type.ClassType result = new Type.ClassType(null, array.getLeafElementType().getClassInfo().getClassName(), null);
        for (int i = array.getDimension() - 1; i >= 0; i--) {
            result = new Type.ClassType(null, "teajay/lang/Array", Arrays.asList((Type) result));
        }
        return result;
    }

    Type toTJArrayType(Type array) {
        if (array == null) {
            throw new NullPointerException("array");
        }
        if (!(array instanceof Type.ArrayType)) {
            throw new TypeException("expecting array type");
        }
        return array.accept(new TJToJavaTypeVisitor(context), null);
    }
}
