
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.Properties;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

/**
 * Put this as main class into your jar, and it can be started like any other
 * java program, by typing
 *     java -jar &lt; jarfile &gt;
 * The TeaJay-libraries must be installed correctly or be found in the class path.
 * META-INF/MANIFEST.MF:
 *     Class-Path: &lt;path&gt;/TeaJayUtils.jar &lt;path&gt;/TeaJayRuntime.jar
 * Additionally the jar must have a META-INF/TEAJAY.MF.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class TeaJayExecuteWrapper {

    private static final String[] teajayFiles = new String[]{
        "TeaJayUtils.jar", "TeaJayRuntime.jar"
    };

    public static void main(String[] args) throws Throwable {
        URL location = TeaJayExecuteWrapper.class.getProtectionDomain().getCodeSource().getLocation();
        ClassLoader loader = TeaJayExecuteWrapper.class.getClassLoader();
        File file = new File(location.toURI());
        if (file.isFile() && file.canRead() && file.getName().toLowerCase().endsWith(".jar")) {
            String[] newArgs = new String[args.length + 2];
            newArgs[0] = "-jar";
            newArgs[1] = file.toString();
            System.arraycopy(args, 0, newArgs, 2, args.length);
            try {
                loader.loadClass("teajay.runtime.Run");
                adjustClassPath(file);
                Class<?> run = loader.loadClass("teajay.runtime.Run");
                Method method = run.getMethod("main", String[].class);
                method.invoke(null, (Object) newArgs);
            } catch (ClassNotFoundException ex) {
                String env = System.getenv("TEAJAYHOME");
                if (env != null) {
                    File dist = new File(new File(env), "dist");
                    File lib = new File(dist, "lib");
                    String cp = "";
                    for (String str : teajayFiles) {
                        File jar = new File(lib, str);
                        if (!jar.canRead()) {
                            throw new Error("cannot read " + jar);
                        }
                        cp += new File(lib, str).getAbsoluteFile() + System.getProperty("path.separator");
                    }
                    cp = cp.substring(0, cp.length() - 1);
                    String[] newerArgs = new String[newArgs.length + 4];
                    newerArgs[0] = new File(new File(System.getProperty("java.home"), "bin"), "java").getAbsolutePath();
                    newerArgs[1] = "-cp";
                    newerArgs[2] = cp;
                    newerArgs[3] = "teajay.runtime.Run";
                    System.arraycopy(newArgs, 0, newerArgs, 4, newArgs.length);
                    ProcessBuilder builder = new ProcessBuilder(newerArgs);
                    Process process = builder.inheritIO().start();
                    newerArgs = newArgs = args = null;
                    cp = null;
                    file = dist = lib = null;
                    ex = null;
                    loader = null;
                    location = null;
                    System.gc();
                    process.waitFor();
                } else {
                    throw new Error("unable to locate TeaJay runtime");
                }
            }
        } else {
            throw new Error("not beeing executed inside a jar");
        }
    }

    private static void adjustClassPath(File jarFile) throws ZipException, IOException {
        try (ZipFile zipFile = new ZipFile(jarFile)) {
            ZipEntry entry = zipFile.getEntry("META-INF/MANIFEST.MF");
            if (entry == null) {
                throw new Error("jar file doesn't specify META-INF/MANIFEST.MF");
            }
            Properties p = new Properties();
            try (InputStream in = zipFile.getInputStream(entry);) {
                p.load(in);
            }
            String jarClassPath = p.getProperty("Class-Path");
            if (jarClassPath != null) {
                String[] cps = jarClassPath.split(" ");
                for (String cp : cps) {
                    if (!cp.equals("")) {
                        String rp = System.getProperty("java.class.path", "");
                        System.setProperty("java.class.path", rp + (!rp.equals("") ? System.getProperty("path.separator") : "") + cp);
                    }
                }
            }
        }
    }
}
