/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.lang;

import teajay.util.TJType;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 * Type definition of TeaJay's semi builtin type List.
 * All references to == refer to TeaJay's interpretation of the == operator.
 * In Java terms that would be the invocation of TJObjects.equals(Object, Object).
 * @author Michael Kußmann <michi@outx.de>
 */

@TJType(type = TeaJayAttribute.Type.Typedef)
public abstract class List<T> extends TJObject implements Cloneable, RandomAccess<T>, Collection<T> {
    /**
     * Creates an empty list.
     */
    public List() {}
    /**
     * Creates a new List with <tt>head</tt> as head and <tt>tail</tt> as tail.
     * @param head head of the new List
     * @param tail tail of the new List
     */
    public List(T head, List<? extends T> tail) {}
    /**
     * Joins all lists in the argument list.
     * Warning: Cannot be used with List<List<...>> types.
     * @param lists lists to be joined
     */
    public List(List<? extends T>... lists) {}
     /**
     * Creates a list prefilled with values.
     * Warning: Cannot be used with List<List<...>> types.
     */
    public List(T... values) {}

    /**
     * Adds value to the list.
     * @param value item to be added
     */
    @Override
    public abstract void add(T value);
    /**
     * Adds all elements in <tt>vals</tt> to the end of this list.
     * @param vals any collection
     */
    public abstract void addAll(TJIterable<? extends T> vals);

    /**
     * Removes the item at position index from the list.
     * @param index the index of the item to be removed
     * @return the removed item
     * @throws IndexOutOfBoundsException if index >= List.size() or index < 0 or index > Integer.MAX_VALUE or index.isInteger() == false
     */
    public abstract T remove(Number index) throws IndexOutOfBoundsException;

    /**
     * Removes all elements e for which e == elem is true from the list.
     * @param elem prototype of the elements to be removed
     * @return true if and only if at least one element has been removed from the list.
     */
    @Override
    public abstract Boolean removeAll(T elem);
    
    /**
     * Retrieves the element at index from the list.
     * @param index the index in the list
     * @return the element at position index
     * @throws IndexOutOfBoundsException if index >= List.size() or index < 0 or index > Integer.MAX_VALUE or index.isInteger() == false
     */
    @Override
    public abstract T get(Number index) throws IndexOutOfBoundsException;

    /**
     * Sets the 
     * @param index
     * @param value
     * @throws IndexOutOfBoundsException if index < 0 or index > Integer.MAX_VALUE or index.isInteger() == false
     */
    @Override
    public abstract void set(Number index, T value) throws IndexOutOfBoundsException;
    
    /**
     * Inserts value into the specified index. All following elements will be moved to the right.
     * @param index index where to insert value
     * @param value the value
     * @throws IndexOutOfBoundsException if index < 0 or index > Integer.MAX_VALUE or index.isInteger() == false
     */
    public abstract void insert(Number index, T value) throws IndexOutOfBoundsException;

    /**
     * Clones this list. (no deep clone)
     * @return a copy of this list
     */
    @Override
    public abstract List<T> clone();

    @Override
    public abstract TJIterator<T> iterator();

    /**
     * Searches for obj in this list.
     * @param obj the item to be looked for
     * @return the index of obj if found, -1 otherwise
     */
    public abstract Number indexOf(T obj);

    /**
     * @return true if and only if this list does not contain any items
     */
    public abstract Boolean isEmpty();

    public abstract T operator_get(Number index) throws IndexOutOfBoundsException;

    public abstract void operator_set(Number index, T value) throws IndexOutOfBoundsException;

    /**
     * @return the number of elements in this list
     */
    @Override
    public abstract Number size();

    /**
     * Creates a new list that contains the element at fromIndex and all following elements of this list.
     * @param fromIndex first index (inclusive)
     * @return a new list
     */
    public abstract List<T> subList(Number fromIndex) throws IndexOutOfBoundsException;
    
    /**
     * Creates a new list that contains the element at fromIndex and all following elements until endIndex (exclusive).
     * @param fromIndex first index (inclusive)
     * @param endIndex last index (exclusive)
     * @return a new list
     */
    public abstract List<T> subList(Number fromIndex, Number endIndex) throws IndexOutOfBoundsException;
    
    /**
     * Removes all items from this list.
     */
    @Override
    public abstract void clear();
    
    /**
     * @param elem the element to be looked for
     * @return true if and only if this list contains at least one item e for which e == elem is true
     */
    @Override
    public abstract Boolean contains(T elem);
}