/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.lang;

import java.io.Serializable;
import java.lang.ref.Reference;
import java.lang.ref.SoftReference;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.MathContext;
import java.math.RoundingMode;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import static teajay.util.JavaNumberParser.*;
import java.util.Objects;
import teajay.util.TJType;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 * Implementation of TeaJay's Number type. This (and String, Boolean) will be
 * the only types that cannot have another implementation than the one provided
 * here. Thus this file represents the typedef and type implementation of Number
 * simultaneously.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
@TJType(accessType = TeaJayAttribute.AccessType.Final)
public final class Number extends TJObject implements Serializable, TJComparable<Number>, Comparable<Number> {

    private static final long serialVersionUID = 1L;

    /**
     * Caching all results of $valueOf(String) as this method is used by the
     * compiler to push TeaJay number literals onto the stack.
     */
    private static class NumberCache {

        NumberCache() {
        }
        private final Map<String, Reference<Number>> cache = Collections.synchronizedMap(new HashMap<String, Reference<Number>>(1024));

        /**
         *
         * @param number
         * @return
         */
        public Number get(String number) {
            Reference<Number> ref = cache.get(number);
            Number value;
            if (ref == null) {
                value = parseTJNumber(number);
                cache.put(number, new SoftReference<>(value));
            } else {
                value = ref.get();
                if (value == null) {
                    value = parseTJNumber(number);
                    cache.put(number, new SoftReference<>(value));
                }
            }
            return value;
        }
    }
    private static final NumberCache literalCache = new NumberCache();
    private static final Number[] intCache = new Number[512];
    public static final Number ZERO = new Number(0, 1);
    public static final Number ONE = new Number(1, 1);
    public static final Number M_ONE = new Number(-1, 1);
    public static final Number BYTE_MAX_VAL = new Number(BigInteger.valueOf(Byte.MAX_VALUE), BigInteger.ONE);
    public static final Number BYTE_MIN_VAL = new Number(BigInteger.valueOf(Byte.MIN_VALUE), BigInteger.ONE);
    public static final Number SHORT_MAX_VAL = new Number(BigInteger.valueOf(Short.MAX_VALUE), BigInteger.ONE);
    public static final Number SHORT_MIN_VAL = new Number(BigInteger.valueOf(Short.MIN_VALUE), BigInteger.ONE);
    public static final Number INT_MAX_VAL = new Number(BigInteger.valueOf(Integer.MAX_VALUE), BigInteger.ONE);
    public static final Number INT_MIN_VAL = new Number(BigInteger.valueOf(Integer.MIN_VALUE), BigInteger.ONE);
    public static final Number LONG_MAX_VAL = new Number(BigInteger.valueOf(Long.MAX_VALUE), BigInteger.ONE);
    public static final Number LONG_MIN_VAL = new Number(BigInteger.valueOf(Long.MIN_VALUE), BigInteger.ONE);
    private static final BigInteger TEN = BigInteger.valueOf(10);
    /**
     * The sign will always be carried in the numerator.
     */
    private final BigInteger num;
    /**
     * Denom must not be 0 at any given time.
     */
    private final BigInteger denom;

    public Number(Number a) {
        this.num = a.num;
        this.denom = a.denom;
    }

    /**
     * Defaults to zero.
     */
    public Number() {
        this.num = BigInteger.ZERO;
        this.denom = BigInteger.ONE;
    }
    
    public Number(BigInteger integer) {
        this(integer, BigInteger.ONE);
    }

    public Number(BigInteger num, BigInteger denom) throws ArithmeticException {
        if (num == null || denom == null) {
            throw new NullPointerException(num == null ? "num" : "denom");
        }
        if (denom.signum() == 0) {
            throw new ArithmeticException("divison by zero");
        }
        //Normalized representation: The sign is always carried in the numerator.
        if (denom.signum() < 0) {
            num = num.negate();
            denom = denom.abs();
        }
        if (denom.equals(BigInteger.ONE)) {
            this.num = num;
            this.denom = BigInteger.ONE;
        } else {
            BigInteger gcd = num.gcd(denom);
            this.num = num.divide(gcd);
            this.denom = denom.divide(gcd);
        }
    }

    private Number(long num, long denom) {
        this(BigInteger.valueOf(num), BigInteger.valueOf(denom));
    }

    private Number(int num, int denom) {
        this(BigInteger.valueOf(num), BigInteger.valueOf(denom));
    }

    public static Boolean operator_less(Number a, Number b) {
        return a.compareTo(b) < 0;
    }

    public static Boolean operator_leq(Number a, Number b) {
        return a.compareTo(b) <= 0;
    }

    public static Boolean operator_greater(Number a, Number b) {
        return a.compareTo(b) > 0;
    }

    public static Boolean operator_geq(Number a, Number b) {
        return a.compareTo(b) >= 0;
    }
    /**
     * @see #add(teajay.lang.Number) 
     */
    public static Number operator_add(Number a, Number b) {
        return a.add(b);
    }
    /**
     * addition
     * @param a addend
     * @return this + a
     */
    public Number add(Number a) {
        boolean azero = a.isZero();
        if (!azero && !isZero()) {
            if (a.denom.equals(denom)) {
                return new Number(num.add(a.num), denom);
            } else {
                BigInteger a_num = a.num.multiply(denom);
                BigInteger a_denom = a.denom.multiply(denom);
                BigInteger b_num = num.multiply(a.denom);

                return new Number(a_num.add(b_num), a_denom);
            }
        } else if (azero) {
            return this;
        } else {
            return a;
        }
    }
    /**
     * @see #negate() 
     */
    public static Number operator_sub(Number a) {
        return a.negate();
    }
    /**
     * @see #sub(teajay.lang.Number) 
     */
    public static Number operator_sub(Number a, Number b) {
        return a.sub(b);
    }
    /**
     * subtraction
     * @param a subtrahend
     * @return this - a
     */
    public Number sub(Number a) {
        boolean azero = a.isZero();
        if (!azero && !isZero()) {
            if (a.denom.equals(denom)) {
                return new Number(num.subtract(a.num), denom);
            } else {
                BigInteger a_num = a.num.multiply(denom);
                BigInteger a_denom = a.denom.multiply(denom);
                BigInteger b_num = num.multiply(a.denom);

                return new Number(b_num.subtract(a_num), a_denom);
            }
        } else if (azero) {
            return this;
        } else {
            return a.negate();
        }
    }
    /**
     * @see #mult(teajay.lang.Number) 
     */
    public static Number operator_mult(Number a, Number b) {
        return a.mult(b);
    }
    /**
     * multiplication
     * @param a factor
     * @return this * a
     */
    public Number mult(Number a) {
        if (a.isOne()) {
            return this;
        } else if (isOne()) {
            return a;
        } else if (a.isZero() || isZero()) {
            return ZERO;
        }
        return new Number(a.num.multiply(num), a.denom.multiply(denom));
    }
    
    /**
     * invert
     * @return this.pow(-1)
     * @throws ArithmeticException if this == 0
     */
    public Number invert() throws ArithmeticException {
        if (!isZero()) {
            return new Number(denom, num);
        } else {
            throw new ArithmeticException("divison by zero");
        }
    }
    /**
     * @see #div(teajay.lang.Number) 
     */
    public static Number operator_div(Number a, Number b) throws ArithmeticException {
        return a.div(b);
    }
    /**
     * division
     * @param a dividend
     * @return this / a
     * @throws ArithmeticException if a is zero
     */
    public Number div(Number a) throws ArithmeticException {
        if (a.isZero()) {
            throw new ArithmeticException("divison by zero");
        } else if (a.isOne()) {
            return this;
        } else if (isOne()) {
            return a.invert();
        } else if (isZero()) {
            return ZERO;
        }
        return new Number(num.multiply(a.denom), denom.multiply(a.num));
    }
    
    /**
     * exponentiation.
     * @param exponent the exponent must be in signed 32 bit integer range
     * @return this "to the power of" exponent
     */
    public Number pow(Number exponent) {
        if (!exponent.isInteger()) {
            throw new ArithmeticException("pow expects an integer");
        }
        BigInteger a = exponent.num.abs();
        if (a.equals(BigInteger.ZERO)) {
            return ONE;
        }
        if (a.bitLength() > 31) {
            throw new ArithmeticException("pow is not implemented for exponents whose absolute value is greater than 2^31 - 1");
        }
        int pow = a.intValue();
        if (exponent.intSignum() >= 0) {
            return new Number(num.pow(pow), denom.pow(pow));
        } else {
            return new Number(denom.pow(pow), num.pow(pow));
        }
    }
    /**
     * modulo
     * @param n dividend
     * @return the result of the euclidean division this : n
     * @throws ArithmeticException if n is zero
     */
    public Number mod(Number n) throws ArithmeticException {
        if (n.intSignum() == 0) {
            throw new ArithmeticException("divison by zero");
        }
        if (this.isInteger() && n.isInteger()) {
            return new Number(num.remainder(n.num), BigInteger.ONE);
        } else { //This is what java does for float and double modulo
            Number a = this;
            int sign = a.intSignum();
            Number b = n;
            if (sign < 0) {
                a = a.abs();
            }
            if (b.intSignum() < 0) {
                b = b.abs();
            }
            while (a.compareTo(b) >= 0) {
                a = a.sub(b);
            }
            return sign < 0 ? a.negate() : a;
        }
    }
    /**
     * bitwise and.
     * Both numbers must be integer.
     * @param other integer
     * @return this & other
     */
    public Number and(Number other) {
        if (this.isInteger() && other.isInteger()) {
            return new Number(num.and(other.num), BigInteger.ONE);
        } else {
            throw new ArithmeticException("bitwise and can only operate on integers");
        }
    }
    /**
     * bitwise or.
     * Both numbers must be integer.
     * @param other integer
     * @return this | other
     */
    public Number or(Number other) {
        if (this.isInteger() && other.isInteger()) {
            return new Number(num.or(other.num), BigInteger.ONE);
        } else {
            throw new ArithmeticException("bitwise or can only operate on integers");
        }
    }
    /**
     * bitwise negation.
     * Both numbers must be integer.
     * @return ~this
     */
    public Number not() {
        if (this.isInteger()) {
            return new Number(num.not(), BigInteger.ONE);
        } else {
            throw new ArithmeticException("bitwise not can only operate on integers");
        }
    }
    /**
     * bitwise exclusive or.
     * Both numbers must be integer.
     * @param other integer
     * @return this ^ other
     */
    public Number xor(Number other) {
        if (this.isInteger() && other.isInteger()) {
            return new Number(num.xor(other.num), BigInteger.ONE);
        } else {
            throw new ArithmeticException("bitwise xor can only operate on integers");
        }
    }
    /**
     * bitwise left shift.
     * Both numbers must be integer.
     * @param bitCount positive signed 32 bit integer
     * @return this &lt;&lt; bitCount
     */
    public Number shiftLeft(Number bitCount) {
        if (this.isInteger() && bitCount.isInteger()) {
            if (bitCount.intSignum() < 0) {
                throw new ArithmeticException("shiftLeft expects a positive bit count");
            }
            return new Number(num.shiftLeft(bitCount.$int_val()), BigInteger.ONE);
        } else {
            throw new ArithmeticException("shiftLeft can only operate on integers");
        }
    }

    /**
     * signed bitwise right schift.
     * Both numbers must be integer.
     * @param bitCount positive signed 32 bit integer
     * @return this &gt;&gt; bitCount
     */
    public Number shiftRight(Number bitCount) {
        if (this.isInteger() && bitCount.isInteger()) {
            if (bitCount.intSignum() < 0) {
                throw new ArithmeticException("shiftRight expects a positive bit count");
            }
            return new Number(num.shiftRight(bitCount.$int_val()), BigInteger.ONE);
        } else {
            throw new ArithmeticException("shiftRight can only operate on integers");
        }
    }

    /**
     * Checks for this number beeing an integer.
     * @return true if and only if this Number represents an integer
     */
    public Boolean isInteger() {
        return denom.equals(BigInteger.ONE);
    }
    
    /**
     * numerator() : denominator()
     * @return the quotient of the euclidean division numerator() : denominator()
     */
    public Number toInteger() {
        if (isInteger()) {
            return this;
        }
        return new Number(num.divide(denom), BigInteger.ONE);
    }
    /**
     * Checks for this number beeing negative.
     * @return true if and only if this &lt; 0
     */
    public Boolean isNegative() {
        return intSignum() < 0;
    }
    
    /**
     * floor
     * @return floor(this)
     */
    public Number floor() {
        BigInteger[] dar = num.divideAndRemainder(denom);
        BigInteger integer = dar[0];
        BigInteger remainder = dar[1];

        if (remainder.signum() == 0) {
            return this;
        }
        switch (intSignum()) {
            case -1:
                return new Number(integer.subtract(BigInteger.ONE), BigInteger.ONE);
            case 0:
                return ZERO;
            default:
                return new Number(integer, BigInteger.ONE);
        }
    }
    /**
     * ceil
     * @return ceil(this)
     */
    public Number ceil() {
        BigInteger[] dar = num.divideAndRemainder(denom);
        BigInteger integer = dar[0];
        BigInteger remainder = dar[1];

        if (remainder.signum() == 0) {
            return this;
        }
        switch (intSignum()) {
            case -1:
                return new Number(integer, BigInteger.ONE);
            case 0:
                return ZERO;
            default:
                return new Number(integer.add(BigInteger.ONE), BigInteger.ONE);
        }
    }
    /**
     * absolute value
     * @return |this|
     */
    public Number abs() {
        if (intSignum() >= 0) {
            return this;
        } else {
            return this.negate();
        }
    }
    /**
     * denominator
     * @return the denominator of this rational number
     */
    public Number denominator() {
        return new Number(denom, BigInteger.ONE);
    }
    
    /**
     * numerator
     * @return the numerator of this rational number
     */
    public Number numerator() {
        return new Number(num, BigInteger.ONE);
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public BigInteger $denom() {
        return denom;
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public BigInteger $num() {
        return num;
    }
    /**
     * numeric negation
     * @return (-1) * this
     */
    public Number negate() {
        if (intSignum() == 0) {
            return this;
        }
        return new Number(num.negate(), denom);
    }
    /**
     * Checks for a beeing divisible by this.
     * @param a 
     * @return (a / this).isInteger()
     * @throws ArithmeticException if this is zero
     */
    public Boolean divides(Number a) throws ArithmeticException {
        return a.div(this).isInteger();
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public static Number $valueOf(boolean a) {
        return a ? Number.ONE : Number.ZERO;
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public static Number $valueOf(long a) {
        if (a >= -256L && a <= 255L) {
            return $valueOf((int) a);
        }
        return new Number(a, 1);
    }

    /**
     * Solely for internal use. Caches small Numbers.
     *
     * @param a
     * @return
     */
    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public static Number $valueOf(int a) {
        if (a >= -256 && a <= 255) {
            int index;
            if (a < 0) {
                index = -a + 255;
            } else {
                index = a;
            }
            if (intCache[index] == null) {
                intCache[index] = new Number(a, 1);
            }
            return intCache[index];
        } else {
            return new Number(a, 1);
        }
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public static Number $valueOf(float a) {
        if (a == 0f) {
            return Number.ZERO;
        } else if (a == 1f) {
            return Number.ONE;
        }
        return parseBigDecimal(new BigDecimal(a));
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public static Number $valueOf(double a) {
        if (a == 0d) {
            return Number.ZERO;
        } else if (a == 1d) {
            return Number.ONE;
        }
        return parseBigDecimal(new BigDecimal(a));
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public static Number $valueOf(int num, int denom) {
        if (denom == 1) {
            return $valueOf(num);
        }
        return new Number(num, denom);
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public static Number $valueOf(long num, long denom) {
        if (denom == 1L) {
            return $valueOf(num);
        }
        return new Number(num, denom);
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public static Number $valueOf(Number num, Number denom) {
        return num.div(denom);
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public static Number $valueOf(BigInteger num, BigInteger denom) {
        return new Number(num, denom);
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public static Number $valueOf(String val) {
        //return parseTJNumber(val);
        return literalCache.get(val);
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public static Number $valueOf(BigDecimal val) {
        if (val.signum() == 0) {
            return Number.ZERO;
        }
        return parseBigDecimal(val);
    }

    private static Number parseBigDecimal(BigDecimal d) {
        BigInteger num;
        BigInteger denom;
        int scale = d.scale();
        /**
         * @see java.math.BigInteger#scale()
         */
        if (scale >= 0) {
            num = d.unscaledValue();
            if (scale == 0) {
                denom = BigInteger.ONE;
            } else {
                denom = TEN.pow(scale);
            }
        } else {
            num = d.unscaledValue().multiply(TEN.pow(-1 * scale));
            denom = BigInteger.ONE;
        }
        return new Number(num, denom);
    }

    /**
     * Parses numbers of the following format (n = digit).
     * (-+)?n+(.n*(e|E(+-)?n+)?)?
     *
     * @param str a String potentially representing a number
     * @return the number representation of str
     * @throws NumberFormatException if str is not a recognized number
     */
    public static Number parseNumber(String str) throws NumberFormatException {
        if (str == null) {
            throw new NullPointerException("number");
        }
        try {
            str = str.toLowerCase();
            if (str.contains("/")) {
                String[] split = str.split("/");
                if (split.length > 2) {
                    throw new NumberFormatException(str);
                }
                return parseNumber(split[0]).div(parseNumber(split[1]));
            } else if (str.contains(".") || str.contains("e")) {
                return parseBigDecimal(new BigDecimal(str));
            } else { // Integer
                return new Number(new BigInteger(str, 10), BigInteger.ONE);
            }
        } catch (NumberFormatException ex) {
            throw new NumberFormatException(str);
        }
    }
    /**
     * Parses only decimal integer numbers.
     * @param str String potentially representing an integer
     * @return the number representation of str
     * @throws NumberFormatException if str is not a recognized number
     */
    public static Number parseInteger(String str) throws NumberFormatException {
        return parseInteger(str, Number.$valueOf(10));
    }

    /**
     * Parses only integer numbers for radices 2 &le; radix &le; 36.
     *
     * @param str String potentially representing an integer
     * @param radix a radix between 2 and 36 inclusive
     * @return the number representation of str
     * @throws NumberFormatException if str is not a recognized number
     */
    public static Number parseInteger(String str, Number radix) throws NumberFormatException {
        if (str == null || radix == null) {
            throw new NullPointerException(radix == null ? "radix" : "number");
        }
        try {
            return new Number(new BigInteger(str, radix.$int_val()), BigInteger.ONE);
        } catch (NumberFormatException ex) {
            throw new NumberFormatException(str + " (integer)");
        }
    }

    /**
     * Parses java/teajay number literals.
     *
     * @param number
     * @return
     * @throws NumberFormatException
     */
    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public static Number parseTJNumber(String number) throws NumberFormatException {
        if (number == null) {
            throw new NullPointerException("number");
        }
        try {
            int index = number.indexOf("/");
            if (index == 0 || index == number.length() - 1) {
                throw new NumberFormatException(number);
            } else if (index > 0) {
                String num = number.substring(0, index);
                String denom = number.substring(index + 1);
                int sign = 1;
                if (num.startsWith("-")) {
                    sign *= -1;
                    num = num.substring(1);
                } else if (num.startsWith("+")) {
                    num = num.substring(1);
                }
                if (denom.startsWith("-")) {
                    sign *= -1;
                    denom = denom.substring(1);
                } else if (denom.startsWith("+")) {
                    denom = denom.substring(1);
                }
                return new Number(parseBigInteger(num, sign == 1), parseBigInteger(denom, true));
            } else {
                boolean sign = true;
                if (number.startsWith("-")) {
                    sign = false;
                    number = number.substring(1);
                } else if (number.startsWith("+")) {
                    number = number.substring(1);
                }

                try {
                    return new Number(parseBigInteger(number, sign), BigInteger.ONE);
                } catch (NumberFormatException ex) {
                    return parseBigDecimal(parseBigFloat(number, sign, MathContext.UNLIMITED));
                }
            }
        } catch (NumberFormatException ex) {
            throw new NumberFormatException(number);
        }
    }

    /**
     * Defaults to 16 digits.
     *
     * @return
     */
    public Number round() {
        BigDecimal a = new BigDecimal(num);
        BigDecimal b = new BigDecimal(denom);
        return parseBigDecimal(a.divide(b, MathContext.DECIMAL64));
    }
    /**
     * Rounds this number to precision digits.
     * @param precision number of digits
     * @return a rounded version of this number
     * @throws ArithmeticException 
     */
    public Number round(Number precision) throws ArithmeticException {
        return round(precision, RoundingMode.HALF_UP);
    }
    /**
     * Rounds this number to precision digits with rounding mode mode.
     * @param precision number of digits
     * @param mode a rounding mode
     * @return a rounded version of this number
     * @throws ArithmeticException if precision &lt; 0 or precision exceeds the signed 32-bit integer range
     */
    public Number round(Number precision, RoundingMode mode) throws ArithmeticException {
        if (!precision.isInteger() || precision.intSignum() < 1) {
            throw new ArithmeticException("precision must be an integer value greater than 0");
        }
        if (precision.num.bitLength() > 31) {
            throw new ArithmeticException("round does not support a precision greater than 2^31 - 1");
        }
        BigDecimal a = new BigDecimal(num);
        BigDecimal b = new BigDecimal(denom);
        BigDecimal rounded = a.divide(b, new MathContext(precision.$int_val(), mode));

        return parseBigDecimal(rounded);
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public boolean $boolean_val() {
        return this.intSignum() != 0;
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public byte $byte_val() {
        BigInteger intval;
        if (denom.equals(BigInteger.ONE)) {
            intval = num;
        } else {
            intval = num.divide(denom);
        }
        if (intval.bitLength() <= 7) {
            return intval.byteValue();
        } else {
            throw new ArithmeticException("8-bit signed integer range exceeded (byte)");
        }
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public short $short_val() {
        BigInteger intval;
        if (denom.equals(BigInteger.ONE)) {
            intval = num;
        } else {
            intval = num.divide(denom);
        }
        if (intval.bitLength() <= 15) {
            return intval.shortValue();
        } else {
            throw new ArithmeticException("16-bit signed integer range exceeded (short)");
        }
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public char $char_val() {
        BigInteger intval;
        if (denom.equals(BigInteger.ONE)) {
            intval = num;
        } else {
            intval = num.divide(denom);
        }
        if (intval.bitLength() <= 16 && intval.signum() >= 0) {
            return (char) intval.intValue();
        } else {
            throw new ArithmeticException("16-bit unsigned integer range exceeded (char)");
        }
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public int $int_val() {
        BigInteger intval;
        if (denom.equals(BigInteger.ONE)) {
            intval = num;
        } else {
            intval = num.divide(denom);
        }
        if (intval.bitLength() <= 31) {
            return intval.intValue();
        } else {
            throw new ArithmeticException("32-bit signed integer range exceeded (int)");
        }
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public long $long_val() {
        BigInteger intval;
        if (denom.equals(BigInteger.ONE)) {
            intval = num;
        } else {
            intval = num.divide(denom);
        }
        if (intval.bitLength() <= 63) {
            return intval.longValue();
        } else {
            throw new ArithmeticException("64-bit signed integer range exceeded (long)");
        }
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public double $double_val() {
        BigDecimal a = new BigDecimal(num);
        BigDecimal b = new BigDecimal(denom);
        return a.divide(b, MathContext.DECIMAL64).doubleValue();
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public float $float_val() {
        BigDecimal a = new BigDecimal(num);
        BigDecimal b = new BigDecimal(denom);
        return a.divide(b, MathContext.DECIMAL32).floatValue();
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public BigDecimal $bigdec_val(MathContext context) {
        BigDecimal a = new BigDecimal(num);
        BigDecimal b = new BigDecimal(denom);
        return a.divide(b, context);
    }

    @TJType(accessType = TeaJayAttribute.AccessType.Hidden)
    public BigInteger $bigint_val() {
        return num.divide(denom);
    }

    private boolean isOne() {
        return num.equals(BigInteger.ONE) && denom.equals(BigInteger.ONE);
    }

    private boolean isZero() {
        return num.signum() == 0;
    }

    private int intSignum() {
        return num.signum() * denom.signum();
    }

    public Number signum() {
        int signum = intSignum();
        switch (signum) {
            case -1:
                return M_ONE;
            case 0:
                return ZERO;
            case 1:
                return ONE;
            default:
                throw new AssertionError("signum out of bounds: " + signum);
        }
    }

    @Override
    public Boolean operator_eq(Object obj) {
        if (obj instanceof Number) {
            Number other = (Number) obj;
            return this.denom.equals(other.denom) && this.num.equals(other.num);
        }
        return false;
    }

    @Override
    public int compareTo(Number n) {
        int cmp;
        if (denom.equals(n.denom)) {
            cmp = num.compareTo(n.num);
        } else {
            BigInteger a = this.num.multiply(n.denom);
            BigInteger b = this.denom.multiply(n.num);
            cmp = a.compareTo(b);
        }
        return cmp;
    }

    @Override
    public Number tjCompareTo(Number other) {
        int cmp = this.compareTo(other);
        switch (cmp) {
            case -1:
                return M_ONE;
            case 0:
                return ZERO;
            case 1:
                return ONE;
            default:
                throw new AssertionError("compare out of bounds: " + cmp);
        }
    }

    @Override
    public Number tjHashCode() {
        int hash = 7;
        hash = 37 * hash + Objects.hashCode(this.num);
        hash = 43 * hash + Objects.hashCode(this.denom);
        return Number.$valueOf(hash);
    }

    @Override
    public String toString() {
        if (isInteger()) {
            return num.toString();
        } else {
            return toDecimalString();
        }
    }

    public String toFractionString() {
        if (isInteger()) {
            return num.toString();
        } else {
            return num + "/" + denom;
        }
    }

    public String toDecimalString() {
        return toDecimalString(MathContext.DECIMAL128);
    }

    public String toDecimalString(Number precision) {
        return toDecimalString(new MathContext(precision.$int_val()));
    }

    public String toDecimalString(Number precision, RoundingMode rmode) {
        return toDecimalString(new MathContext(precision.$int_val(), rmode));
    }

    private String toDecimalString(MathContext context) {
        return $bigdec_val(context).toString();
    }

    public String toEngineeringString() {
        return toEngineeringString(MathContext.DECIMAL32);
    }

    public String toEngineeringString(Number precison) {
        return toEngineeringString(new MathContext(precison.$int_val()));
    }

    public String toEngineeringString(Number precison, RoundingMode rmode) {
        return toEngineeringString(new MathContext(precison.$int_val(), rmode));
    }

    private String toEngineeringString(MathContext context) {
        return $bigdec_val(context).toEngineeringString();
    }
}
