/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.util.impl;

import java.math.BigDecimal;
import java.math.MathContext;
import teajay.util.tjattribute.TeaJayAttribute;
import teajay.lang.Number;
import teajay.util.TJType;

/**
 * Nur ein prototyp!
 * @author Michael Kußmann <michi@outx.de>
 */
@TJType(type = TeaJayAttribute.Type.Impl)
public final class MathImpl extends teajay.util.Math {
    private static final Number defaultPrecision = Number.$valueOf(16);
    private static final Number TWO = Number.$valueOf(2);
    private static final Number ONE = Number.ONE;
    private static final Number ZERO = Number.ZERO;
    private static final Number TEN = Number.$valueOf(10);
    
    private MathImpl(Math rthis) {
    }
    
    public static Number exp(Number x) {
        return exp(x, defaultPrecision);
    }

    public static Number exp(Number x, Number precision) {
        //Number threshold = TEN.pow(precision.negate());
        Number doublePrecision = precision.mult(TWO);
        Number result = Number.ZERO;
        for (int i = 0; i < 1000; i++) {
            result = result.add(exp_elem(x, Number.$valueOf(i), doublePrecision)).round(doublePrecision);
        }
        return result;
    }
    
    private static Number exp_elem(Number x, Number i, Number doublePrecision) {
        return x.pow(i).round(doublePrecision).div(faculty(i)).round(doublePrecision);
    }
    
    public static Number pow(Number x, Number n) {
        return pow(x, n, defaultPrecision);
    }
    
    public static Number pow(Number x, Number n, Number precision) {
        if (x.signum().equals(ZERO)) {
            return ZERO;
        } else if (x.equals(ONE)) {
            return ONE;
        }
        MathContext doublePrecision = new MathContext(precision.$int_val() * 2);
        BigDecimal tmp = x.$bigdec_val(doublePrecision);
        
        return extractRoot(tmp.pow(n.numerator().$int_val(), doublePrecision), n.denominator(), doublePrecision);
    }
    
    private static Number extractRoot(BigDecimal x, Number nn, MathContext context) {
        if (x.signum() < 0 && !nn.$num().testBit(0)) {
            //IEEE: NaN
            throw new ArithmeticException("Expecting non negative radicand with even denominators.");
        }
        BigDecimal threshold = BigDecimal.ONE.divide(BigDecimal.TEN.pow(context.getPrecision()));
        BigDecimal last = BigDecimal.ZERO;
        BigDecimal n = new BigDecimal(nn.$num());
        
        BigDecimal y = x.add(BigDecimal.ONE).divide(n, context);
        while (y.subtract(last).abs().compareTo(threshold) > 0) {
            last = y;
            y = n.subtract(BigDecimal.ONE, context).multiply(y.pow(n.intValue(), context), context).add(x, context).divide(n.multiply(y.pow(n.subtract(BigDecimal.ONE).intValue(), context), context), context);
        }
        return Number.$valueOf(y.round(new MathContext(context.getPrecision() / 2)));
    }    
   
    public static Number faculty(Number x) {
        if (!x.isInteger() && !x.isNegative()) {
            throw new IllegalArgumentException("expecting positive integer argument");
        }
        Number result = Number.ONE;
        for (; x.compareTo(Number.ZERO) > 0; x = x.sub(Number.ONE)) {
            result = result.mult(x);
        }
        
        return result;
    }
    
    public static Number ln(Number a) {
        return ln(a, defaultPrecision);
    }

    public static Number ln(Number a, Number precision) {
        if (a.signum().compareTo(ZERO) <= 0) {
            throw new ArithmeticException("ln expects positive numbers");
        }
        Number threshold = TEN.pow(precision.negate());
        Number doublePrecision = precision.mult(TWO);
        Number x = a;
        int m = 0;
        Number sqrt_2 = sqrt(TWO, precision);
        Number sqrt_2_inv = sqrt_2.invert();
        int cmp = x.compareTo(sqrt_2);
        if (cmp > 0) {
            m = 2;
            while ((x = x.div(TWO)).compareTo(sqrt_2) > 0) {
                m += 2;
            }
        } else if (cmp < 0) {
            m = -2;
            while ((x = x.mult(TWO)).compareTo(sqrt_2_inv) < 0) {
                m -= 2;
            }
        }
        Number result = ZERO;
        Number next;
        for (Number i = ZERO; ln_error(x, next = i.add(ONE)).compareTo(threshold) > 0; i = next) {
            result = result.add(ln_elem(x, i)).round(doublePrecision);
        }
        return result.add(Number.$valueOf(m).mult(ln_sqrt_2(threshold, doublePrecision, sqrt_2)));
    }

    private static Number ln_sqrt_2(Number threshold, Number doublePrecision, Number sqrt_2) {
        Number x = sqrt_2;
        Number result = ZERO;
        Number next;
        for (Number i = ZERO; ln_error(x, next = i.add(ONE)).compareTo(threshold) > 0; i = next) {
            result = result.add(ln_elem(x, i)).round(doublePrecision);
        }
        return result;
    }

    private static Number ln_elem(Number x, Number k) {
        Number k21 = TWO.mult(k).add(ONE);
        Number a = TWO.div(k21);
        Number b = x.sub(ONE).div(x.add(ONE)).pow(k21);
        return a.mult(b);
    }

    private static Number ln_error(Number x, Number n) {
        Number s1 = x.sub(ONE);
        Number a = s1.pow(TWO).div(TWO.mult(x.abs()));
        Number b = s1.div(x.add(ONE)).pow(TWO.mult(n));
        return a.mult(b);
    }
    
    public static Number sqrt(Number x){
        return sqrt(x, defaultPrecision);
    }
    
    public static Number sqrt(Number x, Number precision) {
        Number threshold = TEN.pow(precision.negate());
        Number precisionPlusOne = precision.add(Number.ONE);
        Number tmp = x.add(Number.ONE).div(TWO);
        Number last = ZERO;
        while (tmp.sub(last).abs().compareTo(threshold) > 0) {
            last = tmp;
            tmp = tmp.add(x.div(tmp)).div(TWO).round(precisionPlusOne);
        }
        return tmp.round(precision);
    }
}
