/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package teajay.util.io;

import teajay.runtime.exceptions.DuplicateQualifiedIdentifierException;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.nio.file.Path;
import java.util.HashSet;
import java.util.Set;
import teajay.bytecode.ClassAccessFlag;
import teajay.bytecode.ClassFile;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.runtime.exceptions.ResolveException;
import teajay.util.io.AbstractResourceFinder.QID;

/**
 * Tool for class resolving.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class ClassFinder extends AbstractResourceFinder<ClassFileRef> {

    private static final String[] primitives = {"null", "void", "boolean", "byte", "short", "char", "int", "long", "float", "double"};
    private static ClassFinder instance;
    private final FilesVisitor classFindVisitor = new ClassFindVisitor();
    private final Set<String> strongPackages = new HashSet<>();

    private class ClassFindVisitor implements FilesVisitor {

        private String getClassName(String filename) {
            int index = filename.toLowerCase().lastIndexOf(".class");
            if (index < 0) {
                throw new Error("illegal invokation of getClassName(String)");
            }
            return filename.substring(0, index);
        }

        @Override
        public void visit(String javaPackage, String filename, Input in) throws StopScan, DuplicateQualifiedIdentifierException {
            String className = getClassName(filename);
            String qID = (javaPackage != null ? javaPackage + "/" : "") + className;
            ClassFileRef ref;
            if (javaPackage != null && strongPackages.contains(javaPackage)) {
                ref = new StrongClassFileRef(in, qID);
            } else {
                ref = new SoftClassFileRef(in, qID);
            }
            putRef(javaPackage, className, ref, false);
        }

        @Override
        public void failed(String file, IOException reason) throws StopScan {
            if (!file.toLowerCase().endsWith("sunrsasign.jar")) {
                System.err.println("[ClassFinder, I/O Error] " + reason.getLocalizedMessage());
            }
        }

        @Override
        public boolean accept(Path file) {
            String name = file.getFileName().toString();
            return name.toLowerCase().endsWith(".class") && name.length() > 6 && !name.equals("module-info.class");
        }
    }

    private ClassFinder() {
        try {
            for (String primitive : primitives) {
                addClassFile(new ClassFile(primitive, (String) null, ClassAccessFlag.PUBLIC, ClassAccessFlag.ABSTRACT));
            }
        } catch (ByteCodeException | DuplicateQualifiedIdentifierException ex) {
            throw new Error("Failed to initialize primitive dummies: " + ex.getMessage());
        }
        strongPackages.add("teajay/lang");
    }

    public synchronized static ClassFinder instance() {
        if (instance == null) {
            instance = new ClassFinder();
        }
        return instance;
    }

    /**
     * Adds new paths to the finder.
     *
     * @param paths an array containing path strings in the format PATH:PATH:...
     */
    public void addPaths(String... paths) throws DuplicateQualifiedIdentifierException {
        super.addPaths(classFindVisitor, true, paths);
    }

    /**
     * Adds new paths to the finder.
     *
     * @param paths an array containing path strings in the format PATH:PATH:...
     */
    public void addPaths(Path... paths) throws DuplicateQualifiedIdentifierException {
        super.addPaths(classFindVisitor, true, paths);
    }

    public synchronized void addClassFile(ClassFile file) throws DuplicateQualifiedIdentifierException {
        QID q = split(file.getThis().getClassName());
        putRef(q.javaPackage, q.simpleName, new DirectClassFileRef(file), true);
    }

    @Override
    public ClassFileRef find(String qid) {
        if (!qid.startsWith("[")) {
            ClassFileRef find = super.find(qid);
            /*if (find == null) {
                //final URL res = ClassLoader.getSystemResource(qid + ".class");
                final URL res = ClassFinder.class.getResource(qid + ".class");
                if (res != null) {
                    ClassFileRef ref = new SoftClassFileRef(new Input() {
                        @Override
                        public InputStream openStream() throws IOException {
                            return res.openStream();
                        }

                        @Override
                        public File getFile() {
                            return null;
                        }

                    }, qid);
                    QID q = split(qid);
                    putRef(q.javaPackage, q.simpleName, ref, false);
                    find = ref;
                }
            }*/
            return find;
        } else {
            ClassFileRef result = super.find(null, qid);
            if (result == null) {
                try {
                    ClassFile dummy = super.find("void").get();
                    ClassFileRef ref = new ArrayDummyClassFileRef(qid, dummy);
                    putRef(null, qid, ref, false);
                    return ref;
                } catch (ByteCodeException | IOException | ResolveException ex) {
                    throw new Error("internal error", ex);
                }
            } else {
                return result;
            }
        }
    }
}
