/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package teajay.util.io;

import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.DirectoryStream;
import java.nio.file.FileSystem;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class FileIterator implements Closeable {

    private static final DirectoryStream.Filter<Path> dirFilter = new DirectoryStream.Filter<Path>() {
        @Override
        public boolean accept(Path pathname) {
            return Files.isDirectory(pathname);
        }
    };
    private static final String pathSeparator = System.getProperty("path.separator");
    private static final String fileSeparator = System.getProperty("file.separator");
    private static final boolean isNotUnixStyleFileSeparator = !fileSeparator.equals("/");
    public static final String CLASSPATH = System.getProperty("java.class.path");
    public static final String BOOTPATH = System.getProperty("sun.boot.class.path");
    private final boolean filesOnly;
    private final List<Path> paths;
    private Set<Path> scannedDirs = new HashSet<>();
    private final Collection<Closeable> streams = Collections.synchronizedList(new LinkedList<Closeable>());
    private Path currentPath = null;

    private static String makePackageName(String path) {
        if (path == null) {
            return null;
        }
        if (isNotUnixStyleFileSeparator) {
            path = path.replace(fileSeparator, "/");
        }
        if (path.endsWith("/")) {
            path = path.substring(0, path.length() - 1);
        }
        return path;
    }

    public FileIterator(File... paths) {
        this(false, paths);
    }

    public FileIterator(Path... paths) {
        this(false, Arrays.asList(paths));
    }

    public FileIterator(String... paths) {
        this(false, paths);
    }

    public FileIterator(boolean filesOnly, File... paths) {
        Path[] tmp = new Path[paths.length];
        for (int i = 0; i < paths.length; i++) {
            tmp[i] = paths[i].toPath();
        }
        this.filesOnly = filesOnly;
        this.paths = Arrays.asList(tmp);
    }

    public FileIterator(boolean filesOnly, String... paths) {
        Path[] tmp = new Path[paths.length];
        for (int i = 0; i < paths.length; i++) {
            tmp[i] = new File(paths[i]).toPath();
        }
        this.filesOnly = filesOnly;
        this.paths = Arrays.asList(tmp);
    }

    private FileIterator(boolean filesOnly, List<Path> paths) {
        this.filesOnly = filesOnly;
        this.paths = paths;
    }

    public static FileIterator fromPathStrings(String... pathStrings) {
        return fromPathStrings(false, pathStrings);
    }

    public static FileIterator fromPathStrings(boolean filesOnly, String... pathStrings) {
        List<Path> paths = new ArrayList<>(pathStrings.length);
        for (String path : pathStrings) {
            for (String s : path.split(pathSeparator)) {
                paths.add(new File(s).toPath());
            }
        }
        return new FileIterator(filesOnly, paths);
    }

    public void scan(FilesVisitor visitor, boolean scanJars) {
        try {
            for (Path file : paths) {
                try {
                    this.currentPath = file;
                    if (filesOnly && Files.isRegularFile(file)) {
                        scan(file, visitor, scanJars);
                    } else if (!filesOnly) {
                        scan(file, visitor, scanJars);
                    }
                } catch (IOException ex) {
                    visitor.failed(file.toString(), ex);
                }
            }
        } catch (StopScan stop) {
        } finally {
            this.currentPath = null;
            this.scannedDirs = new HashSet<>();
        }
    }

    private void scan(final Path file, final FilesVisitor visitor, boolean scanJars) throws StopScan, IOException {
        String filename = file.getFileName().toString();
        if (Files.isDirectory(file)) {
            Path abs = file.toAbsolutePath();
            if (!scannedDirs.contains(abs)) {
                scannedDirs.add(abs);
                try (DirectoryStream<Path> files = Files.newDirectoryStream(file, new DirectoryStream.Filter<Path>() {
                    @Override
                    public boolean accept(Path entry) throws IOException {
                        return visitor.accept(entry);
                    }

                })) {

                    for (Path f : files) {
                        scanFile(f, visitor);
                    }

                    try (DirectoryStream<Path> dirs = Files.newDirectoryStream(file, dirFilter)) {
                        for (Path f : dirs) {
                            try {
                                scan(f, visitor, false);
                            } catch (IOException ex) {
                                visitor.failed(f.toString(), ex);
                            }
                        }
                    }
                }
            }
        } else if (scanJars && filename.toLowerCase().endsWith(".jar")) {
            final ZipFile zipFile = new ZipFile(file.toFile());
            streams.add(zipFile);
            Enumeration<? extends ZipEntry> entries = zipFile.entries();

            while (entries.hasMoreElements()) {
                final ZipEntry entry = entries.nextElement();
                File f = new File(entry.getName());
                if (visitor.accept(f.toPath())) {
                    String parent = makePackageName(f.getParent());
                    visitor.visit(parent != null ? parent : null, f.getName(), new Input() {
                        @Override
                        public InputStream openStream() throws IOException {
                            return zipFile.getInputStream(entry);
                        }

                        @Override
                        public String toString() {
                            return file + "!" + entry.getName();
                        }

                        @Override
                        public Path getPath() {
                            return null;
                        }
                    });
                }
            }
        } else if (visitor.accept(file)) {
            scanFile(file, visitor);
        }
    }

    private void scanFile(final Path file, FilesVisitor visitor) throws StopScan {
        if (Files.exists(file)) {
            Input input = new Input() {
                @Override
                public InputStream openStream() throws IOException {
                    return Files.newInputStream(file);
                }

                @Override
                public String toString() {
                    return file.toString();
                }

                @Override
                public Path getPath() {
                    return file;
                }
            };
            Path parent = currentPath.relativize(file).getParent();

            visitor.visit(parent != null ? makePackageName(parent.toString()) : null, file.getFileName().toString(), input);
        }
    }

    @Override
    public void close() throws IOException {
        IOException ex = null;
        synchronized (streams) {
            for (Closeable c : streams) {
                try {
                    c.close();
                } catch (IOException e) {
                    if (ex == null) {
                        ex = e;
                    }
                }
            }
        }
        if (ex != null) {
            throw ex;
        }
    }
}
