/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.util.io;

import teajay.runtime.exceptions.DuplicateQualifiedIdentifierException;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Path;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public abstract class SuffixFinder extends AbstractResourceFinder<Input> {
    
    private class SuffixFindVisitor implements FilesVisitor {

        private String getClassName(String filename) {
            int index = filename.toLowerCase().lastIndexOf(suffix);
            if (index < 0) {
                throw new RuntimeException("illegal invokation of getClassName(String)");
            }
            return filename.substring(0, index);
        }

        @Override
        public void visit(String javaPackage, String filename, Input in) throws StopScan, DuplicateQualifiedIdentifierException {
            putRef(javaPackage, getClassName(filename), in, false);
        }

        @Override
        public void failed(String file, IOException reason) throws StopScan {
            System.err.println("[" + getClass().getSimpleName() + ", I/O Error] " + reason.getLocalizedMessage());
        }

        @Override
        public boolean accept(Path file) {
            File f = file.toFile();
            if (file.toFile().isHidden() && ignoreHiddenFiles) {
                return false;
            }
            String name = file.getFileName().toString();
            return name.toLowerCase().endsWith(suffix) && name.length() > suffix.length();
        }
    }
    
    private final String suffix;
    private final boolean ignoreHiddenFiles;
    
    private final SuffixFindVisitor suffixVisitor = new SuffixFindVisitor();

    protected SuffixFinder(String suffix) {
        this.suffix = suffix;
        this.ignoreHiddenFiles = true;
    }
    
    protected SuffixFinder(String suffix, boolean ignoreHiddenFiles) {
        this.suffix = suffix;
        this.ignoreHiddenFiles = ignoreHiddenFiles;
    }

    public void addPaths(String... paths) throws DuplicateQualifiedIdentifierException {
        super.addPaths(suffixVisitor, false, paths);
    }
    
    public synchronized void addSourceFile(String javaPackage, String name, final File file) throws DuplicateQualifiedIdentifierException {
        if (suffixVisitor.accept(file.toPath())) {
            Input in = new Input() {
                @Override
                public InputStream openStream() throws IOException {
                    return new FileInputStream(file);
                }
                
                @Override
                public String toString() {
                    return file.toString();
                }
                
                @Override
                public Path getPath() {
                    return file.toPath();
                }
            };
            putRef(javaPackage, name, in, true);
        } else {
            throw new IllegalArgumentException("Invalid suffix: " + file.getName());
        }
    }
}
